package com.provisionnetworks.common;

import java.io.*;
import java.util.*;

import org.apache.log4j.Logger;

public class RDPSettings {
	
	Logger logger = Logger.getLogger(RDPSettings.class.getName());
	
    public static final int APPTYPE_APPLICATION = 1;
    public static final int APPTYPE_DESKTOP = 2;
    public static final int APPTYPE_CONTENTSERVER = 3;
    public static final int APPTYPE_CONTENTCLIENT = 4;
    
    // General tab
	public String computer;
	public String application;
    
    // Credentials tab
    public boolean enableSSO;
    public boolean enableKerberos;
    public String username;
    public String password;
    public String domain;
    public boolean savePassword;
    
    // Display tab
    public int screenModeId;
    public int desktopWidth;
    public int desktopHeight;
    public int sessionBitsPerPixel;
    public boolean displayConnectionBar;
    public boolean pinConnectionBar;
    public boolean smartSizing;
    
    // Local Resources tab
    public int audioMode;
    public int keyboardHook;
    public boolean redirectDrives;
    public boolean redirectPrinters;
    public boolean redirectComPorts;
    public boolean redirectSmartCards;
    public boolean redirectHandhelds;
    public boolean redirectUniversalPrinters;
    public boolean redirectClipboard;
    
    // Programs tab
    public String programFullPath;
    public String programWorkingDir;
    public boolean sessionSharing;
    public boolean seamless;
    
    // Experience tab
    public boolean enableWallpaper;
    public boolean enableFullWindowDrag;
    public boolean enableAnimation;
    public boolean enableThemes;
    public boolean enableFontSmoothing;
    public boolean enableDesktopComposition;
    public boolean enableBitmapCaching;
    public boolean enableAutoReconnection;
    public boolean bitmapAcceleration; 
    
    // Connectivity dialog.
    //LLMSETTINGS LLMSettings;
    public String sslGateway;
    public boolean useProxy;
    public String proxyServer;
    public String proxyBypass;
    
    // Non-displayable settings.
    public int appType;
    public boolean deleteFile;
    public String connectingText;
    public String serverName;
    public String serverAddr;
    public String farmName;
    public String ticket;
    public boolean compression;
    public boolean encryption;
    
    // Indicates settings have were encrypted.
    public boolean encrypted;
    
    public static final int DEFAULT_DESKTOP_WIDTH = 800;
    public static final int DEFAULT_DESKTOP_HEIGHT = 600;
    public static final int DEFAULT_SESSION_BPP = 16;
    
    public RDPSettings()
    {
    	init();
    }
    
    public void init()
    {
        computer = "";
        application = "";
        
        enableSSO = false;
        enableKerberos = false;
        username = "";
        password = "";
        domain = "";
        savePassword = false;

        screenModeId = 1;
        desktopWidth = DEFAULT_DESKTOP_WIDTH;
        desktopHeight = DEFAULT_DESKTOP_HEIGHT;
        sessionBitsPerPixel = DEFAULT_SESSION_BPP;
        displayConnectionBar = true;
        pinConnectionBar = true;
        smartSizing = false;

        audioMode = 0;
        keyboardHook = 0;
        redirectDrives = false;
        redirectPrinters = false;
        redirectComPorts = false;
        redirectSmartCards = false;
        redirectHandhelds = false;
        redirectUniversalPrinters = false;
        redirectClipboard = false;

        programFullPath = "";
        programWorkingDir = "";
        sessionSharing = false;
        seamless = false;

        enableWallpaper = false;
        enableFullWindowDrag = false;
        enableAnimation = false;
        enableThemes = false;
        enableFontSmoothing = false;
        enableDesktopComposition = false;
        enableBitmapCaching = false;
        enableAutoReconnection = true;
        bitmapAcceleration = false;
        
        sslGateway = "";
        useProxy = true;
        proxyServer = "";
        proxyBypass = "";
        
        appType = 0;
        deleteFile = false;
        connectingText = "";
        serverName = "";
        serverAddr = "";
        farmName = "";
        ticket = "";
        compression = true;
        encryption = true;
        
        encrypted = false;
    }

    public boolean loadPITFile(String filename)
    {
        try {
            // Open the file.
            FileInputStream is = new FileInputStream(filename);

            // Load the settings.
            return loadPITStream(is);
        }
        catch (FileNotFoundException e) {
            return false;
        }
    }
    
    public boolean savePITFile(String filename) {
        try {
          // Open the file.
          DataOutputStream os = new DataOutputStream(new FileOutputStream(filename));

          // Write the settings.
          try {
            os.writeBytes("computer:s:" + this.computer + "\n");
            os.writeBytes("application:s:" + this.application + "\n");
            
            os.writeBytes("enablesso:i:" + (this.enableSSO ? 1 : 0) + "\n");
            os.writeBytes("enablekerberos:i:" + (this.enableKerberos ? 1 : 0) + "\n");
            os.writeBytes("username:s:" + this.username + "\n");
            if (this.savePassword) {
            	os.writeBytes("password:s:" + encrypt256(this.password) + "\n");
            }
            os.writeBytes("domain:s:" + this.domain + "\n");
            
            os.writeBytes("screen mode id:i:" + this.screenModeId + "\n");
            os.writeBytes("desktopwidth:i:" + this.desktopWidth + "\n");
            os.writeBytes("desktopheight:i:" + this.desktopHeight + "\n");
            os.writeBytes("session bpp:i:" + this.sessionBitsPerPixel + "\n");
            os.writeBytes("displayconnectionbar:i:" + (this.displayConnectionBar ? 1 : 0) + "\n");
            os.writeBytes("pinconnectionbar:i:" + (this.pinConnectionBar ? 1 : 0) + "\n");
            os.writeBytes("smartsizing:i:" + (this.smartSizing ? 1 : 0) + "\n");
            
            os.writeBytes("audiomode:i:" + this.audioMode + "\n");
            os.writeBytes("keyboardhook:i:" + this.keyboardHook + "\n");
            os.writeBytes("redirectdrives:i:" + (this.redirectDrives ? 1 : 0) + "\n");
            os.writeBytes("redirectprinters:i:" + (this.redirectPrinters ? 1 : 0) + "\n");
            os.writeBytes("redirectcomports:i:" + (this.redirectComPorts ? 1 : 0) + "\n");
            os.writeBytes("redirectsmartcards:i:" + (this.redirectSmartCards ? 1 : 0) + "\n");
            os.writeBytes("redirecthandhelds:i:" + (this.redirectHandhelds ? 1 : 0) + "\n");
            os.writeBytes("redirectuniversalprinters:i:" + (this.redirectUniversalPrinters ? 1 : 0) + "\n");
            os.writeBytes("redirectclipboard:i:" + this.redirectClipboard);
            
            os.writeBytes("alternate shell:s:" + encrypt1024(this.programFullPath) + "\n");
            os.writeBytes("shell working directory:s:" + this.programWorkingDir + "\n");
            os.writeBytes("session sharing:i:" + (this.sessionSharing ? 1 : 0) + "\n");
            os.writeBytes("seamless:i:" + (this.seamless ? 1 : 0) + "\n");
            
            os.writeBytes("disable wallpaper:i:" + (this.enableWallpaper ? 0 : 1) + "\n");
            os.writeBytes("disable full window drag:i:" + (this.enableFullWindowDrag ? 0 : 1) + "\n");
            os.writeBytes("disable menu anims:i:" + (this.enableAnimation ? 0 : 1) + "\n");
            os.writeBytes("disable themes:i:" + (this.enableThemes ? 0 : 1) + "\n");
            os.writeBytes("disable font smoothing:i:" + (this.enableFontSmoothing ? 0 : 1) + "\n");
            os.writeBytes("disable desktop composition:i:" + (this.enableDesktopComposition ? 0 : 1) + "\n");
            os.writeBytes("bitmapcachepersistenable:i:" + (this.enableBitmapCaching ? 1 : 0) + "\n");
            os.writeBytes("autoreconnection enabled:i:" + (this.enableAutoReconnection ? 1 : 0) + "\n");
            os.writeBytes("bitmapacceleration:i:" +  + (this.bitmapAcceleration ? 1 : 0) + "\n");
            
            os.writeBytes("sslgateway:s:" + this.sslGateway + "\n");
            os.writeBytes("useproxy:i:" + (this.useProxy ? 1 : 0) + "\n");
            os.writeBytes("proxyserver:s:" + this.proxyServer + "\n");
            os.writeBytes("proxybypass:s:" + this.proxyBypass + "\n");
            
            os.writeBytes("apptype:i:" + this.appType + "\n");
            os.writeBytes("deletefile:i:" + (this.deleteFile ? 1 : 0) + "\n");
            os.writeBytes("connectingtext:s:" + this.connectingText + "\n");
            os.writeBytes("servername:s:" + this.serverName + "\n");
            os.writeBytes("serveraddr:s:" + this.serverAddr + "\n");
            os.writeBytes("farmname:s:" + this.farmName + "\n");
            os.writeBytes("ticket:s:" + this.ticket + "\n");
            os.writeBytes("compression:i:" + (this.compression ? 1 : 0) + "\n");
            os.writeBytes("encryption:i:" + (this.encryption ? 1 : 0) + "\n");
          }
          catch (IOException e) {
          }

          // Close the file.
          try {
            os.close();
          }
          catch (IOException e) {
          }
        }
        catch (FileNotFoundException e) {
          return false;
        }

        return true;
      }

    public boolean loadPITStream(InputStream is)
    {
    	// Decrypt the stream.
    	InputStream decryptedStream = PNCipher.decryptStream(is);
    	if (decryptedStream != null) {
    		is = decryptedStream;
    		encrypted = true;
    	}
    	
        // Read the settings.
        for (;;) {
          String line = new String();

          // Read the next line.
          try {
            for (;;) {
              char nextChar = (char)is.read();
              if (nextChar == (char)-1) {
                line = null;
                break;
              }
              if (nextChar == '\n') break;
              if (nextChar == '\r') continue;
              line += nextChar;
            }
          }
          catch (IOException e) {
            line = null;
          }
          if (line == null) break;
          
          // Parse the line.
          StringTokenizer tokenizer = new StringTokenizer(line, ":");
          String name, type, value;
          try { name = tokenizer.nextToken(); } catch (Exception e) { name = ""; }
          try { type = tokenizer.nextToken(); } catch (Exception e) { type = ""; }
          try { value = tokenizer.nextToken(); } catch (Exception e) { value = ""; }
          for (;;) {
            String extra;
            try { extra = tokenizer.nextToken(); } catch (Exception e) { extra = ""; }
            if (extra.length() > 0) {
              value = value + ":" + extra;
            } else break;
          }
          
          //JOptionPane.showMessageDialog(null, name + "=" + value);

          if (name.equals("computer")) {
        	  this.computer = value;
          }
          else if (name.equals("application")) {
        	  this.application = value;
          }
          else if (name.equals("enablesso")) {
        	  this.enableSSO = toBoolean(value);
          }
          else if (name.equals("enablekerberos")) {
        	  this.enableKerberos = toBoolean(value);
          }
          else if (name.equals("username")) {
        	  this.username = value;
          }
          else if (name.equals("password")) {
        	  this.password = decrypt256(value);
        	  if (this.password.length() > 0)
        	  {
        		  this.savePassword = true;
        	  }
          }
          else if (name.equals("domain")) {
        	  this.domain = value;
          }
          else if (name.equals("screen mode id")) {
        	  this.screenModeId = toInteger(value);
          }
          else if (name.equals("desktopwidth")) {
        	  this.desktopWidth = toInteger(value);
          }
          else if (name.equals("desktopheight")) {
        	  this.desktopHeight = toInteger(value);
          }
          else if (name.equals("session bpp")) {
        	  this.sessionBitsPerPixel = toInteger(value);
          }
          else if (name.equals("displayconnectionbar")) {
        	  this.displayConnectionBar = toBoolean(value);
          }
          else if (name.equals("pinconnectionbar")) {
        	  this.pinConnectionBar = toBoolean(value);
          }
          else if (name.equals("smartsizing")) {
        	  this.smartSizing = toBoolean(value);
          }
          else if (name.equals("audiomode")) {
        	  this.audioMode = toInteger(value);
          }
          else if (name.equals("keyboardhook")) {
        	  this.keyboardHook = toInteger(value);
          }
          else if (name.equals("redirectdrives")) {
        	  this.redirectDrives = toBoolean(value);
          }
          else if (name.equals("redirectprinters")) {
        	  this.redirectPrinters = toBoolean(value);
          }
          else if (name.equals("redirectcomports")) {
        	  this.redirectComPorts = toBoolean(value);
          }
          else if (name.equals("redirectsmartcards")) {
        	  this.redirectSmartCards = toBoolean(value);
          }
          else if (name.equals("redirecthandhelds")) {
        	  this.redirectHandhelds = toBoolean(value);
          }
          else if (name.equals("redirectUniversalPrinters")) {
        	  this.redirectUniversalPrinters = toBoolean(value);
          }
          else if (name.equals("alternate shell")) {
        	  this.programFullPath = decrypt1024(value);
          }
          else if (name.equals("shell working directory")) {
        	  this.programWorkingDir = value;
          }
          else if (name.equals("session sharing")) {
        	  this.sessionSharing = toBoolean(value);
          }
          else if (name.equals("seamless")) {
        	  this.seamless = toBoolean(value);
          }
          else if (name.equals("disable wallpaper")) {
        	  this.enableWallpaper = !toBoolean(value);
          }
          else if (name.equals("disable full window drag")) {
        	  this.enableFullWindowDrag = !toBoolean(value);
          }
          else if (name.equals("disable menu anims")) {
        	  this.enableAnimation = !toBoolean(value);
          }
          else if (name.equals("disable themes")) {
        	  this.enableThemes = !toBoolean(value);
          }
          else if (name.equals("disable font smoothing")) {
              this.enableFontSmoothing = !toBoolean(value);
          }
          else if (name.equals("disable desktop composition")) {
              this.enableDesktopComposition = !toBoolean(value);
          }
          else if (name.equals("bitmapcachepersistenable")) {
        	  this.enableBitmapCaching = toBoolean(value);
          }
          else if (name.equals("autoreconnection enabled")) {
        	  this.enableAutoReconnection = toBoolean(value);
          }
          else if (name.equals("sslgateway")) {
        	  this.sslGateway = value;
          }
          else if (name.equals("useproxy")) {
        	  this.useProxy = toBoolean(value);
          }
          else if (name.equals("proxyserver")) {
        	  this.proxyServer = value;
          }
          else if (name.equals("proxybypass")) {
        	  this.proxyBypass = value;
          }
          else if (name.equals("apptype")) {
        	  this.appType = toInteger(value);
          }
          else if (name.equals("deletefile")) {
        	  this.deleteFile = toBoolean(value);
          }
          else if (name.equals("connectingtext")) {
        	  this.connectingText = value;
          }
          else if (name.equals("servername")) {
        	  this.serverName = value;
          }
          else if (name.equals("serveraddr")) {
        	  this.serverAddr = value;
          }
          else if (name.equals("farmname")) {
        	  this.farmName = value;
          }
          else if (name.equals("ticket")) {
        	  this.ticket = value;
          }
          else if (name.equals("compression")) {
        	  this.compression = true;
        	  if (!toBoolean(value))
        	  {
        		  logger.warn("Incorrect bitmap compression parameter. Compression set to TRUE");
        	  }
          }
          else if (name.equals("encryption")) {
        	  this.encryption = toBoolean(value);
          } else if (name.equals("redirectclipboard")) {
                  this.redirectClipboard = toBoolean(value);
          } else if (name.equals("bitmapacceleration")) {
        	  this.bitmapAcceleration = toBoolean(value);
          } else {
        	  logger.info("Unknown variable in PIT file: " + name);
          }
        }

        // Close the file.
        try {
          is.close();
        }
        catch (IOException e) {
        }
        
        return true;
    }
    
    private int toInteger(String value)
    {
    	try
    	{
    		return Integer.parseInt(value);
    	}
    	catch (Exception e) { return 0; }
    }
    
    private boolean toBoolean(String value)
    {
    	try
    	{
    		return Integer.parseInt(value) != 0 ? true : false;
    	}
    	catch (Exception e) { return false; }
    }
    
    private String decrypt256(String value)
    {
    	try
    	{
	        byte[] bin = PNCipher.hexToBin(value);
	        PNCipher.decrypt(bin);
	        return PNCipher.decode256(bin);
    	}
    	catch (Exception e) { return ""; }
    }
    
    private String decrypt1024(String value)
    {
    	try
    	{
	    	byte[] bin = PNCipher.hexToBin(value);
	    	PNCipher.decrypt(bin);
	    	return PNCipher.decode1024(bin);
    	}
    	catch (Exception e) { return ""; }
    }
    
    private String encrypt256(String value)
    {
    	try
    	{
	    	byte[] bin = PNCipher.encode256(value);
	    	PNCipher.encrypt(bin);
	    	return PNCipher.binToHex(bin);
    	}
    	catch (Exception e) { return ""; }
    }
    
    private String encrypt1024(String value)
    {
    	try
    	{
	    	byte[] bin = PNCipher.encode1024(value);
	    	PNCipher.encrypt(bin);
	    	return PNCipher.binToHex(bin);
    	}
    	catch (Exception e) { return ""; }
    }
}
