package com.provisionnetworks.common;

import java.io.InputStream;
import java.net.InetSocketAddress;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.security.cert.Certificate;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import org.apache.log4j.Logger;

public class SSLUtil {
	
	static Logger logger = Logger.getLogger(SSLUtil.class);
	
	public static Certificate[] getCertificates(String host, int port) {
		Certificate[] certs = getCertificates(host, port, true);
		if (certs == null) {
			certs = getCertificates(host, port, false);
		}
		return certs;
	}
	
	public static Certificate[] getCertificates(String host, int port, boolean enableTLS) {
		logger.info("Get SSL certificates from host");
    	Certificate[] certs = null;

    	//Disable/Enable TLS protocol
    	//Control TLS for connecting via WS2008
    	final String httpsProperty = "https.protocols";
    	if (enableTLS) {
    		logger.info("Enable TLS protocol");
    		System.setProperty(httpsProperty, "SSLv3,TLSv1");
    	} else {
    		logger.info("Disable TLS protocol");
    		System.setProperty(httpsProperty, "SSLv3");
    	}
    	
    	try {
    		URL url = new URL("https://" + host + ":" + port);
    		HttpsURLConnection con = (HttpsURLConnection)url.openConnection();
    		con.connect();
    		certs = con.getServerCertificates();
    		con.disconnect();
    	} catch (Exception e) {
    		logger.warn(e.getMessage());
    	}
    	if (certs != null && certs.length > 0) {
    		logger.debug(certs[0].toString());
    	}
    	return certs;
    }
	
    public static boolean disableCertChecking() {
    	logger.warn("Enabling trusting to any SSL certificate");
    	//Creating new trust manager which trust all certificates
	    TrustManager[] trustAllCerts = new TrustManager[]{
	        new X509TrustManager() {
	            public java.security.cert.X509Certificate[] getAcceptedIssuers() {
	                return null;
	            }
	            public void checkClientTrusted(
	                    java.security.cert.X509Certificate[] certs, String authType) {
	            }
	            public void checkServerTrusted(
	                    java.security.cert.X509Certificate[] certs, String authType) {
	            }
	        }
	    };

	    //Creating new host names verifier
	    HostnameVerifier verifier = new HostnameVerifier() {
	        public boolean verify(String string, SSLSession sSLSession) {
	            return true;
	        }
	    };
	
	    //Set new trust manager and host verifier
	    try {
	        SSLContext sc = SSLContext.getInstance("TLS");
	        sc.init(null, trustAllCerts, new java.security.SecureRandom());
	        HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
	        HttpsURLConnection.setDefaultHostnameVerifier(verifier);
	    } catch (Exception e) {
	    	logger.warn(e.getMessage());
	    	return false;
	    }
	    return true;
    }
    
    public static boolean isServerCertificateTrusted(String host, int port) {
    	logger.info("SSL certificate checking");
    	try {
	    	SSLSocket sock = (SSLSocket)SSLSocketFactory.getDefault().createSocket();
	    	sock.connect(new InetSocketAddress(host,port));
	    	sock.setSoTimeout(1000);
	    	InputStream is = sock.getInputStream();
	    	is.read();
	    	is.close();
	    	sock.close();
    	} catch (SocketTimeoutException sockExcept) {
    		logger.info("Socket closed by timeout: SSL certificate is trusted by Java");
	    	return true;
	    } catch (Exception e) {
    		logger.warn("SSL certificate not trusted by Java."); 
    		logger.warn("Socket was closed with error: " + e.getMessage());
    		return false;
    	}
	    logger.info("SSL certificate is trusted by Java");
    	return true;
    }
}
