
package com.provisionnetworks.pntsc;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import org.apache.log4j.Logger;

import quest.android.appportal.connection.PNException;

import android.app.Activity;

import com.provisionnetworks.common.OnRdpControlListener;
import com.provisionnetworks.common.RDPSettings;
import com.provisionnetworks.rdp5.Rdesktop;

public class PNTSC
{
    public void destroy() 
    {
        activity = null;
        rdpSettings = null;
        if ( rdesktop != null )
        {
            rdesktop.destroy();
            rdesktop = null;
        }
    }
    
    Logger logger            = Logger.getLogger( PNTSC.class );

    private Activity   activity       = null;
    
    public RDPSettings rdpSettings    = null;
    private Rdesktop rdesktop = null;
    
    public PNTSC( Activity activity )
    {
        this.activity = activity;
    }

    public void main( String args ) throws PNException
    {
        logger.debug( "PNTSC started. Arguments: " + args );
        
        try
        {
            System.setProperty( "java.net.useSystemProxies", "true" );

            InputStream is = new ByteArrayInputStream( args.getBytes() );

            rdpSettings = new RDPSettings();
            rdpSettings.loadPITStream( is );

            if ( rdpSettings == null )
            {
                throw new PNException( "Error getting connection parameters", 0 );
            }
            
            startRDPClient();
        }
        catch( Exception e )
        {
            logger.error( "Error occurred: ", e );
            throw new PNException( e.getMessage(), 0 );
        }

        logger.info( "Exiting" );
    }

    private void startRDPClient()
    {
        try
        {
            rdesktop = new Rdesktop( activity, rdpSettings );
            rdesktop.setOnRdpControlListener( rdpControlListener );
            rdesktop.main( );
        }
        catch( Exception e )
        {
            logger.error( "Error in RDP connection: " + e.getMessage(), e );
            StackTraceElement[] stackTrace = e.getStackTrace();
            for ( int i = 0; i < stackTrace.length; ++i )
            {
                logger.debug( stackTrace[i].getFileName() + " - " + stackTrace[i].getMethodName() + ":" + stackTrace[i].getLineNumber() );
            }
        }
    }
    
    public void stop()
    {
        if ( rdesktop != null )
        {
            rdesktop.stop();
        }
    }

    private OnRdpControlListener rdpControlListener = null;

    public void setOnRdpControlListener( OnRdpControlListener listener )
    {
        this.rdpControlListener = listener;
    }
}
