package com.provisionnetworks.rdp5;

import java.io.IOException;

import net.propero.rdp.Options;
import net.propero.rdp.RdesktopException;
import net.propero.rdp.RdpPacket;
import net.propero.rdp.RdpPacket_Localised;
import net.propero.rdp.crypto.CryptoException;
import net.propero.rdp.rdp5.VChannel;

import org.apache.log4j.Logger;


public class ChannelPNTSC extends VChannel
{
    protected Logger logger = Logger.getLogger(ChannelPNTSC.class.getName());

    private VChannel pntssChannel = null;
    private PNOptions pnOptions = null;
    private Options options = null;
    
    public ChannelPNTSC(VChannel pntssChannel, Options options, PNOptions pnOptions)
    {
        super( options );
        this.pntssChannel = pntssChannel;
        this.options = options;
        this.pnOptions = pnOptions;
    }
    
    public int flags()
    {
        return 0;
            //VChannels.CHANNEL_OPTION_INITIALIZED | 
            //VChannels.CHANNEL_OPTION_ENCRYPT_RDP |
            //VChannels.CHANNEL_OPTION_COMPRESS_RDP | 
            //VChannels.CHANNEL_OPTION_SHOW_PROTOCOL;
    }

    public String name()
    {
        return "PNTSC";
    }

    public void process(RdpPacket data) throws RdesktopException, IOException, CryptoException
    {
        logger.debug("PNTSC channel packet");
        try 
        {
            byte packetData[] = new byte[data.getEnd() - data.getPosition()];
            data.copyToByteArray(packetData, 0, data.getPosition(), packetData.length);
            String msg = new String(packetData, "UTF-16LE");
            logger.debug("Data=\"" + msg + "\"");
            
            if (msg.equals("ID"))
            {
                if (options.command != null && options.command.length() > 0)
                {
                    logger.info("Sended LOGIN message to PNTSS channel");
                    String login = "LI" + pnOptions.programFullPath + "\n" + pnOptions.programWorkingDir;
                    RdpPacket_Localised packet = new RdpPacket_Localised(login.length() * 2 + 2);
                    packet.outUnicodeString(login, login.length() * 2);
                    packet.markEnd();
                    try { pntssChannel.send_packet(packet); } catch (Exception e) { logger.error(e.getMessage(), e); }
                }
            }
        } 
        catch (Exception e)
        {
            logger.error( "Exception: ", e );
        }
    }
}
