/*
 * Rdesktop.java Component: ProperJavaRDP Revision: $Revision: 1.7 $ Author:
 * $Author: telliott $ Date: $Date: 2005/09/27 14:15:39 $ Copyright (c) 2005
 * Propero Limited Purpose: Main class, launches session This program is free
 * software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version. This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License for more details. You
 * should have received a copy of the GNU General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA (See gpl.txt for details of the
 * GNU General Public License.)
 */

package com.provisionnetworks.rdp5;

import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.StringTokenizer;

import net.propero.rdp.ConnectionException;
import net.propero.rdp.Constants;
import net.propero.rdp.Options;
import net.propero.rdp.OrderException;
import net.propero.rdp.RdesktopException;
import net.propero.rdp.RdesktopFrame;
import net.propero.rdp.RdesktopFrame_Localised;
import net.propero.rdp.Rdp;
import net.propero.rdp.Version;
import net.propero.rdp.rdp5.Rdp5;
import net.propero.rdp.rdp5.VChannels;

import org.apache.log4j.Logger;

import android.app.Activity;
import android.graphics.Rect;

import com.provisionnetworks.common.OnRdpControlListener;
import com.provisionnetworks.common.RDPSettings;
import com.provisionnetworks.rdp5.pnrdpp.PnrdppChannel;
import com.provisionnetworks.rdp5.rdpdr.RdpdrChannel;
import com.provisionnetworks.rdp5.rdpsnd.RdpsndChannel;

public class Rdesktop
{
    
    public void destroy()
    {
        if ( pnOptions != null )
        {
            pnOptions.destroy();
            pnOptions = null;
        }
        if ( options != null )
        {
            options.destroy();
            options = null;
        }
        activity = null;
        
        if ( RdpLayer != null )
        {
            RdpLayer.destroy();
            RdpLayer = null;
        }
    }
    
    /**
     * Translate a disconnect code into a textual description of the reason for
     * the disconnect
     * 
     * @param reason
     *            Integer disconnect code received from server
     * @return Text description of the reason for disconnection
     */
    static String textDisconnectReason( int reason )
    {
        String text;

        switch ( reason )
        {
            case exDiscReasonNoInfo:
                text = "No information available";
                break;

            case exDiscReasonAPIInitiatedDisconnect:
                text = "Server initiated disconnect";
                break;

            case exDiscReasonAPIInitiatedLogoff:
                text = "Server initiated logoff";
                break;

            case exDiscReasonServerIdleTimeout:
                text = "Server idle timeout reached";
                break;

            case exDiscReasonServerLogonTimeout:
                text = "Server logon timeout reached";
                break;

            case exDiscReasonReplacedByOtherConnection:
                text = "Another user connected to the session";
                break;

            case exDiscReasonOutOfMemory:
                text = "The server is out of memory";
                break;

            case exDiscReasonServerDeniedConnection:
                text = "The server denied the connection";
                break;

            case exDiscReasonServerDeniedConnectionFips:
                text = "The server denied the connection for security reason";
                break;

            case exDiscReasonLicenseInternal:
                text = "Internal licensing error";
                break;

            case exDiscReasonLicenseNoLicenseServer:
                text = "No license server available";
                break;

            case exDiscReasonLicenseNoLicense:
                text = "No valid license available";
                break;

            case exDiscReasonLicenseErrClientMsg:
                text = "Invalid licensing message";
                break;

            case exDiscReasonLicenseHwidDoesntMatchLicense:
                text = "Hardware id doesn't match software license";
                break;

            case exDiscReasonLicenseErrClientLicense:
                text = "Client license error";
                break;

            case exDiscReasonLicenseCantFinishProtocol:
                text = "Network error during licensing protocol";
                break;

            case exDiscReasonLicenseClientEndedProtocol:
                text = "Licensing protocol was not completed";
                break;

            case exDiscReasonLicenseErrClientEncryption:
                text = "Incorrect client license enryption";
                break;

            case exDiscReasonLicenseCantUpgradeLicense:
                text = "Can't upgrade license";
                break;

            case exDiscReasonLicenseNoRemoteConnections:
                text = "The server is not licensed to accept remote connections";
                break;

            default:
                if ( reason > 0x1000 && reason < 0x7fff )
                {
                    text = "Internal protocol error";
                }
                else
                {
                    text = null;
                }
        }
        return text;
    }

    /* RDP5 disconnect PDU */
    public static final int exDiscReasonNoInfo                        = 0x0000;
    public static final int exDiscReasonAPIInitiatedDisconnect        = 0x0001;
    public static final int exDiscReasonAPIInitiatedLogoff            = 0x0002;
    public static final int exDiscReasonServerIdleTimeout             = 0x0003;
    public static final int exDiscReasonServerLogonTimeout            = 0x0004;
    public static final int exDiscReasonReplacedByOtherConnection     = 0x0005;
    public static final int exDiscReasonOutOfMemory                   = 0x0006;
    public static final int exDiscReasonServerDeniedConnection        = 0x0007;
    public static final int exDiscReasonServerDeniedConnectionFips    = 0x0008;
    public static final int exDiscReasonLicenseInternal               = 0x0100;
    public static final int exDiscReasonLicenseNoLicenseServer        = 0x0101;
    public static final int exDiscReasonLicenseNoLicense              = 0x0102;
    public static final int exDiscReasonLicenseErrClientMsg           = 0x0103;
    public static final int exDiscReasonLicenseHwidDoesntMatchLicense = 0x0104;
    public static final int exDiscReasonLicenseErrClientLicense       = 0x0105;
    public static final int exDiscReasonLicenseCantFinishProtocol     = 0x0106;
    public static final int exDiscReasonLicenseClientEndedProtocol    = 0x0107;
    public static final int exDiscReasonLicenseErrClientEncryption    = 0x0108;
    public static final int exDiscReasonLicenseCantUpgradeLicense     = 0x0109;
    public static final int exDiscReasonLicenseNoRemoteConnections    = 0x010a;

    Logger           logger                                    = Logger.getLogger( "com.provisionnetworks.rdp5" );

    boolean                 keep_running;
    public boolean          loggedon;
    public boolean          readytosend;

    Rdp5 RdpLayer = null;

    // Save the RDP settings.
    PNOptions               pnOptions                                 = new PNOptions();
    Options                 options                                   = new Options();

    Activity activity = null;

    public Rdesktop( Activity activity, RDPSettings settings )
    {
        this.activity = activity;
        this.setRDPSettings( settings );
    }

    /**
     * @param args
     * @throws OrderException
     * @throws RdesktopException
     */
    public void main( ) throws OrderException,
            net.propero.rdp.RdesktopException
    {

        // Ensure that variables are properly initialised
        keep_running = true;

        int logonflags = Rdp.RDP_LOGON_NORMAL | Rdp.RDP_LOGON_AUTO;

        String server = options.hostname;

        VChannels channels = new VChannels( options );

        // ClipChannel clipChannel = null;
        if ( pnOptions.rdpSettings.redirectClipboard )
        {
            // clipChannel = new ClipChannel();
        }
        else
        {
            options.map_clipboard = false;
        }

        RdpsndChannel rdpsndChannel = null;
        RdpdrChannel rdpdrChannel = null;
        PnrdppChannel pnrdppChannel = null;

        if ( pnOptions.rdpSettings.redirectDrives
                || pnOptions.rdpSettings.redirectPrinters )
        {
            rdpdrChannel = new RdpdrChannel( options,
                    pnOptions.rdpSettings.redirectDrives,
                    pnOptions.rdpSettings.redirectPrinters, false, false );
        }

        if ( pnOptions.rdpSettings.audioMode == 0 )
        {
            rdpsndChannel = new RdpsndChannel( options );
        } 
        else if ( pnOptions.rdpSettings.audioMode == 1 )
        {
            logonflags |= Rdp.INFO_REMOTECONSOLEAUDIO;
        }
        else if ( pnOptions.rdpSettings.audioMode == 2 )
        {
            logonflags |= Rdp.INFO_NOAUDIOPLAYBACK;
        }

        if ( pnOptions.rdpSettings.bitmapAcceleration )
        {
            pnrdppChannel = new PnrdppChannel( options );
        }
        
        ChannelPNKERB pnkerb = null;
        if ( (pnOptions.rdpSettings.ticket != null)
                && (pnOptions.rdpSettings.ticket.length() > 0) )
        {
            pnkerb = new ChannelPNKERB( options, pnOptions );
        }
        ChannelPNTSS pntss = new ChannelPNTSS( options );
        ChannelPNTSC pntsc = new ChannelPNTSC( pntss, options, pnOptions );
        
        // Initialise all RDP5 channels
        if ( options.use_rdp5 )
        {
            /**********************************************************/
            if ( true )
            {
                if ( (rdpdrChannel != null)
                        && !channels.register( rdpdrChannel ) )
                    logger.error( "Can't register channel RDPDR" );

                if ( rdpsndChannel != null
                        && !channels.register( rdpsndChannel ) )
                    logger.error( "Can't register channel RDPSND" );
            }
            /**********************************************************/
            // if ((clipChannel != null) && !channels.register(clipChannel))
            // logger.error("Can't register channel Clipboard");
            
            if ( (pnkerb != null) && !channels.register( pnkerb ) )
                logger.error( "Can't register channel PNKERB" );
            if ( !channels.register( pntsc ) )
                logger.error( "Can't register channel PNTSC" );
            if ( !channels.register( pntss ) )
                logger.error( "Can't register channel PNTSS" );
            
            if ( pnrdppChannel != null && !channels.register( pnrdppChannel ) )
                logger.error( "Can't register channel PNRDP" );
        }

        /*****************************************************************************/
        // Now do the startup...

        logger.info( "properJavaRDP version " + Version.version );
        /*
         * if (args.length == 0) usage();
         */
        String java = System.getProperty( "java.specification.version" );
        logger.info( "Java version is " + java );

        String os = System.getProperty( "os.name" );
        String osvers = System.getProperty( "os.version" );

        if ( os.equals( "Windows 2000" ) || os.equals( "Windows XP" ) )
            options.built_in_licence = true;

        logger.info( "Operating System is " + os + " version " + osvers );

        if ( os.startsWith( "Linux" ) )
            Constants.OS = Constants.LINUX;
        else if ( os.startsWith( "Windows" ) )
            Constants.OS = Constants.WINDOWS;
        else if ( os.startsWith( "Mac" ) )
            Constants.OS = Constants.MAC;

        if ( Constants.OS == Constants.MAC )
            options.caps_sends_up_and_down = false;

        RdesktopFrame window = new RdesktopFrame_Localised( activity, options );
        
        boolean[] deactivated = new boolean[1];
        int[] ext_disc_reason = new int[1];

        String msg = null;

        logger.debug( "keep_running = " + keep_running );
        while ( keep_running )
        {
            logger.debug( "Initialising RDP layer..." );
            try
            {
                RdpLayer = new Rdp5( activity, channels, options, pnOptions );
            }
            catch( Throwable e )
            {
                logger.fatal( "Error: " + e.getMessage(), e );
                throw new RdesktopException( "Not enough memory" );
            }
            if ( RdpLayer != null )
            {
                options.common.rdp = RdpLayer;
                logger.debug( "Registering drawing surface..." );
                RdpLayer.registerDrawingSurface( window );
                logger.debug( "Registering comms layer..." );
                window.registerCommLayer( RdpLayer );

                loggedon = false;
                readytosend = false;
                logger.info( "Connecting to " + server + ":" + options.port
                        + " ..." );

                if ( server.equalsIgnoreCase( "localhost" ) )
                    server = "127.0.0.1";

                // Attempt to connect to server on port Options.port
                try
                {
                    RdpLayer.connect( options.username,
                            InetAddress.getByName( server ), logonflags,
                            options.domain, options.password, options.command,
                            options.directory );

                    // srv.doneConnect(pntss);

                    // Remove to get rid of sendEvent tool
                    /*
                     * if (showTools) { toolFrame = new SendEvent(RdpLayer);
                     * toolFrame.setVisible(true); }
                     */
                    // End

                    if ( keep_running )
                    {

                        /*
                         * By setting encryption to False here, we have an
                         * encrypted login packet but unencrypted transfer of
                         * other packets
                         */
                        if ( !options.packet_encryption )
                            options.encryption = false;

                        logger.info( "Connection successful" );
                        
                        // now show window after licence negotiation
                        window.setOnRdpControlListener( rdpControlListener );
                        RdpLayer.mainLoop( deactivated, ext_disc_reason );

                        if ( deactivated[0] )
                        {
                            /* clean disconnect */
                            exit( 0, RdpLayer, window, true );
                            // return 0;
                        }
                        else
                        {
                            if ( ext_disc_reason[0] == exDiscReasonAPIInitiatedDisconnect
                                    || ext_disc_reason[0] == exDiscReasonAPIInitiatedLogoff )
                            {
                                /*
                                 * not so clean disconnect, but nothing to worry
                                 * about
                                 */
                                exit( 0, RdpLayer, window, true );
                                // return 0;
                            }

                            if ( ext_disc_reason[0] >= 2 )
                            {
                                String reason = textDisconnectReason( ext_disc_reason[0] );
                                if ( reason != null )
                                {
                                    msg = "Connection terminated: " + reason;
                                }
                                logger.warn( "Connection terminated: " + reason );
                                exit( 0, RdpLayer, window, true );
                                
                            }

                        }

                        keep_running = false; // exited main loop
                        /***************************************
                         * if (!readytosend) { // maybe the licence server was
                         * having a comms // problem, retry? String msg1 =
                         * "The terminal server disconnected before licence negotiation completed."
                         * ; String msg2 =
                         * "Possible cause: terminal server could not issue a licence."
                         * ; String[] msg = { msg1, msg2 }; logger.warn(msg1);
                         * logger.warn(msg2); window.showErrorDialog(msg); }
                         ***************************************/
                    } // closing bracket to if(running)

                }
                catch( ConnectionException e )
                {
                    String errorMsg = e.getMessage();
                    logger.error( errorMsg, e );
                    msg = "Could not establish a connection to the server";
                    exit( 0, RdpLayer, window, true );
                }
                catch( UnknownHostException e )
                {
                    msg = "Unknown Host Exception: " + e.getMessage();
                    error( e, RdpLayer, window, true );
                }
                catch( SocketException s )
                {
                    msg = s.getMessage();

                    if ( RdpLayer.isConnected() )
                    {
                        logger.fatal(
                                s.getClass().getName() + " " + s.getMessage(),
                                s );
                        error( s, RdpLayer, window, true );
                        exit( 0, RdpLayer, window, true );
                    }
                }
                catch( RdesktopException e )
                {
                    String msg1 = e.getClass().getName();
                    String msg2 = e.getMessage();
                    logger.fatal( msg1 + ": " + msg2, e );

                    e.printStackTrace( System.err );

                    if ( !readytosend )
                    {
                        // maybe the licence server was having a comms
                        // problem, retry?
                        msg = "The terminal server reset connection before licence negotiation completed.\nPossible cause: "
                                + e.getMessage();
                        boolean retry = false;
                        if ( !retry )
                        {
                            logger.info( "Selected not to retry." );
                            exit( 0, RdpLayer, window, true );
                        }
                        else
                        {
                            if ( RdpLayer != null && RdpLayer.isConnected() )
                            {
                                logger.info( "Disconnecting ..." );
                                RdpLayer.disconnect();
                                logger.info( "Disconnected" );
                            }
                            logger.info( "Retrying connection..." );
                            keep_running = true; // retry
                            continue;
                        }
                    }
                    else
                    {
                        //msg = e.getMessage();
                        exit( 0, RdpLayer, window, true );
                    }
                }
                catch( Exception e )
                {
                    logger.warn( e.getClass().getName() + " " + e.getMessage() );
                    StackTraceElement[] list = e.getStackTrace();
                    for( int i = 0; i < list.length; ++i )
                        logger.info( i + ": " + list[i].toString() );
                    error( e, RdpLayer, window, true );
                }
            }
            else
            { // closing bracket to if(!rdp==null)
                logger.fatal( "The communications layer could not be initiated!" );
                msg = "The communications layer could not be initiated!";
            }
        }
        
        exit( 0, RdpLayer, window, true );

        if ( rdpControlListener != null )
        {
            rdpControlListener.OnRdpConnectionStopped( window, msg );
        }
    }

    public void stop()
    {
        keep_running = false;

        // if ( srv != null )
        // {
        // srv.disconnect();
        // }

        if ( RdpLayer != null && RdpLayer.isConnected() )
        {
            logger.info( "Disconnecting ..." );
            RdpLayer.disconnect();
            logger.info( "Disconnected" );
        }
    }
    
    /**
     * Disconnects from the server connected to through rdp and destroys the
     * RdesktopFrame window.
     * <p>
     * Exits the application iff sysexit == true, providing return value n to
     * the operating system.
     * 
     * @param n
     * @param rdp
     * @param window
     * @param sysexit
     */
    public void exit( int n, Rdp rdp, RdesktopFrame window, boolean sysexit )
    {
        keep_running = false;

        // if ( srv != null )
        // {
        // srv.disconnect();
        // }
        
        if ( window != null )
        {
            window.destroy();
        }
        
        if ( rdp != null && rdp.isConnected() )
        {
            logger.info( "Disconnecting ..." );
            rdp.disconnect();
            logger.info( "Disconnected" );
        }
    }

     /**
     * Displays details of the Exception e in an error dialog via the
     * RdesktopFrame window and reports this through the logger, then prints a
     * stack trace.
     * <p>
     * The application then exits iff sysexit == true
     * 
     * @param e
     * @param RdpLayer
     * @param window
     * @param sysexit
     */
    public void error( Exception e, Rdp RdpLayer, RdesktopFrame window,
            boolean sysexit )
    {
        try
        {
            String msg1 = e.getClass().getName();
            String msg2 = e.getMessage();

            logger.fatal( msg1 + ": " + msg2 );
        }
        catch( Exception ex )
        {
            logger.warn( "Exception in Rdesktop.error: "
                    + ex.getClass().getName() + ": " + ex.getMessage() );
        }

        exit( 0, RdpLayer, window, sysexit );
    }

    public void setRDPSettings( RDPSettings rdpSettings )
    {
        int pos;

        pnOptions.rdpSettings = rdpSettings;

        // Determine the server to which a connection is to be established.
        String server = rdpSettings.computer;
        if ( (rdpSettings.serverName != null)
                && (rdpSettings.serverName.length() > 0) )
        {
            // Retrieve the first server name in the list.
            StringTokenizer tokenizer = new StringTokenizer(
                    rdpSettings.serverName, "," );
            rdpSettings.serverName = tokenizer.nextToken();
            server = rdpSettings.serverName;
        }
        if ( (rdpSettings.serverAddr != null)
                && (rdpSettings.serverAddr.length() > 0) )
        {
            // Retrieve the first server address in the list.
            StringTokenizer tokenizer = new StringTokenizer(
                    rdpSettings.serverAddr, "," );
            rdpSettings.serverAddr = tokenizer.nextToken();
            server = rdpSettings.serverAddr;
        }

        // If an SSL gateway is being used.
        if ( (rdpSettings.sslGateway != null)
                && (rdpSettings.sslGateway.length() > 0) )
        {
            pnOptions.sslGateway = rdpSettings.sslGateway;

            // Parse the SSL gateway of the form "host:port".
            pos = rdpSettings.sslGateway.indexOf( ':' );
            if ( pos > 0 )
            {
                options.hostname = rdpSettings.sslGateway.substring( 0, pos );
                options.port = Integer.parseInt( rdpSettings.sslGateway
                        .substring( pos + 1 ) );
            }
            else
            {
                options.hostname = rdpSettings.sslGateway;
                options.port = 443;
            }

            // Parse the server name of the form "host:port".
            pos = server.indexOf( ':' );
            if ( pos > 0 )
            {
                pnOptions.targetHost = server.substring( 0, pos );
                pnOptions.targetPort = Integer.parseInt( server
                        .substring( pos + 1 ) );
            }
            else
            {
                pnOptions.targetHost = server;
                pnOptions.targetPort = 3389;
            }
        }
        else
        {
            // Parse the server name of the form "host:port".
            pos = server.indexOf( ':' );
            if ( pos > 0 )
            {
                options.hostname = server.substring( 0, pos );
                options.port = Integer.parseInt( server.substring( pos + 1 ) );
            }
            else
            {
                options.hostname = server;
                options.port = 3389;
            }
        }

        // Determine the screen size.
        // Dimension screenSize = new
        // Dimension(AndroidGraphics2D.getAndroidCanvas().getWidth(),
        // AndroidGraphics2D.getAndroidCanvas().getHeight());//Toolkit.getDefaultToolkit().getScreenSize();
        // Dimension screenSize = new Dimension(480,
        // 800);//Toolkit.getDefaultToolkit().getScreenSize();
        Rect screenSize = new Rect( 0, 0, 1024, 768 );

        // Set the remaining options.
        options.windowTitle = "vWorkspace Connector for Android - "
                + pnOptions.rdpSettings.serverName + " - "
                + pnOptions.rdpSettings.serverAddr;

        if ( (rdpSettings.ticket == null) || (rdpSettings.ticket.length() == 0) )
        {
            options.domain = rdpSettings.domain;
            options.username = rdpSettings.username;
            options.password = rdpSettings.password;
        }
        else
        {
            options.domain = "";
            options.username = "";
            options.password = "";
        }

        if ( rdpSettings.screenModeId == 1 )
        {
            options.fullscreen = false;
            options.width = rdpSettings.desktopWidth;
            options.height = rdpSettings.desktopHeight;
        }
        else
        {
            options.fullscreen = true;
            options.width = screenSize.right;
            options.height = screenSize.bottom;
        }
        options.set_bpp( rdpSettings.sessionBitsPerPixel != 0 ? rdpSettings.sessionBitsPerPixel
                : 16 );

        options.command = "pnstart.exe";

        options.bitmap_caching = rdpSettings.enableBitmapCaching;

        options.bitmap_compression = rdpSettings.compression;
        options.encryption = rdpSettings.encryption;

        if ( rdpSettings.enableWallpaper )
        {
            options.rdp5_performanceflags &= ~Rdp.RDP5_NO_WALLPAPER;
        }
        if ( rdpSettings.enableAnimation )
        {
            options.rdp5_performanceflags &= ~Rdp.RDP5_NO_MENUANIMATIONS;
        }
        if ( rdpSettings.enableFullWindowDrag )
        {
            options.rdp5_performanceflags &= ~Rdp.RDP5_NO_FULLWINDOWDRAG;
        }
        if ( rdpSettings.enableThemes )
        {
            options.rdp5_performanceflags &= ~Rdp.RDP5_NO_THEMING;
        }
        if ( rdpSettings.enableFontSmoothing )
        {
            options.rdp5_performanceflags |= Rdp.PERF_ENABLE_FONT_SMOOTHING;
        }
        if ( rdpSettings.enableDesktopComposition )
        {
            options.rdp5_performanceflags |= Rdp.PERF_ENABLE_DESKTOP_COMPOSITION;
        }

        if ( (rdpSettings.programFullPath == null)
                || (rdpSettings.programFullPath.length() == 0) )
        {
            pnOptions.programFullPath = "explorer.exe";
            pnOptions.programWorkingDir = "";
        }
        else
        {
            pnOptions.programFullPath = rdpSettings.programFullPath;
            pnOptions.programWorkingDir = rdpSettings.programWorkingDir;
        }
    }

    OnRdpControlListener rdpControlListener = null;

    public void setOnRdpControlListener( OnRdpControlListener listener )
    {
        rdpControlListener = listener;
    }
}
