package com.provisionnetworks.rdp5.pnrdpp;

import java.io.IOException;

import org.apache.log4j.Logger;

import net.propero.rdp.Options;
import net.propero.rdp.RdesktopException;
import net.propero.rdp.RdpPacket;
import net.propero.rdp.RdpPacket_Localised;
import net.propero.rdp.crypto.CryptoException;
import net.propero.rdp.rdp5.VChannel;

public class PnrdppChannel extends VChannel {

	protected Logger logger = Logger.getLogger(PnrdppChannel.class);
	
	public static final String PNRDP_NEG_REQ_STRING 	= "CHECK";
	public static final String PNRDP_NEG_RSP_STRING 	= "READY";
	public static final String PNRDP_NEG_CNF_STRING 	= "OK";
	public static final String PNRDP_VERSION			= "1";
	
	public static int ProtocolVersion = 1;
	
	public PnrdppChannel( Options options )
	{
	    super( options );
	}
	
	public int flags() {
		return 0;
		//VChannels.CHANNEL_OPTION_INITIALIZED | 
        //VChannels.CHANNEL_OPTION_ENCRYPT_RDP |
        //VChannels.CHANNEL_OPTION_COMPRESS_RDP | 
        //VChannels.CHANNEL_OPTION_SHOW_PROTOCOL;
	}

	public String name() {
		return "PNRDPP";
	}

	public void process(RdpPacket packet) throws RdesktopException, IOException,
			CryptoException {
		logger.info("PNRDPP Process Start");
		byte[] buf = new byte[packet.size()];
		packet.copyToByteArray(buf, 0, 0, packet.size());
		String msg = new String(buf);
		logger.info("Server message = " + msg);
		if(msg.startsWith(PNRDP_NEG_REQ_STRING))
		{
			String serverVersionString = msg.substring(PNRDP_NEG_REQ_STRING.length());
			int clientVersion = Integer.getInteger(PNRDP_VERSION).intValue();
			int serverVersion = Integer.getInteger(serverVersionString).intValue();
			ProtocolVersion = Math.min(clientVersion, serverVersion);
			if (ProtocolVersion <= 0 )
			{
				logger.error("Incorrect Server or Client version");
				return;
			}
			logger.info("Negotiation Request");
			RdpPacket_Localised outPacket = new RdpPacket_Localised(6);
			outPacket.copyFromByteArray(PNRDP_NEG_RSP_STRING.getBytes(), 0, 0, 5);
			outPacket.copyFromByteArray(PNRDP_VERSION.getBytes(), 0, 5, 1);
			logger.info("Send Negotiation Response");
			
			try{ send_packet(outPacket); } catch(Exception e) { logger.error(e.getMessage()); };			
		}
		else if(msg.startsWith(PNRDP_NEG_CNF_STRING))
		{
			logger.info("Negotiation OK");
		}
		
	}

}
