package com.provisionnetworks.rdp5.rdpdr;

import java.io.*;
import java.util.regex.*;

public class FileInfo implements FilenameFilter
{
	protected File file = null;
	protected DiskMonitor monitor = null;
	protected RandomAccessFile raf = null;
	protected Pattern queryPattern = null;
	protected File[] queryResults = null;
	protected int queryIndex = 0;
	
	public FileInfo(File file) {
		this.file = file;
	}
	
	public File getFile() {
		return this.file;
	}
	
	public void close() {
		stopMonitor();
		
		if (this.raf != null) {
			try {
				this.raf.close();
			} catch (Exception e) {}
			this.raf = null;
		}
	}
	
	public void startMonitor(IoRequestPacket irp) {
		if (this.monitor != null) {
			stopMonitor();
		}
		this.monitor = new DiskMonitor(this.file, irp);
	}
	
	public void stopMonitor() {
		if (this.monitor != null) {
			this.monitor.close();
			this.monitor = null;
		}
	}
	
	public void executeQuery(String pattern) {
		clearQueryResults();

		int index = pattern.lastIndexOf(File.separator);
		if (index >= 0) {
			pattern = pattern.substring(index + 1);
		}
		pattern = pattern.replace(".", "\\.");
		pattern = pattern.replace("*", ".*");
		pattern = pattern.replace("?", ".");
		
		try {
			this.queryPattern = Pattern.compile(pattern, Pattern.CASE_INSENSITIVE);
			this.queryResults = this.file.listFiles(this);
			if (this.queryResults.length == 0) {
				this.queryResults = null;
			}
			this.queryIndex = 0;
		} catch (Exception e) {}
	}

	public File prevQueryResult() {
		if (!moreQueryResults()) return null;
		
		return (this.queryIndex > 0) ? this.queryResults[--this.queryIndex] : null;
	}
	
	public File nextQueryResult() {
		if (!moreQueryResults()) return null;
		
		File file = this.queryResults[this.queryIndex++];
		if (this.queryIndex >= this.queryResults.length) {
			clearQueryResults();
		}
		
		return file;
	}
	
	public boolean moreQueryResults() {
		return this.queryResults != null ? true : false;
	}
	
	public void clearQueryResults() {
		this.queryPattern = null;
		this.queryResults = null;
		this.queryIndex = 0;
	}
	
	////////////////////////////////////////////////////////////////////
	//
	// Implementation of FilenameFilter interface
	//
	public boolean accept(File dir, String name) {
		Matcher matcher = queryPattern.matcher(name);
		return matcher.matches();
	}
}
