package com.provisionnetworks.rdp5.rdpdr;

import java.io.*;
import java.net.*;
//import javax.print.*;
import net.propero.rdp.*;
import net.propero.rdp.crypto.*;
import net.propero.rdp.rdp5.*;
import org.apache.log4j.*;

public class RdpdrChannel extends VChannel
{
    protected static Logger logger = Logger.getLogger(RdpdrChannel.class);

    protected static final int RDPDR_CTYP_CORE = 0x4472;
    protected static final int RDPDR_CTYP_PRN = 0x5052;
    
    protected static final int RDPDR_CORE_SERVER_ANNOUNCE = 0x496E;
    protected static final int RDPDR_CORE_CLIENTID_CONFIRM = 0x4343;
    protected static final int RDPDR_CORE_CLIENT_NAME = 0x434E;
    protected static final int RDPDR_CORE_SERVER_CAPABILITY = 0x5350;
    protected static final int RDPDR_CORE_CLIENT_CAPABILITY = 0x4350;
    protected static final int RDPDR_CORE_CLIENT_DISPLAY_NAME = 0x444E;
    protected static final int RDPDR_CORE_DEVICE_ANNOUNCE = 0x6461;
    protected static final int RDPDR_CORE_DEVICELIST_ANNOUNCE = 0x4441;
    protected static final int RDPDR_CORE_DEVICE_IOREQUEST = 0x4952;
    protected static final int RDPDR_CORE_DEVICE_IOCOMPLETION = 0x4943;
    protected static final int RDPDR_CORE_DEVICELIST_REPLY = 0x4452;
    protected static final int RDPDR_CORE_DEVICE_REPLY = 0x6472;
    protected static final int RDPDR_CORE_DEVICE_REMOVE = 0x646D;
    protected static final int RDPDR_CORE_DEVICELIST_REMOVE = 0x444D;
    
    protected static final int RDPDR_PRN_CACHE_DATA = 0x5043;
    
    protected static final int RDPDR_GENERAL_CAPABILITY = 0x01;
    protected static final int RDPDR_PRINTER_CAPABILITY = 0x02;
    protected static final int RDPDR_COMPORT_CAPABILITY = 0x03;
    protected static final int RDPDR_DRIVE_CAPABILITY = 0x04;
    protected static final int RDPDR_SMARTCARD_CAPABILITY = 0x05;
    
    protected static final int RDPDR_OS_TYPE_UNKNOWN = 0x00;
    protected static final int RDPDR_OS_TYPE_WIN9X = 0x01;
    protected static final int RDPDR_OS_TYPE_WINNT = 0x02;
    
    protected static final int RDPDR_DTYP_COMPORT = 0x01;
    protected static final int RDPDR_DTYP_LPTPORT = 0x02;
    protected static final int RDPDR_DTYP_PRINTER = 0x04;
    protected static final int RDPDR_DTYP_DISK = 0x08;
    protected static final int RDPDR_DTYP_PRINTERPORT = 0x10;
    protected static final int RDPDR_DTYP_SMARTCARD = 0x20;
    
    protected static final int RDPDR_DEVICE_REPLY_SUCCESS = 0;
    protected static final int RDPDR_DEVICE_REPLY_REJECTED = 1;
    
    protected static final int IRP_MJ_CREATE = 0x00;
    protected static final int IRP_MJ_CLOSE = 0x02;
    protected static final int IRP_MJ_READ = 0x03;
    protected static final int IRP_MJ_WRITE = 0x04;
    protected static final int IRP_MJ_QUERY_INFORMATION = 0x05;
    protected static final int IRP_MJ_SET_INFORMATION = 0x06;
    protected static final int IRP_MJ_QUERY_VOLUME_INFORMATION = 0x0A;
    protected static final int IRP_MJ_SET_VOLUME_INFORMATION = 0x0B;
    protected static final int IRP_MJ_DIRECTORY_CONTROL = 0x0C;
    protected static final int IRP_MJ_FILE_SYSTEM_CONTROL = 0x0D;
    protected static final int IRP_MJ_DEVICE_CONTROL = 0x0E;
    protected static final int IRP_MJ_LOCK_CONTROL = 0x11;
    protected static final int IRP_MJ_QUERY_SECURITY = 0x14;
    protected static final int IRP_MJ_SET_SECURITY = 0x15;
    
    protected static final int IRP_MN_QUERY_DIRECTORY = 0x01;
    protected static final int IRP_MN_NOTIFY_CHANGE_DIRECTORY = 0x02;

    protected static final int GENERIC_READ = 0x80000000;
    protected static final int GENERIC_WRITE = 0x40000000;
    protected static final int GENERIC_EXECUTE = 0x20000000;
    protected static final int GENERIC_ALL = 0x10000000;
    
    protected static final int FILE_READ_DATA = 0x0001;
    protected static final int FILE_WRITE_DATA = 0x0002;
    protected static final int FILE_APPEND_DATA = 0x0004;
    protected static final int FILE_EXECUTE = 0x0020;
    
    protected static final int FILE_SUPERSEDE = 0;
    protected static final int FILE_OPEN = 1;
    protected static final int FILE_CREATE = 2;
    protected static final int FILE_OPEN_IF = 3;
    protected static final int FILE_OVERWRITE = 4;
    protected static final int FILE_OVERWRITE_IF = 5;
    
    protected static final int FILE_DIRECTORY_FILE = 0x00000001;
    protected static final int FILE_WRITE_THROUGH = 0x00000002;
    protected static final int FILE_SEQUENTIAL_ONLY = 0x00000004;
    protected static final int FILE_NON_DIRECTORY_FILE = 0x00000040;
    protected static final int FILE_RANDOM_ACCESS = 0x00000800;
    protected static final int FILE_DELETE_ON_CLOSE = 0x00001000;
    
    protected static final int FILE_ATTRIBUTE_READONLY = 0x00000001;
    protected static final int FILE_ATTRIBUTE_HIDDEN = 0x00000002;
    protected static final int FILE_ATTRIBUTE_SYSTEM = 0x00000004;
    protected static final int FILE_ATTRIBUTE_DIRECTORY = 0x00000010;
    protected static final int FILE_ATTRIBUTE_ARCHIVE = 0x00000020;
    protected static final int FILE_ATTRIBUTE_DEVICE = 0x00000040;
    protected static final int FILE_ATTRIBUTE_NORMAL = 0x00000080;

    protected static final int STATUS_SUCCESS = 0x00000000;
    protected static final int STATUS_NOT_IMPLEMENTED = 0x00000001;
    protected static final int STATUS_NO_MORE_FILES = 0x80000006;
    protected static final int STATUS_PENDING = 0x00000103;
    protected static final int STATUS_INFO_LENGTH_MISMATCH = 0xC0000004;
    protected static final int STATUS_INVALID_HANDLE = 0xC0000008;
    protected static final int STATUS_INVALID_PARAMETER = 0xC000000D;
    protected static final int STATUS_NO_SUCH_FILE = 0xC000000F;
    protected static final int STATUS_ACCESS_DENIED = 0xC0000022;
    protected static final int STATUS_FILE_IS_A_DIRECTORY = 0xC00000BA;
    protected static final int STATUS_NOT_SUPPORTED = 0xC00000BB;
    protected static final int STATUS_CANCELLED = 0xC0000120;
    
    private static RdpdrChannel singleton = null;

    private boolean redirectDrives = true;
    private boolean redirectPrinters = true;
    private boolean redirectComPorts = true;
    private boolean redirectSmartCards = true;
    
    // Server information.
    private int serverVersion;
    private int serverOsType;
    private int serverOsVersion;
    private int serverMajorVersion;
    private int serverMinorVersion;
    private int serverIoRequestFlags;
    private int serverExtendedPduFlags;
    
    // Client information.
    private int clientId;
    private String clientName = null;
    private byte[] clientAddr = null;
    private int clientVersion = 1;
    private int clientOsType = RDPDR_OS_TYPE_WINNT;
    private int clientOsVersion = 0x00050002;
    private int clientMajorVersion = 1;
    private int clientMinorVersion = 5;
    private int clientIoRequestFlags = 0xFFFF;
    private int clientExtendedPduFlags = 0x03;
    
    // Device information.
    private Device[] deviceList = null;
    
    public RdpdrChannel( Options options, boolean redirectDrives, boolean redirectPrinters, boolean redirectComPorts, boolean redirectSmartCards)
    {
    	super( options );
    	
    	singleton = this;

        this.redirectDrives = redirectDrives;
        this.redirectPrinters = redirectPrinters;
        this.redirectComPorts = redirectComPorts;
        this.redirectSmartCards = redirectSmartCards;

    	try
    	{
    		InetAddress inetAddress = InetAddress.getLocalHost();
    		clientName = inetAddress.getHostName().toUpperCase();
    		clientAddr = inetAddress.getAddress();
    	}
    	catch (Exception e) {}
    }
    
    public int flags()
    {
        return 0; 
            //VChannels.CHANNEL_OPTION_INITIALIZED | 
            //VChannels.CHANNEL_OPTION_ENCRYPT_RDP |
            //VChannels.CHANNEL_OPTION_COMPRESS_RDP | 
            //VChannels.CHANNEL_OPTION_SHOW_PROTOCOL;
    }

    public String name()
    {
        return "RDPDR";
    }

    public void process(RdpPacket packet) throws RdesktopException, IOException, CryptoException
    {
    	int component = packet.getLittleEndian16();
    	int packetId = packet.getLittleEndian16();
    	
    	logger.info("component=" + Integer.toHexString(component) + ",packetId=" + Integer.toHexString(packetId));

    	try {
    	switch (component) {
    	case RDPDR_CTYP_CORE:
    		switch (packetId) {
    		case RDPDR_CORE_SERVER_ANNOUNCE:
    			processServerAnnounce(packet);
    			break;
    		case RDPDR_CORE_CLIENTID_CONFIRM:
    			processClientIdConfirm(packet);
    			break;
    		case RDPDR_CORE_SERVER_CAPABILITY:
    			processServerCapability(packet);
    			break;
    		case RDPDR_CORE_DEVICELIST_REPLY:
    			processDeviceListReply(packet);
    			break;
    		case RDPDR_CORE_DEVICE_REPLY:
    			processDeviceReply(packet);
    			break;
    		case RDPDR_CORE_DEVICELIST_REMOVE:
    			processDeviceListRemove(packet);
    			break;
    		case RDPDR_CORE_DEVICE_REMOVE:
    			processDeviceRemove(packet);
    			break;
    		case RDPDR_CORE_DEVICE_IOREQUEST:
    			processDeviceIoRequest(packet);
    		default:
    			break;
    		}
    		break;
    	case RDPDR_CTYP_PRN:
    		switch (packetId) {
    		case RDPDR_PRN_CACHE_DATA:
    			processPrinterCacheData(packet);
    			break;
    		default:
    			break;
    		}
    		break;
    	default:
    		break;
    	}
    	} catch (Exception e) {
    		logger.error(e.getMessage(), e.fillInStackTrace());
    	}
    }
    
    private void processServerAnnounce(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_SERVER_ANNOUNCE packet...");
    	
    	// Parse the incoming packet.
    	serverMajorVersion = packet.getLittleEndian16();
    	serverMinorVersion = packet.getLittleEndian16();
    	clientId = packet.getLittleEndian32();
    	
    	// Send the response.
    	sendClientIdConfirm();
    	sendClientName();
    }
    
    private void processClientIdConfirm(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_CLIENTID_CONFIRM packet...");
    	
    	// Parse the incoming packet.
    	serverMajorVersion = packet.getLittleEndian16();
    	serverMinorVersion = packet.getLittleEndian16();
    	clientId = packet.getLittleEndian32();

	// Send the response.
	sendDeviceListAnnounce();
    }
    
    private void processServerCapability(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_SERVER_CAPABILITY packet...");

    	boolean supportsRedirectDrives = false;
    	boolean supportsRedirectPrinters = false;
    	boolean supportsRedirectComPorts = false;
    	boolean supportsRedirectSmartCards = false;

    	// Parse the incoming packet.
    	int numCapabilities = packet.getLittleEndian32();
    	for (int i = 0; i < numCapabilities; i++)
    	{
    		int capabilityType = packet.getLittleEndian16();
    		int capabilityLength = packet.getLittleEndian16();
    		int bytesRead = 4;
    		switch (capabilityType) {
    		case RDPDR_GENERAL_CAPABILITY:
    			serverVersion = packet.getLittleEndian32();
    			serverOsType = packet.getLittleEndian32();
    			serverOsVersion = packet.getLittleEndian32();
    			serverMajorVersion = packet.getLittleEndian16();
    			serverMinorVersion = packet.getLittleEndian16();
    			serverIoRequestFlags = packet.getLittleEndian32();
    			packet.getLittleEndian32();
    			serverExtendedPduFlags = packet.getLittleEndian32();
    			packet.getLittleEndian32();
    			packet.getLittleEndian32();
    			bytesRead += 36;
    			break;
    		case RDPDR_PRINTER_CAPABILITY:
    			supportsRedirectPrinters = packet.getLittleEndian32() != 0 ? true : false;
    			bytesRead += 4;
    			break;
    		case RDPDR_COMPORT_CAPABILITY:
    			supportsRedirectComPorts = packet.getLittleEndian32() != 0 ? true : false;
    			bytesRead += 4;
    			break;
    		case RDPDR_DRIVE_CAPABILITY:
    			supportsRedirectDrives = packet.getLittleEndian32() != 0 ? true : false;
    			bytesRead += 4;
    			break;
    		case RDPDR_SMARTCARD_CAPABILITY:
    			supportsRedirectSmartCards = packet.getLittleEndian32() != 0 ? true : false;
    			bytesRead += 4;
    			break;
    		default:
    			break;
    		}

			while (bytesRead < capabilityLength) {
				packet.get8();
				bytesRead++;
			}
    	}

    	if (!supportsRedirectDrives) {
			this.redirectDrives = false;
		}
		if (!supportsRedirectPrinters) {
			this.redirectPrinters = false;
		}
		if (!supportsRedirectComPorts) {
			this.redirectComPorts = false;
		}
		if (!supportsRedirectSmartCards) {
			this.redirectSmartCards = false;
		}
		
		// Send the response.
		sendClientCapability();
    }

    private void processDeviceListReply(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_DEVICELIST_REPLY packet...");
    	
    	// Process the incoming packet.
    	int numDevices = packet.getLittleEndian32();
    	for (int i = 0; i < numDevices; i++)
    	{
    		int deviceId = packet.getLittleEndian32();
    		int resultCode = packet.getLittleEndian32();
    		
    		deviceReply(deviceId, resultCode);
    	}
    }
    
    private void processDeviceReply(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_DEVICE_REPLY packet...");

    	// Process the incoming packet.
    	int deviceId = packet.getLittleEndian32();
    	int resultCode = packet.getLittleEndian32();

    	deviceReply(deviceId, resultCode);
    }
    
    private void processDeviceListRemove(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_DEVICELIST_REMOVE packet...");
    	
    	// Process the incoming packet.
    	int numDevices = packet.getLittleEndian32();
    	for (int i = 0; i < numDevices; i++)
    	{
    		int deviceId = packet.getLittleEndian32();
    		
    		deviceRemove(deviceId);
    	}
    }
    
    private void processDeviceRemove(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_DEVICE_REMOVE packet...");

    	// Process the incoming packet.
    	int deviceId = packet.getLittleEndian32();

    	deviceRemove(deviceId);
    }
    
    private void processDeviceIoRequest(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_CORE_DEVICE_IOREQUEST packet...");
    	
    	// Process the incoming packet.
    	IoRequestPacket irp = new IoRequestPacket(packet);

    	logger.info("deviceId=" + irp.deviceId + ",fileId=" + irp.fileId + ",completionId=" + irp.completionId + ",majorFunction=" + irp.majorFunction + ",minorFunction=" + irp.minorFunction);
    	
    	Device device = deviceList[irp.deviceId];

    	device.processIoRequest(irp);
    	if (irp.status != STATUS_PENDING) {
    		sendIoCompletion(irp);
    	}
    }
    
    private void processPrinterCacheData(RdpPacket packet)
    {
    	logger.info("Processing RDPDR_PRN_CACHE_DATA packet...");
    }

    private void sendClientIdConfirm()
    {
    	logger.info("Sending RDPDR_CORE_CLIENTID_CONFIRM packet...");

        RdpPacket_Localised packet = new RdpPacket_Localised(12);
        packet.setLittleEndian16(RDPDR_CTYP_CORE);
        packet.setLittleEndian16(RDPDR_CORE_CLIENTID_CONFIRM);
        packet.setLittleEndian16(clientMajorVersion);
        packet.setLittleEndian16(clientMinorVersion);
        for (int i = 0; i < 4; i++) {
        	if (i < clientAddr.length) {
        		packet.set8(clientAddr[i]);
        	}
        	else {
        		packet.set8(0);
        	}
        }
        packet.markEnd();
        try { send_packet(packet); } catch (Exception e) { logger.error(e.getMessage(), e); }
    }
    
    private void sendClientName()
    {
    	logger.info("Sending RDPDR_CORE_CLIENT_NAME packet...");
    	
    	int nameLength = (clientName.length() + 1) * 2;
    	
    	RdpPacket_Localised packet = new RdpPacket_Localised(16 + nameLength);
        packet.setLittleEndian16(RDPDR_CTYP_CORE);
        packet.setLittleEndian16(RDPDR_CORE_CLIENT_NAME);
    	packet.setLittleEndian32(1);  // ANSI=0, Unicode=1
    	packet.setLittleEndian32(0);  // ANSI code page (0 if Unicode)
    	packet.setLittleEndian32(nameLength);  // client name length
    	packet.outUnicodeString(clientName, nameLength - 2);  // client name
    	packet.markEnd();
    	try { send_packet(packet); } catch (Exception e) { logger.error(e.getMessage(), e); }
    }
    
    private void sendClientCapability()
    {
    	logger.info("Sending RDPDR_CORE_CLIENT_CAPABILITY packet...");
    	
    	RdpPacket_Localised packet = new RdpPacket_Localised(80);
    	packet.setLittleEndian16(RDPDR_CTYP_CORE);
    	packet.setLittleEndian16(RDPDR_CORE_CLIENT_CAPABILITY);    	
    	packet.setLittleEndian32(5);   // numCapabilities
    	
    	packet.setLittleEndian16(RDPDR_GENERAL_CAPABILITY);
    	packet.setLittleEndian16(40);  // capabilityLength
    	packet.setLittleEndian32(clientVersion);
    	packet.setLittleEndian32(clientOsType);
    	packet.setLittleEndian32(clientOsVersion);
    	packet.setLittleEndian16(clientMajorVersion);
    	packet.setLittleEndian16(clientMinorVersion);
    	packet.setLittleEndian32(clientIoRequestFlags);
    	packet.setLittleEndian32(0);
    	packet.setLittleEndian32(clientExtendedPduFlags);
    	packet.setLittleEndian32(0);
    	packet.setLittleEndian32(0);
    	
    	packet.setLittleEndian16(RDPDR_PRINTER_CAPABILITY);
    	packet.setLittleEndian16(8);   // capabilityLength    
    	packet.setLittleEndian32(1);
    	
    	packet.setLittleEndian16(RDPDR_COMPORT_CAPABILITY);
    	packet.setLittleEndian16(8);   // capabilityLength
    	packet.setLittleEndian32(1);
    	
    	packet.setLittleEndian16(RDPDR_DRIVE_CAPABILITY);
    	packet.setLittleEndian16(8);   // capabilityLength
    	packet.setLittleEndian32(1);
    	
    	packet.setLittleEndian16(RDPDR_SMARTCARD_CAPABILITY);
    	packet.setLittleEndian16(8);   // capabilityLength
    	packet.setLittleEndian32(1);
    	
    	packet.markEnd();
    	try { send_packet(packet); } catch (Exception e) { logger.error(e.getMessage(), e); }
    }
    
    private void sendDeviceListAnnounce()
    {
    	logger.info("Sending RDPDR_CORE_DEVICELIST_ANNOUNCE packet...");
    	
    	DiskDevice diskDeviceList[] = null;
    	PrinterDevice printerDeviceList[] = null;
    	
    	int numDisks = 0;
    	int numPrinters = 0;
    	int deviceId = 0;
    	int extraData = 0;

    	if (this.redirectDrives)
    	{
    		diskDeviceList = enumerateDisks();
    		numDisks = (diskDeviceList != null) ? diskDeviceList.length : 0;
    	}
    	
    	if (this.redirectPrinters)
    	{
    		printerDeviceList = enumeratePrinters();
    		numPrinters = (printerDeviceList != null) ? printerDeviceList.length : 0;
    	}
    	
    	int numDevices = numDisks + numPrinters;
    	
    	deviceList = new Device[numDevices];
    	for (int i = 0; i < numDisks; i++)
    	{
    		deviceList[deviceId++] = diskDeviceList[i];
    	}
    	for (int i = 0; i < numPrinters; i++)
    	{
    		//deviceList[deviceId++] = printerDeviceList[i];
    		String driverName = printerDeviceList[i].getDriverName();
    		String printerName = printerDeviceList[i].getPrinterName();
    		int driverNameLength = ((driverName != null) ? (driverName.length() + 1) * 2 : 0);
    		int printerNameLength = ((printerName != null) ? (printerName.length() + 1) * 2 : 0);
    		int cachedFieldsLength = 0;
    		extraData += (24 + driverNameLength + printerNameLength + cachedFieldsLength);
    	}
    	
    	RdpPacket_Localised packet = new RdpPacket_Localised(8 + numDevices * 20 + extraData);
    	packet.setLittleEndian16(RDPDR_CTYP_CORE);
    	packet.setLittleEndian16(RDPDR_CORE_DEVICELIST_ANNOUNCE);
    	packet.setLittleEndian32(numDevices);
    	for (int i = 0; i < numDevices; i++)
    	{
	    	packet.setLittleEndian32(deviceList[i].getDeviceType());
	    	packet.setLittleEndian32(i);
	    	packet.out_uint8p(deviceList[i].getDeviceName(), 8);
	    	switch (deviceList[i].getDeviceType())
	    	{
	    	case RDPDR_DTYP_PRINTER:
	    		PrinterDevice printerDevice = (PrinterDevice)deviceList[i];
	    		String driverName = printerDevice.getDriverName();
	    		String printerName = printerDevice.getPrinterName();
	    		int driverNameLength = ((driverName != null) ? (driverName.length() + 1) * 2 : 0);
	    		int printerNameLength = ((printerName != null) ? (printerName.length() + 1) * 2 : 0);
	    		int cachedFieldsLength = 0;
	    		packet.setLittleEndian32(24 + driverNameLength + printerNameLength + cachedFieldsLength);
	    		packet.setLittleEndian32(0);  // flags
	    		packet.setLittleEndian32(0);  // code page
	    		packet.setLittleEndian32(0);  // reserved
	    		packet.setLittleEndian32(driverNameLength);
	    		packet.setLittleEndian32(printerNameLength);
	    		packet.setLittleEndian32(cachedFieldsLength);
	    		if (driverNameLength > 0) {
	    			packet.outUnicodeString(driverName, driverNameLength - 2);
	    		}
	    		if (printerNameLength > 0) {
	    			packet.outUnicodeString(printerName, printerNameLength - 2);
	    		}
	    		break;
	    	default:
	    		packet.setLittleEndian32(0);
	    		break;
	    	}
    	}
    	packet.markEnd();
    	try { send_packet(packet); } catch (Exception e) { logger.error(e.getMessage(), e); }
    }

    protected static synchronized void sendIoCompletion(IoRequestPacket irp)
    {
    	logger.info("Sending RDPDR_CORE_DEVICE_IOCOMPLETION packet...");

        RdpPacket_Localised packet = new RdpPacket_Localised(16 + (irp.result != null ? irp.result.getPosition() : 8));
        packet.setLittleEndian16(RDPDR_CTYP_CORE);
        packet.setLittleEndian16(RDPDR_CORE_DEVICE_IOCOMPLETION);
        packet.setLittleEndian32(irp.deviceId);
        packet.setLittleEndian32(irp.completionId);
        packet.setLittleEndian32(irp.status);
        if (irp.result != null && irp.status == RdpdrChannel.STATUS_SUCCESS)
        {
        	int length = irp.result.getPosition();
        	packet.copyFromPacket(irp.result, 0, packet.getPosition(), length);
        	packet.incrementPosition(length);
        } else {
        	packet.setLittleEndian64(0); // pad
        }
        
        packet.markEnd();
        try { singleton.send_packet(packet); } catch (Exception e) { logger.error(e.getMessage(), e); }
    }
    
    private DiskDevice[] enumerateDisks()
    {
    	try {
    		// Enumerate file systems.
        	File roots[] = File.listRoots();
        	DiskDevice[] deviceList = null;
        	int numDisks = (roots != null) ? roots.length : 0;
        	if (numDisks > 0)
        	{
        		deviceList = new DiskDevice[numDisks];
	        	for (int i = 0; i < numDisks; i++) {
	        		logger.info("path=" + roots[i].getPath() +",parent=" + roots[i].getParent());
	        		deviceList[i] = new DiskDevice(roots[i]);
	        	}
        	}
        	return deviceList;
        } catch (Exception e) { return null; }        	
    }
    
    private PrinterDevice[] enumeratePrinters()
    {
    	//PrintService printServices[] = PrintServiceLookup.lookupPrintServices(null, null);
    	PrinterDevice deviceList[] = null;
    	/*
    	int numPrinters = (printServices != null) ? printServices.length : 0;
    	if (numPrinters > 0)
    	{
    		deviceList = new PrinterDevice[numPrinters];
    		for (int i = 0; i < numPrinters; i++) {
    			logger.info("printer=" + printServices[i].getName());
    			deviceList[i] = new PrinterDevice("PRN" + i, printServices[i]);
    		}
    	}
    	*/
    	return deviceList;
    }
    
    private void deviceReply(int deviceId, int resultCode)
    {
    	logger.info("deviceReply: deviceId=" + deviceId + ",resultCode=" + resultCode);
    	
    	switch (resultCode)
    	{
    	case RDPDR_DEVICE_REPLY_SUCCESS:
    		break;
    	case RDPDR_DEVICE_REPLY_REJECTED:
    		deviceList[deviceId] = null;
    		break;
    	}
    }
    
    private void deviceRemove(int deviceId)
    {
    	logger.info("deviceRemove: deviceId=" + deviceId);
    	
    	deviceList[deviceId] = null;
    }
}
