package com.provisionnetworks.rdp5.rdpsnd;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import org.apache.log4j.Logger;

import android.media.AudioFormat;
import android.media.AudioManager;
import android.media.AudioTrack;

public class WindowsAudioFormat extends AudioTrack
{
	
	protected static Logger logger = Logger.getLogger(WindowsAudioFormat.class);
	
	protected final static int WAVE_FORMAT_PCM 		= 0x0001;
	protected final static int WAVE_FORMAT_ADPCM 	= 0x0002;
	protected final static int WAVE_FORMAT_IEEE_FLOAT = 0x0003;
	protected final static int WAVE_FORMAT_ALAW 	= 0x0006;
	protected final static int WAVE_FORMAT_MULAW 	= 0x0007;
	protected final static int WAVE_FORMAT_DVI_ADPCM= 0x0011;
	
	protected final static int FORMAT_HEADER_SIZE 	= 18;
	
	public WindowsAudioFormat( WindowsAudioFormat format ) 
    {
	    super( 
                AudioManager.STREAM_MUSIC, 
                (int) format.samplesPerSec, 
                format.channels,
                getEncoding( format.format, format.bitsPerSample ),
                WindowsAudioFormat.getMinBufferSize( format.getSampleRate(), format.getChannelCount(), format.getAudioFormat() ),
                MODE_STREAM
                );
        this.format = format.format;
        this.channels = format.channels;
        this.samplesPerSec = format.samplesPerSec;
        this.avgBytesPerSec = format.avgBytesPerSec;
        this.blockAlign = format.blockAlign;
        this.bitsPerSample = format.bitsPerSample;
        this.dataSize = format.dataSize;
        this.data = format.data;
        
        this.formatSize = FORMAT_HEADER_SIZE + format.dataSize;
    }
	
	public WindowsAudioFormat(int format, int channels, long samplePerSec, long avgBytesPerSec, int blockAlign, int bitsPerSample, int dataSize, byte[] data) 
	{
	    super( 
	            AudioManager.STREAM_MUSIC, 
	            (int)samplePerSec, 
	            channels == 1 ? AudioFormat.CHANNEL_OUT_MONO : AudioFormat.CHANNEL_OUT_STEREO,
                getEncoding( format, bitsPerSample ),
                WindowsAudioFormat.getMinBufferSize( 
                        (int) samplePerSec, 
                        channels == 1 ? AudioFormat.CHANNEL_OUT_MONO : AudioFormat.CHANNEL_OUT_STEREO, 
                        getEncoding( format, bitsPerSample ) ),
                MODE_STREAM
                );
	    
	    logger.info("[AUDIO] AudioTrack initialization: format=" + format +
	            " channels=" + channels + 
	            " samplePerSec=" + samplePerSec +
	            " avgBytesPerSec=" + avgBytesPerSec +
	            " blockAlign=" + blockAlign +
	            " bitsPerSample=" + bitsPerSample +
	            " dataSize=" + dataSize);
	    
		this.format = format;
		this.channels = channels;
		this.samplesPerSec = samplePerSec;
		this.avgBytesPerSec = avgBytesPerSec;
		this.blockAlign = blockAlign;
		this.bitsPerSample = bitsPerSample;
		this.dataSize = dataSize;
		this.data = data;
		
		this.formatSize = FORMAT_HEADER_SIZE + dataSize;
	}
	
	private int format;
	private int channels;
	private long samplesPerSec;
	private long avgBytesPerSec;
	private int blockAlign;
	private int bitsPerSample;
	private int dataSize;
	private byte[] data;
	
	private int formatSize = 0;
		
	public static int getEncoding(int format, int bitsPerSample)
	{
		logger.info("[AUDIO] format=" + format);
		
		switch(format)
		{
		case WAVE_FORMAT_PCM:
			logger.info("[AUDIO] Audio Encoding is PCM");
			if(bitsPerSample == 8)
			{
				return AudioFormat.ENCODING_PCM_8BIT;
			}
			else if(bitsPerSample == 16)
			{
				return AudioFormat.ENCODING_PCM_16BIT;
			}
			return AudioFormat.ENCODING_INVALID;
		case WAVE_FORMAT_ALAW:
			logger.info("[AUDIO] Audio Encoding is WAVE_FORMAT_ALAW");
			return AudioFormat.ENCODING_INVALID;
		case WAVE_FORMAT_MULAW:
			logger.info("[AUDIO] Audio Encoding is WAVE_FORMAT_MULAW");
			return AudioFormat.ENCODING_INVALID;
		case WAVE_FORMAT_ADPCM:
        case WAVE_FORMAT_DVI_ADPCM:
        case WAVE_FORMAT_IEEE_FLOAT:
            logger.info("[AUDIO] Audio Encoding is not supported PCMs");
            return AudioFormat.ENCODING_INVALID;
		default:
			logger.info("[AUDIO] Audio Encoding is UNKNOWN");
			return AudioFormat.ENCODING_INVALID;
		}
	}

	public byte[] toByteArray()
	{
		ByteBuffer bb = ByteBuffer.allocateDirect(this.getSize());
		bb.order(ByteOrder.LITTLE_ENDIAN);
		bb.putShort((short)format);
		bb.putShort((short)channels);
		bb.putInt((int)samplesPerSec);
		bb.putInt((int)avgBytesPerSec);
		bb.putShort((short)blockAlign);
		bb.putShort((short)bitsPerSample);
		bb.putShort((short)dataSize);
		
		if(dataSize > 0) bb.put(data);
		
		byte[] res = new byte[formatSize];
		bb.position(0);
		bb.get(res);
		return res;
	}
	
	public int getSize()
	{
		return formatSize;
	}
	
	public boolean isSupportedFormat()
	{		
		try 
		{
			if ( super.getAudioFormat() == AudioFormat.ENCODING_INVALID )
			{
			    return false;
			}
		} 
		catch (Exception e) 
		{
			logger.info( "[AUDIO] " + e.getMessage(), e );
			return false;
		}
		logger.info("[AUDIO] Supported Format: " + this.toString());
		
		return true;
	}
}
