/*
 * Bitmap.java Component: ProperJavaRDP Revision: $Revision: 1.7 $ Author:
 * $Author: telliott $ Date: $Date: 2005/09/27 14:15:39 $ Copyright (c) 2005
 * Propero Limited Purpose: Provide a class for storage of Bitmap images, along
 * with static methods for decompression and conversion of bitmaps. This program
 * is free software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. This
 * program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA (See gpl.txt for details of the
 * GNU General Public License.)
 */

package net.propero.rdp;

import org.apache.log4j.Logger;

import quest.androidport.IndexColorModel;

public class Bitmap
{

    public int              usage;

    private int[]           highdata = null;

    private int             width    = 0;

    private int             height   = 0;

    private int             x        = 0;

    private int             y        = 0;

    protected static Logger logger   = Logger.getLogger( Bitmap.class );

    public static int convertTo24( int colour, int bpp )
    {
        if ( bpp == 15 )
            return convert15to24( colour );
        if ( bpp == 16 )
            return convert16to24( colour );
        return colour;
    }

    public static int convert15to24( int colour16 )
    {
        int r24 = (colour16 >> 7) & 0xF8;
        int g24 = (colour16 >> 2) & 0xF8;
        int b24 = (colour16 << 3) & 0xFF;

        r24 |= r24 >> 5;
        g24 |= g24 >> 5;
        b24 |= b24 >> 5;

        return (r24 << 16) | (g24 << 8) | b24;
    }

    public static int convert16to24( int colour16 )
    {
        int r24 = (colour16 >> 8) & 0xF8;
        int g24 = (colour16 >> 3) & 0xFC;
        int b24 = (colour16 << 3) & 0xFF;

        r24 |= r24 >> 5;
        g24 |= g24 >> 6;
        b24 |= b24 >> 5;

        return (r24 << 16) | (g24 << 8) | b24;
    }

    private static int bmpCount = 0;

    /**
     * Read integer of a specified byte-length from byte array
     * 
     * @param data
     *            Array to read from
     * @param offset
     *            Offset in array to read from
     * @param bpp
     *            Number of bytes to read
     * @return
     */
    static int cvalx( byte[] data, int offset, int bpp )
    {
        int rv = 0;
        
        if ( data.length <= offset + ((bpp + 7) / 8) - 1 )
        {
            /*
             * TODO:  
             * HOTFIX. Incorrect bpp value.
             */
            logger.error( "Incorrect data size: " + data.length + " " + offset + " " + bpp );
            bpp = 16;
        }
        
        if ( bpp == 15 )
        {
            int lower = data[offset] & 0xFF;
            int full = (data[offset + 1] & 0xFF) << 8 | lower;

            int r24 = (full >> 7) & 0xF8;
            r24 |= r24 >> 5;
            int g24 = (full >> 2) & 0xF8;
            g24 |= g24 >> 5;
            int b24 = (lower << 3) & 0xFF;
            b24 |= b24 >> 5;

            rv = (r24 << 16) | (g24 << 8) | b24;

        }
        else if ( bpp == 16 )
        {
            int lower = data[offset] & 0xFF;
            int full = (data[offset + 1] & 0xFF) << 8 | lower;

            int r24 = (full >> 8) & 0xF8;
            r24 |= r24 >> 5;
            int g24 = (full >> 3) & 0xFC;
            g24 |= g24 >> 6;
            int b24 = (lower << 3) & 0xFF;
            b24 |= b24 >> 5;

            rv = (r24 << 16) | (g24 << 8) | b24;

        }
        else if ( bpp >= 24 )
        {
            for( int i = ((bpp >> 3) - 1); i >= 0; i-- )
            {
                rv = rv << 8;
                rv |= data[offset + i] & 0xFF;
            }
        }

        return rv;
    }

    /**
     * @param input
     * @param startOffset
     * @param offset
     * @param Bpp
     * @return
     */
    static int getli( byte[] input, int startOffset, int offset, int Bpp )
    {
        int rv = 0;

        int rOffset = startOffset + (offset * Bpp);
        for( int i = 0; i < Bpp; i++ )
        {
            rv = rv << 8;
            rv |= (input[rOffset + (Bpp - i - 1)]) & 0xFF;
        }
        return rv;
    }

    /**
     * @param input
     * @param startlocation
     * @param offset
     * @param value
     * @param Bpp
     */
    static void setli( byte[] input, int startlocation, int offset, int value,
            int Bpp )
    {
        int location = startlocation + offset * Bpp;

        input[location] = (byte) (value & 0xFF);
        if ( Bpp > 1 )
            input[location + 1] = (byte) ((value & 0xFF00) >> 8);
        if ( Bpp > 2 )
            input[location + 2] = (byte) ((value & 0xFF0000) >> 16);
    }

    /**
     * Convert byte array representing a bitmap into integer array of pixels
     * 
     * @param bitmap
     *            Byte array of bitmap data
     * @param Bpp
     *            Bytes-per-pixel for bitmap
     * @return Integer array of pixel data representing input image data
     */
    static int[] convertImage( byte[] bitmap, int Bpp, int bpp,
            IndexColorModel cm )
    {
        int[] out = new int[bitmap.length / Bpp];

        final int alpha = 0xFF000000;
        
        if ( Bpp == 1 )
        {
            if ( cm != null )
            {
                for ( int i = 0; i < out.length; ++i )
                {
                    out[i] = cm.getRGB( bitmap[i] & 0xFF );
                }
            }
            else
            {
                for ( int i = 0; i < out.length; ++i )
                {
                    out[i] = bitmap[i] & 0xFF;
                }
            }
        }
        else if ( bpp == 15 )
        {
            int current = 0;
            for ( int i = 0; i < out.length; ++i )
            {
                out[i] = ((bitmap[current + 1] & 0xFF) << 8) | (bitmap[current] & 0xFF);
                out[i] = Bitmap.convert15to24( out[i] ) | alpha;
                current += Bpp;
            }
        }
        else if ( bpp == 16 )
        {
            int current = 0;
            for ( int i = 0; i < out.length; ++i )
            {
                out[i] = ((bitmap[current + 1] & 0xFF) << 8) | (bitmap[current] & 0xFF);
                out[i] = Bitmap.convert16to24( out[i] ) | alpha;
                current += Bpp;
            }
        }
        else if ( bpp == 24 )
        {
            int current = 0;
            for ( int i = 0; i < out.length; ++i )
            {
                out[i] = ((bitmap[current + 2] & 0xFF) << 16)
                | ((bitmap[current + 1] & 0xFF) << 8)
                | (bitmap[current] & 0xFF)
                | alpha;
                current += Bpp;
            }
        }
        else if ( bpp == 32 )
        {
            int current = 0;
            
            for ( int i = 0; i < out.length; ++i )
            {
                out[i] = ((bitmap[current + 3] & 0xFF) << 16)
                | ((bitmap[current + 2] & 0xFF) << 8)
                | (bitmap[current + 1] & 0xFF)
                | (bitmap[current] & 0xFF);
                current += Bpp;
            }
        }
        
//        for( int i = 0; i < out.length; i++ )
//        {
//            if ( Bpp == 1 )
//            {
//                out[i] = cm != null ? cm.getRGB( bitmap[i] & 0xFF )
//                        : bitmap[i] & 0xFF;
//            }
//            else if ( Bpp == 2 )
//            {
//                out[i] = ((bitmap[i * Bpp + 1] & 0xFF) << 8)
//                        | (bitmap[i * Bpp] & 0xFF);
//                out[i] = Bitmap.convertTo24( out[i], bpp );
//            }
//            else if ( Bpp == 3 )
//            {
//                out[i] = ((bitmap[i * Bpp + 2] & 0xFF) << 16)
//                        | ((bitmap[i * Bpp + 1] & 0xFF) << 8)
//                        | (bitmap[i * Bpp] & 0xFF)
//                        | 0xFF000000;
//            }
//            else if ( Bpp == 4 )
//                out[i] = ((bitmap[i * Bpp + 3] & 0xFF) << 24)
//                        | ((bitmap[i * Bpp + 2] & 0xFF) << 16)
//                        | ((bitmap[i * Bpp + 1] & 0xFF) << 8)
//                        | (bitmap[i * Bpp] & 0xFF);
//        }

        return out;
    }

    /**
     * Constructor for Bitmap based on integer pixel values
     * 
     * @param data
     *            Array of pixel data, one integer per pixel. Should have a
     *            length of width*height.
     * @param width
     *            Width of bitmap represented by data
     * @param height
     *            Height of bitmap represented by data
     * @param x
     *            Desired x-coordinate of bitmap
     * @param y
     *            Desired y-coordinate of bitmap
     */
    public Bitmap( int[] data, int width, int height, int x, int y )
    {
        this.highdata = data;
        this.width = width;
        this.height = height;
        this.x = x;
        this.y = y;
    }

    /**
     * Constructor for Bitmap based on
     * 
     * @param data
     *            Array of pixel data, each pixel represented by Bpp bytes.
     *            Should have a length of width*height*Bpp.
     * @param width
     *            Width of bitmap represented by data
     * @param height
     *            Height of bitmap represented by data
     * @param x
     *            Desired x-coordinate of bitmap
     * @param y
     *            Desired y-coordinate of bitmap
     * @param Bpp
     *            Number of bytes per pixel in image represented by data
     */
    public Bitmap( byte[] data, int width, int height, int x, int y, int Bpp,
            int bpp )
    {
        this.highdata = Bitmap.convertImage( data, Bpp, bpp, null );
        this.width = width;
        this.height = height;
        this.x = x;
        this.y = y;
    }

    /**
     * Retrieve data representing this Bitmap, as an array of integer pixel
     * values
     * 
     * @return Bitmap pixel data
     */
    public int[] getBitmapData()
    {
        return this.highdata;
    }

    /**
     * Retrieve width of the bitmap represented by this object
     * 
     * @return Bitmap width
     */
    public int getWidth()
    {
        return this.width;
    }

    /**
     * Retrieve height of the bitmap represented by this object
     * 
     * @return Bitmap height
     */
    public int getHeight()
    {
        return this.height;
    }

    /**
     * Retrieve desired x-coordinate of the bitmap represented by this object
     * 
     * @return x-coordinate of this bitmap
     */
    public int getX()
    {
        return this.x;
    }

    /**
     * Retrieve desired y-coordinate of the bitmap represented by this object
     * 
     * @return y-coordinate of this bitmap
     */
    public int getY()
    {
        return this.y;
    }

    /**
     * Decompress bitmap data from packet and output directly to supplied image
     * object
     * 
     * @param width
     *            Width of bitmap to decompress
     * @param height
     *            Height of bitmap to decompress
     * @param size
     *            Size of compressed data in bytes
     * @param data
     *            Packet containing bitmap data
     * @param Bpp
     *            Bytes per-pixel for bitmap
     * @param cm
     *            Colour model of bitmap
     * @param left
     *            X offset for drawing bitmap
     * @param top
     *            Y offset for drawing bitmap
     * @param w
     *            Image to draw bitmap to
     * @return Original image object, with decompressed bitmap drawn at
     *         specified coordinates
     * @throws RdesktopException
     */
    public static android.graphics.Bitmap decompressImgDirect( int width,
            int height, int size, RdpPacket_Localised data, int Bpp, int left,
            int top, android.graphics.Bitmap w ) throws RdesktopException
    {

        // android.graphics.Bitmap w = null;

        byte[] compressed_pixel = new byte[size];
        data.copyToByteArray( compressed_pixel, 0, data.getPosition(), size );
        data.incrementPosition( size );

        int previous = -1, line = 0, prevY = 0;
        int input = 0, end = size;
        int opcode = 0, count = 0, offset = 0, x = width;
        int lastopcode = -1, fom_mask = 0;
        int code = 0, color1 = 0, color2 = 0;
        byte mixmask = 0;
        int mask = 0;
        int mix = 0xffffffff;

        boolean insertmix = false, bicolor = false, isfillormix = false;

        while ( input < end )
        {
            fom_mask = 0;
            code = (compressed_pixel[input++] & 0x000000ff);
            opcode = code >> 4;

            /* Handle different opcode forms */
            switch ( opcode )
            {
                case 0xc:
                case 0xd:
                case 0xe:
                    opcode -= 6;
                    count = code & 0xf;
                    offset = 16;
                    break;

                case 0xf:
                    opcode = code & 0xf;
                    if ( opcode < 9 )
                    {
                        count = (compressed_pixel[input++] & 0xff);
                        count |= ((compressed_pixel[input++] & 0xff) << 8);
                    }
                    else
                    {
                        count = (opcode < 0xb) ? 8 : 1;
                    }
                    offset = 0;
                    break;

                default:
                    opcode >>= 1;
                    count = code & 0x1f;
                    offset = 32;
                    break;
            }

            /* Handle strange cases for counts */
            if ( offset != 0 )
            {
                isfillormix = ((opcode == 2) || (opcode == 7));

                if ( count == 0 )
                {
                    if ( isfillormix )
                        count = (compressed_pixel[input++] & 0x000000ff) + 1;
                    else
                        count = (compressed_pixel[input++] & 0x000000ff)
                                + offset;
                }
                else if ( isfillormix )
                {
                    count <<= 3;
                }
            }

            switch ( opcode )
            {
                case 0: /* Fill */
                    if ( (lastopcode == opcode)
                            && !((x == width) && (previous == -1)) )
                        insertmix = true;
                    break;
                case 8: /* Bicolor */
                    color1 = cvalx( compressed_pixel, input, Bpp );
                    // (compressed_pixel[input++]&0x000000ff);
                    input += Bpp;
                case 3: /* Color */
                    color2 = cvalx( compressed_pixel, input, Bpp );
                    // color2 = (compressed_pixel[input++]&0x000000ff);
                    input += Bpp;
                    break;
                case 6: /* SetMix/Mix */
                case 7: /* SetMix/FillOrMix */
                    // mix = compressed_pixel[input++];
                    mix = cvalx( compressed_pixel, input, Bpp );
                    input += Bpp;
                    opcode -= 5;
                    break;
                case 9: /* FillOrMix_1 */
                    mask = 0x03;
                    opcode = 0x02;
                    fom_mask = 3;
                    break;
                case 0x0a: /* FillOrMix_2 */
                    mask = 0x05;
                    opcode = 0x02;
                    fom_mask = 5;
                    break;

            }

            lastopcode = opcode;
            mixmask = 0;

            /* Output body */
            while ( count > 0 )
            {
                if ( x >= width )
                {
                    if ( height <= 0 )
                        throw new RdesktopException(
                                "Decompressing bitmap failed! Height = "
                                        + height );
                    x = 0;
                    height--;

                    previous = line;
                    prevY = previous / width;
                    line = height * width;
                }

                switch ( opcode )
                {
                    case 0: /* Fill */
                        if ( insertmix )
                        {
                            if ( previous == -1 )
                            {
                                // pixel[line+x] = mix;
                                w.setPixel( left + x, top + height, mix );
                            }
                            else
                            {
                                w.setPixel( left + x, top + height,
                                        w.getPixel( left + x, top + prevY )
                                                ^ mix );
                                // pixel[line+x] = (pixel[previous+x] ^ mix);
                            }

                            insertmix = false;
                            count--;
                            x++;
                        }

                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line+x] = 0;
                                    w.setPixel( left + x, top + height, 0 );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line+x] = 0;
                                w.setPixel( left + x, top + height, 0 );
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line + x] = pixel[previous + x];
                                    w.setPixel( left + x, top + height,
                                            w.getPixel( left + x, top + prevY ) );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line + x] = pixel[previous + x];
                                w.setPixel( left + x, top + height,
                                        w.getPixel( left + x, top + prevY ) );
                                count--;
                                x++;
                            }
                        }
                        break;

                    case 1: /* Mix */
                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line + x] = mix;
                                    w.setPixel( left + x, top + height, mix );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line + x] = mix;
                                w.setPixel( left + x, top + height, mix );
                                count--;
                                x++;
                            }
                        }
                        else
                        {

                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line + x] = pixel[previous + x] ^
                                    // mix;
                                    w.setPixel( left + x, top + height,
                                            w.getPixel( left + x, top + prevY )
                                                    ^ mix );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line + x] = pixel[previous + x] ^ mix;
                                w.setPixel( left + x, top + height,
                                        w.getPixel( left + x, top + prevY )
                                                ^ mix );
                                count--;
                                x++;
                            }

                        }
                        break;
                    case 2: /* Fill or Mix */
                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                    {
                                        // pixel[line + x] = (byte) mix;
                                        w.setPixel( left + x, top + height,
                                                (byte) mix );
                                    }
                                    else
                                    {
                                        // pixel[line + x] = 0;
                                        w.setPixel( left + x, top + height, 0 );
                                    }
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                {
                                    // pixel[line + x] = mix;
                                    w.setPixel( left + x, top + height, mix );
                                }
                                else
                                {
                                    // pixel[line + x] = 0;
                                    w.setPixel( left + x, top + height, 0 );
                                }
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                    {
                                        // pixel[line + x] = (pixel[previous +
                                        // x] ^ mix);
                                        w.setPixel(
                                                left + x,
                                                top + height,
                                                w.getPixel( left + x, prevY
                                                        + top )
                                                        ^ mix );
                                    }
                                    else
                                    {
                                        // pixel[line + x] = pixel[previous +
                                        // x];
                                        w.setPixel(
                                                left + x,
                                                top + height,
                                                w.getPixel( left + x, prevY
                                                        + top ) );
                                    }
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                {
                                    // pixel[line + x] = (pixel[previous + x] ^
                                    // mix);
                                    w.setPixel( left + x, top + height,
                                            w.getPixel( left + x, prevY + top )
                                                    ^ mix );
                                }
                                else
                                {
                                    // pixel[line + x] = pixel[previous + x];
                                    w.setPixel( left + x, top + height,
                                            w.getPixel( left + x, prevY + top ) );
                                }
                                count--;
                                x++;
                            }

                        }
                        break;

                    case 3: /* Color */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = color2;
                                w.setPixel( left + x, top + height, color2 );
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = color2;
                            w.setPixel( left + x, top + height, color2 );
                            count--;
                            x++;
                        }

                        break;

                    case 4: /* Copy */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = cvalx(compressed_pixel,
                                // input, Bpp);
                                w.setPixel( left + x, top + height,
                                        cvalx( compressed_pixel, input, Bpp ) );
                                input += Bpp;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = cvalx(compressed_pixel, input,
                            // Bpp);
                            w.setPixel( left + x, top + height,
                                    cvalx( compressed_pixel, input, Bpp ) );
                            input += Bpp;
                            count--;
                            x++;
                        }
                        break;

                    case 8: /* Bicolor */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                if ( bicolor )
                                {
                                    // pixel[line + x] = color2;
                                    w.setPixel( left + x, top + height, color2 );
                                    bicolor = false;
                                }
                                else
                                {
                                    // pixel[line + x] = color1;
                                    w.setPixel( left + x, top + height, color1 );
                                    bicolor = true;
                                    count++;
                                }
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            if ( bicolor )
                            {
                                // pixel[line + x] = color2;
                                w.setPixel( left + x, top + height, color2 );
                                bicolor = false;
                            }
                            else
                            {
                                // pixel[line + x] = color1;
                                w.setPixel( left + x, top + height, color1 );
                                bicolor = true;
                                count++;
                            }
                            count--;
                            x++;
                        }

                        break;

                    case 0xd: /* White */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = 0xffffff;
                                w.setPixel( left + x, top + height, 0xffffff );
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = 0xffffff;
                            w.setPixel( left + x, top + height, 0xffffff );
                            count--;
                            x++;
                        }
                        break;

                    case 0xe: /* Black */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = 0x00;
                                w.setPixel( left + x, top + height, 0x00 );
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = 0x00;
                            w.setPixel( left + x, top + height, 0x00 );
                            count--;
                            x++;
                        }

                        break;
                    default:
                        throw new RdesktopException(
                                "Unimplemented decompress opcode " + opcode );// ;
                }
            }
        }

        /*
         * if(Options.server_bpp == 16){ for(int i = 0; i < pixel.length; i++)
         * pixel[i] = Bitmap.convert16to24(pixel[i]); }
         */

        return w;
    }

    /**
     * Decompress bitmap data from packet and output as an Image
     * 
     * @param width
     *            Width of bitmap
     * @param height
     *            Height of bitmap
     * @param size
     *            Size of compressed data in bytes
     * @param data
     *            Packet containing bitmap data
     * @param Bpp
     *            Bytes per-pixel for bitmap
     * @param cm
     *            Colour model for bitmap (if using indexed palette)
     * @return Decompressed bitmap as Image object
     * @throws RdesktopException
     */
    public static android.graphics.Bitmap decompressImg( int width, int height,
            int size, RdpPacket_Localised data, int Bpp )
            throws RdesktopException
    {

        android.graphics.Bitmap w = null;

        byte[] compressed_pixel = new byte[size];
        data.copyToByteArray( compressed_pixel, 0, data.getPosition(), size );
        data.incrementPosition( size );

        int previous = -1, line = 0, prevY = 0;
        int input = 0, end = size;
        int opcode = 0, count = 0, offset = 0, x = width;
        int lastopcode = -1, fom_mask = 0;
        int code = 0, color1 = 0, color2 = 0;
        byte mixmask = 0;
        int mask = 0;
        int mix = 0xffffffff;

        boolean insertmix = false, bicolor = false, isfillormix = false;

        /*
         * if(cm == null) new android.graphics.Bitmap(width, height,
         * BufferedImage.TYPE_INT_RGB); else w = new
         * android.graphics.Bitmap(width, height, BufferedImage.TYPE_INT_RGB,
         * cm);
         */
        w = android.graphics.Bitmap.createBitmap( width, height, null );

        while ( input < end )
        {
            fom_mask = 0;
            code = (compressed_pixel[input++] & 0x000000ff);
            opcode = code >> 4;

            /* Handle different opcode forms */
            switch ( opcode )
            {
                case 0xc:
                case 0xd:
                case 0xe:
                    opcode -= 6;
                    count = code & 0xf;
                    offset = 16;
                    break;

                case 0xf:
                    opcode = code & 0xf;
                    if ( opcode < 9 )
                    {
                        count = (compressed_pixel[input++] & 0xff);
                        count |= ((compressed_pixel[input++] & 0xff) << 8);
                    }
                    else
                    {
                        count = (opcode < 0xb) ? 8 : 1;
                    }
                    offset = 0;
                    break;

                default:
                    opcode >>= 1;
                    count = code & 0x1f;
                    offset = 32;
                    break;
            }

            /* Handle strange cases for counts */
            if ( offset != 0 )
            {
                isfillormix = ((opcode == 2) || (opcode == 7));

                if ( count == 0 )
                {
                    if ( isfillormix )
                        count = (compressed_pixel[input++] & 0x000000ff) + 1;
                    else
                        count = (compressed_pixel[input++] & 0x000000ff)
                                + offset;
                }
                else if ( isfillormix )
                {
                    count <<= 3;
                }
            }

            switch ( opcode )
            {
                case 0: /* Fill */
                    if ( (lastopcode == opcode)
                            && !((x == width) && (previous == -1)) )
                        insertmix = true;
                    break;
                case 8: /* Bicolor */
                    color1 = cvalx( compressed_pixel, input, Bpp );
                    // (compressed_pixel[input++]&0x000000ff);
                    input += Bpp;
                case 3: /* Color */
                    color2 = cvalx( compressed_pixel, input, Bpp );
                    // color2 = (compressed_pixel[input++]&0x000000ff);
                    input += Bpp;
                    break;
                case 6: /* SetMix/Mix */
                case 7: /* SetMix/FillOrMix */
                    // mix = compressed_pixel[input++];
                    mix = cvalx( compressed_pixel, input, Bpp );
                    input += Bpp;
                    opcode -= 5;
                    break;
                case 9: /* FillOrMix_1 */
                    mask = 0x03;
                    opcode = 0x02;
                    fom_mask = 3;
                    break;
                case 0x0a: /* FillOrMix_2 */
                    mask = 0x05;
                    opcode = 0x02;
                    fom_mask = 5;
                    break;

            }

            lastopcode = opcode;
            mixmask = 0;

            /* Output body */
            while ( count > 0 )
            {
                if ( x >= width )
                {
                    if ( height <= 0 )
                        throw new RdesktopException(
                                "Decompressing bitmap failed! Height = "
                                        + height );
                    x = 0;
                    height--;

                    previous = line;
                    prevY = previous / width;
                    line = height * width;
                }

                switch ( opcode )
                {
                    case 0: /* Fill */
                        if ( insertmix )
                        {
                            if ( previous == -1 )
                            {
                                // pixel[line+x] = mix;
                                w.setPixel( x, height, mix );
                            }
                            else
                            {
                                w.setPixel( x, height, w.getPixel( x, prevY )
                                        ^ mix );
                                // pixel[line+x] = (pixel[previous+x] ^ mix);
                            }

                            insertmix = false;
                            count--;
                            x++;
                        }

                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line+x] = 0;
                                    w.setPixel( x, height, 0 );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line+x] = 0;
                                w.setPixel( x, height, 0 );
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line + x] = pixel[previous + x];
                                    w.setPixel( x, height,
                                            w.getPixel( x, prevY ) );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line + x] = pixel[previous + x];
                                w.setPixel( x, height, w.getPixel( x, prevY ) );
                                count--;
                                x++;
                            }
                        }
                        break;

                    case 1: /* Mix */
                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line + x] = mix;
                                    w.setPixel( x, height, mix );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line + x] = mix;
                                w.setPixel( x, height, mix );
                                count--;
                                x++;
                            }
                        }
                        else
                        {

                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // pixel[line + x] = pixel[previous + x] ^
                                    // mix;
                                    w.setPixel( x, height,
                                            w.getPixel( x, prevY ) ^ mix );
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // pixel[line + x] = pixel[previous + x] ^ mix;
                                w.setPixel( x, height, w.getPixel( x, prevY )
                                        ^ mix );
                                count--;
                                x++;
                            }

                        }
                        break;
                    case 2: /* Fill or Mix */
                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                    {
                                        // pixel[line + x] = (byte) mix;
                                        w.setPixel( x, height, (byte) mix );
                                    }
                                    else
                                    {
                                        // pixel[line + x] = 0;
                                        w.setPixel( x, height, 0 );
                                    }
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                {
                                    // pixel[line + x] = mix;
                                    w.setPixel( x, height, mix );
                                }
                                else
                                {
                                    // pixel[line + x] = 0;
                                    w.setPixel( x, height, 0 );
                                }
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                    {
                                        // pixel[line + x] = (pixel[previous +
                                        // x] ^ mix);
                                        w.setPixel( x, height,
                                                w.getPixel( x, prevY ) ^ mix );
                                    }
                                    else
                                    {
                                        // pixel[line + x] = pixel[previous +
                                        // x];
                                        w.setPixel( x, height,
                                                w.getPixel( x, prevY ) );
                                    }
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                {
                                    // pixel[line + x] = (pixel[previous + x] ^
                                    // mix);
                                    w.setPixel( x, height,
                                            w.getPixel( x, prevY ) ^ mix );
                                }
                                else
                                {
                                    // pixel[line + x] = pixel[previous + x];
                                    w.setPixel( x, height,
                                            w.getPixel( x, prevY ) );
                                }
                                count--;
                                x++;
                            }

                        }
                        break;

                    case 3: /* Color */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = color2;
                                w.setPixel( x, height, color2 );
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = color2;
                            w.setPixel( x, height, color2 );
                            count--;
                            x++;
                        }

                        break;

                    case 4: /* Copy */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = cvalx(compressed_pixel,
                                // input, Bpp);
                                w.setPixel( x, height,
                                        cvalx( compressed_pixel, input, Bpp ) );
                                input += Bpp;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = cvalx(compressed_pixel, input,
                            // Bpp);
                            w.setPixel( x, height,
                                    cvalx( compressed_pixel, input, Bpp ) );
                            input += Bpp;
                            // pixel[line+x] = compressed_pixel[input++];
                            count--;
                            x++;
                        }
                        break;

                    case 8: /* Bicolor */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                if ( bicolor )
                                {
                                    // pixel[line + x] = color2;
                                    w.setPixel( x, height, color2 );
                                    bicolor = false;
                                }
                                else
                                {
                                    // pixel[line + x] = color1;
                                    w.setPixel( x, height, color1 );
                                    bicolor = true;
                                    count++;
                                }
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            if ( bicolor )
                            {
                                // pixel[line + x] = color2;
                                w.setPixel( x, height, color2 );
                                bicolor = false;
                            }
                            else
                            {
                                // pixel[line + x] = color1;
                                w.setPixel( x, height, color1 );
                                bicolor = true;
                                count++;
                            }
                            count--;
                            x++;
                        }

                        break;

                    case 0xd: /* White */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = 0xffffff;
                                w.setPixel( x, height, 0xffffff );
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = 0xffffff;
                            w.setPixel( x, height, 0xffffff );
                            count--;
                            x++;
                        }
                        break;

                    case 0xe: /* Black */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // pixel[line + x] = 0x00;
                                w.setPixel( x, height, 0x00 );
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // pixel[line + x] = 0x00;
                            w.setPixel( x, height, 0x00 );
                            count--;
                            x++;
                        }

                        break;
                    default:
                        throw new RdesktopException(
                                "Unimplemented decompress opcode " + opcode );// ;
                }
            }
        }

        /*
         * if(Options.server_bpp == 16){ for(int i = 0; i < pixel.length; i++)
         * pixel[i] = Bitmap.convert16to24(pixel[i]); }
         */

        return w;
    }

    /**
     * Decompress bitmap data from packet and store in array of integers
     * 
     * @param width
     *            Width of bitmap
     * @param height
     *            Height of bitmap
     * @param size
     *            Size of compressed data in bytes
     * @param data
     *            Packet containing bitmap data
     * @param Bpp
     *            Bytes per-pixel for bitmap
     * @return Integer array of pixels containing decompressed bitmap data
     * @throws RdesktopException
     */
    public static int[] decompressInt( int width, int height, int size,
            RdpPacket_Localised data, int Bpp, int bpp )
            throws RdesktopException
    {

        byte[] compressed_pixel = new byte[size];
        data.copyToByteArray( compressed_pixel, 0, data.getPosition(), size );
        data.incrementPosition( size );

        int previous = -1, line = 0;
        int input = 0, output = 0, end = size;
        int opcode = 0, count = 0, offset = 0, x = width;
        int lastopcode = -1, fom_mask = 0;
        int code = 0, color1 = 0, color2 = 0;
        byte mixmask = 0;
        int mask = 0;
        int mix = 0xffffffff;

        final int blackPixel = 0xFF000000;
        
        boolean insertmix = false, bicolor = false, isfillormix = false;

        int[] pixel = new int[width * height];
        
        while ( input < end )
        {
            fom_mask = 0;
            code = (compressed_pixel[input++] & 0x000000ff);
            opcode = code >> 4;

            /* Handle different opcode forms */
            switch ( opcode )
            {
                case 0xc:
                case 0xd:
                case 0xe:
                    opcode -= 6;
                    count = code & 0xf;
                    offset = 16;
                    break;

                case 0xf:
                    opcode = code & 0xf;
                    if ( opcode < 9 )
                    {
                        count = (compressed_pixel[input++] & 0xff);
                        count |= ((compressed_pixel[input++] & 0xff) << 8);
                    }
                    else
                    {
                        count = (opcode < 0xb) ? 8 : 1;
                    }
                    offset = 0;
                    break;

                default:
                    opcode >>= 1;
                    count = code & 0x1f;
                    offset = 32;
                    break;
            }

            /* Handle strange cases for counts */
            if ( offset != 0 )
            {
                isfillormix = ((opcode == 2) || (opcode == 7));

                if ( count == 0 )
                {
                    if ( isfillormix )
                        count = (compressed_pixel[input++] & 0x000000ff) + 1;
                    else
                        count = (compressed_pixel[input++] & 0x000000ff)
                                + offset;
                }
                else if ( isfillormix )
                {
                    count <<= 3;
                }
            }

            switch ( opcode )
            {
                case 0: /* Fill */
                    if ( (lastopcode == opcode)
                            && !((x == width) && (previous == -1)) )
                        insertmix = true;
                    break;
                case 8: /* Bicolor */
                    color1 = cvalx( compressed_pixel, input, bpp ) | blackPixel;
                    // (compressed_pixel[input++]&0x000000ff);
                    input += Bpp;
                case 3: /* Color */
                    color2 = cvalx( compressed_pixel, input, bpp ) | blackPixel;
                    // color2 = (compressed_pixel[input++]&0x000000ff);
                    input += Bpp;
                    break;
                case 6: /* SetMix/Mix */
                case 7: /* SetMix/FillOrMix */
                    // mix = compressed_pixel[input++];
                    mix = cvalx( compressed_pixel, input, bpp ) | blackPixel;
                    input += Bpp;
                    opcode -= 5;
                    break;
                case 9: /* FillOrMix_1 */
                    mask = 0x03;
                    opcode = 0x02;
                    fom_mask = 3;
                    break;
                case 0x0a: /* FillOrMix_2 */
                    mask = 0x05;
                    opcode = 0x02;
                    fom_mask = 5;
                    break;

            }

            lastopcode = opcode;
            mixmask = 0;

            /* Output body */
            while ( count > 0 )
            {
                if ( x >= width )
                {
                    if ( height <= 0 )
                    {
                        throw new RdesktopException(
                                "Decompressing bitmap failed! Height = "
                                        + height );
                    }

                    x = 0;
                    height--;

                    previous = line;
                    line = output + height * width;
                }

                switch ( opcode )
                {
                    case 0: /* Fill */
                        if ( insertmix )
                        {
                            if ( previous == -1 )
                            {
                                pixel[line + x] = mix;
                            }
                            else
                            {
                                pixel[line + x] = (pixel[previous + x] ^ mix) | blackPixel;
                            }

                            insertmix = false;
                            count--;
                            x++;
                        }

                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, 0);
                                    pixel[line + x] = blackPixel;
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, 0);
                                pixel[line + x] = blackPixel;
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, pixel[previous +
                                    // x]);
                                    pixel[line + x] = pixel[previous + x];
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, pixel[previous + x]);
                                pixel[line + x] = pixel[previous + x];
                                count--;
                                x++;
                            }
                        }
                        break;

                    case 1: /* Mix */
                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, mix, Bpp);
                                    pixel[line + x] = mix;
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, mix, Bpp);
                                pixel[line + x] = mix;
                                count--;
                                x++;
                            }
                        }
                        else
                        {

                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, pixel[previous + x]
                                    // ^ mix);
                                    pixel[line + x] = (pixel[previous + x] ^ mix) | blackPixel;
                                    // setli(pixel, line, x, 0, Bpp);
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, pixel[previous + x] ^
                                // mix);
                                pixel[line + x] = (pixel[previous + x] ^ mix) | blackPixel;
                                // setli(pixel, line, x, 0, Bpp);
                                count--;
                                x++;
                            }

                        }
                        break;
                    case 2: /* Fill or Mix */
                        if ( previous == -1 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                        // setli(pixel, line, x, mix, Bpp);
                                        pixel[line + x] = ((byte) mix) | blackPixel;
                                    else
                                        // setli(pixel, line, x, 0, Bpp);
                                        pixel[line + x] = blackPixel;
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                    // setli(pixel, line, x, mix, Bpp);
                                    pixel[line + x] = mix;
                                else
                                    // setli(pixel, line, x, 0, Bpp);
                                    pixel[line + x] = blackPixel;
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                        // setli(pixel, line, x,
                                        // getli(pixel,previous, x, Bpp) ^ mix,
                                        // Bpp);
                                        pixel[line + x] = (pixel[previous + x] ^ mix) | blackPixel;
                                    else
                                        // setli(pixel, line, x, getli(pixel,
                                        // previous, x, Bpp), Bpp);
                                        pixel[line + x] = pixel[previous + x];
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                    // setli(pixel, line, x, getli(pixel,
                                    // previous,
                                    // x, Bpp) ^ mix, Bpp);
                                    pixel[line + x] = (pixel[previous + x] ^ mix) | blackPixel;
                                else
                                    // setli(pixel, line, x, getli(pixel,
                                    // previous,
                                    // x, Bpp), Bpp);
                                    pixel[line + x] = pixel[previous + x];
                                count--;
                                x++;
                            }

                        }
                        break;

                    case 3: /* Color */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, color2, Bpp);
                                pixel[line + x] = color2;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, color2, Bpp);
                            pixel[line + x] = color2;
                            count--;
                            x++;
                        }

                        break;

                    case 4: /* Copy */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, cvalx(compressed_pixel,
                                // input, Bpp), Bpp);
                                // pixel[line+x] = compressed_pixel[input++];
                                pixel[line + x] = cvalx( compressed_pixel,
                                        input, bpp ) | blackPixel;
                                input += Bpp;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, cvalx(compressed_pixel,
                            // input,Bpp), Bpp);
                            pixel[line + x] = cvalx( compressed_pixel, input,
                                    bpp ) | blackPixel;
                            input += Bpp;
                            // pixel[line+x] = compressed_pixel[input++];
                            count--;
                            x++;
                        }
                        break;

                    case 8: /* Bicolor */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                if ( bicolor )
                                {
                                    // setli(pixel, line, x, color2, Bpp);
                                    pixel[line + x] = color2;
                                    bicolor = false;
                                }
                                else
                                {
                                    // setli(pixel, line, x, color1, Bpp);
                                    pixel[line + x] = color1;
                                    bicolor = true;
                                    count++;
                                }
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            if ( bicolor )
                            {
                                // setli(pixel, line, x, color2, Bpp);
                                pixel[line + x] = color2;
                                bicolor = false;
                            }
                            else
                            {
                                // setli(pixel, line, x, color1, Bpp);
                                pixel[line + x] = color1;
                                bicolor = true;
                                count++;
                            }
                            count--;
                            x++;
                        }

                        break;

                    case 0xd: /* White */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, 0xffffff, Bpp);
                                pixel[line + x] = 0xffffffff;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, 0xffffff, Bpp);
                            pixel[line + x] = 0xffffffff;
                            count--;
                            x++;
                        }
                        break;

                    case 0xe: /* Black */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, 0x00, Bpp);
                                pixel[line + x] = blackPixel;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, 0x00, Bpp);
                            pixel[line + x] = blackPixel;
                            count--;
                            x++;
                        }

                        break;
                    default:
                        throw new RdesktopException(
                                "Unimplemented decompress opcode " + opcode );// ;
                }
            }
        }

        return pixel;
    }

    /**
     * Decompress bitmap data from packet and store in array of bytes
     * 
     * @param width
     *            Width of bitmap
     * @param height
     *            Height of bitmap
     * @param size
     *            Size of compressed data in bytes
     * @param data
     *            Packet containing bitmap data
     * @return Byte array of pixels containing decompressed bitmap data
     * @throws RdesktopException
     */
    public static byte[] decompress( int width, int height, int size,
            RdpPacket_Localised data ) throws RdesktopException
    {

        byte[] compressed_pixel = new byte[size];
        data.copyToByteArray( compressed_pixel, 0, data.getPosition(), size );
        data.incrementPosition( size );

        int previous = 0, line = 0;
        int input = 0, output = 0, end = size;
        int opcode = 0, count = 0, offset = 0, x = width;
        int lastopcode = -1, fom_mask = 0;
        int code = 0, color1 = 0, color2 = 0;
        byte mixmask = 0;
        int mask = 0;
        int mix = 0xffffffff;

        boolean insertmix = false, bicolor = false, isfillormix = false;

        byte[] pixel = new byte[width * height];
        while ( input < end )
        {
            fom_mask = 0;
            code = (compressed_pixel[input++] & 0x000000ff);
            opcode = code >> 4;

            /* Handle different opcode forms */
            switch ( opcode )
            {
                case 0xc:
                case 0xd:
                case 0xe:
                    opcode -= 6;
                    count = code & 0xf;
                    offset = 16;
                    break;

                case 0xf:
                    opcode = code & 0xf;
                    if ( opcode < 9 )
                    {
                        count = (compressed_pixel[input++] & 0xff);
                        count |= ((compressed_pixel[input++] & 0xff) << 8);
                    }
                    else
                    {
                        count = (opcode < 0xb) ? 8 : 1;
                    }
                    offset = 0;
                    break;

                default:
                    opcode >>= 1;
                    count = code & 0x1f;
                    offset = 32;
                    break;
            }

            /* Handle strange cases for counts */
            if ( offset != 0 )
            {
                isfillormix = ((opcode == 2) || (opcode == 7));

                if ( count == 0 )
                {
                    if ( isfillormix )
                        count = (compressed_pixel[input++] & 0x000000ff) + 1;
                    else
                        count = (compressed_pixel[input++] & 0x000000ff)
                                + offset;
                }
                else if ( isfillormix )
                {
                    count <<= 3;
                }
            }

            switch ( opcode )
            {
                case 0: /* Fill */
                    if ( (lastopcode == opcode)
                            && !((x == width) && (previous == 0)) )
                        insertmix = true;
                    break;
                case 8: /* Bicolor */
                    // color1 = cvalx(compressed_pixel, input, Bpp); //
                    // (compressed_pixel[input++]&0x000000ff);
                    color1 = (compressed_pixel[input++] & 0x000000ff);
                    // input += Bpp;
                case 3: /* Color */
                    // color2 = cvalx(compressed_pixel, input, Bpp);
                    color2 = (compressed_pixel[input++] & 0x000000ff);
                    // input += Bpp;
                    break;
                case 6: /* SetMix/Mix */
                case 7: /* SetMix/FillOrMix */
                    mix = compressed_pixel[input++];
                    // mix = cvalx(compressed_pixel, input, Bpp);
                    // input += Bpp;
                    opcode -= 5;
                    break;
                case 9: /* FillOrMix_1 */
                    mask = 0x03;
                    opcode = 0x02;
                    fom_mask = 3;
                    break;
                case 0x0a: /* FillOrMix_2 */
                    mask = 0x05;
                    opcode = 0x02;
                    fom_mask = 5;
                    break;

            }

            lastopcode = opcode;
            mixmask = 0;

            /* Output body */
            while ( count > 0 )
            {
                if ( x >= width )
                {
                    if ( height <= 0 )
                        throw new RdesktopException(
                                "Decompressing bitmap failed! Height = "
                                        + height );

                    x = 0;
                    height--;

                    previous = line;
                    line = output + height * width;
                }

                switch ( opcode )
                {
                    case 0: /* Fill */
                        if ( insertmix )
                        {
                            if ( previous == 0 )
                            {
                                pixel[line + x] = (byte) mix;
                            }
                            else
                            {
                                // setli(pixel, line, x, getli(pixel, previous,
                                // x, Bpp) ^ mix, Bpp);
                                pixel[line + x] = (byte) (pixel[previous + x] ^ (byte) mix);
                            }

                            insertmix = false;
                            count--;
                            x++;
                        }

                        if ( previous == 0 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, 0, Bpp);
                                    pixel[line + x] = 0;
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, 0, Bpp);
                                pixel[line + x] = 0;
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, getli(pixel,
                                    // previous, x, Bpp), Bpp);
                                    pixel[line + x] = pixel[previous + x];
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, getli(pixel, previous,
                                // x, Bpp), Bpp);
                                pixel[line + x] = pixel[previous + x];
                                count--;
                                x++;
                            }
                        }
                        break;

                    case 1: /* Mix */
                        if ( previous == 0 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    // setli(pixel, line, x, mix, Bpp);
                                    pixel[line + x] = (byte) mix;
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                // setli(pixel, line, x, mix, Bpp);
                                pixel[line + x] = (byte) mix;
                                count--;
                                x++;
                            }
                        }
                        else
                        {

                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    setli( pixel, line, x,
                                            getli( pixel, previous, x, 1 )
                                                    ^ mix, 1 );
                                    // setli(pixel, line, x, 0, Bpp);
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                setli( pixel, line, x,
                                        getli( pixel, previous, x, 1 ) ^ mix, 1 );
                                // setli(pixel, line, x, 0, Bpp);
                                count--;
                                x++;
                            }

                        }
                        break;
                    case 2: /* Fill or Mix */
                        if ( previous == 0 )
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                        // setli(pixel, line, x, mix, Bpp);
                                        pixel[line + x] = (byte) mix;
                                    else
                                        // setli(pixel, line, x, 0, Bpp);
                                        pixel[line + x] = 0;
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                    // setli(pixel, line, x, mix, Bpp);
                                    pixel[line + x] = (byte) mix;
                                else
                                    // setli(pixel, line, x, 0, Bpp);
                                    pixel[line + x] = 0;
                                count--;
                                x++;
                            }
                        }
                        else
                        {
                            while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                            {
                                for( int i = 0; i < 8; i++ )
                                {
                                    mixmask <<= 1;
                                    if ( mixmask == 0 )
                                    {
                                        mask = (fom_mask != 0) ? (byte) fom_mask
                                                : compressed_pixel[input++];
                                        mixmask = 1;
                                    }
                                    if ( (mask & mixmask) != 0 )
                                        // setli(pixel, line, x, getli(pixel,
                                        // previous, x, Bpp)
                                        // ^ mix, Bpp);
                                        pixel[line + x] = (byte) (pixel[previous
                                                + x] ^ (byte) mix);
                                    else
                                        // setli(pixel, line, x, getli(pixel,
                                        // previous, x, Bpp), Bpp);
                                        pixel[line + x] = pixel[previous + x];
                                    count--;
                                    x++;
                                }
                            }
                            while ( (count > 0) && (x < width) )
                            {
                                mixmask <<= 1;
                                if ( mixmask == 0 )
                                {
                                    mask = (fom_mask != 0) ? (byte) fom_mask
                                            : compressed_pixel[input++];
                                    mixmask = 1;
                                }
                                if ( (mask & mixmask) != 0 )
                                    // setli(pixel, line, x, getli(pixel,
                                    // previous, x,
                                    // Bpp)
                                    // ^ mix, Bpp);
                                    pixel[line + x] = (byte) (pixel[previous
                                            + x] ^ (byte) mix);
                                else
                                    // setli(pixel, line, x, getli(pixel,
                                    // previous, x,
                                    // Bpp), Bpp);
                                    pixel[line + x] = pixel[previous + x];
                                count--;
                                x++;
                            }

                        }
                        break;

                    case 3: /* Color */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, color2, Bpp);
                                pixel[line + x] = (byte) color2;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, color2, Bpp);
                            pixel[line + x] = (byte) color2;
                            count--;
                            x++;
                        }

                        break;

                    case 4: /* Copy */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, cvalx(compressed_pixel,
                                // input, Bpp), Bpp);
                                pixel[line + x] = compressed_pixel[input++];
                                // input += Bpp;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, cvalx(compressed_pixel,
                            // input,
                            // Bpp), Bpp);
                            // input += Bpp;
                            pixel[line + x] = compressed_pixel[input++];
                            count--;
                            x++;
                        }
                        break;

                    case 8: /* Bicolor */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                if ( bicolor )
                                {
                                    // setli(pixel, line, x, color2, Bpp);
                                    pixel[line + x] = (byte) color2;
                                    bicolor = false;
                                }
                                else
                                {
                                    // setli(pixel, line, x, color1, Bpp);
                                    pixel[line + x] = (byte) color1;
                                    bicolor = true;
                                    count++;
                                }
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            if ( bicolor )
                            {
                                // setli(pixel, line, x, color2, Bpp);
                                pixel[line + x] = (byte) color2;
                                bicolor = false;
                            }
                            else
                            {
                                // setli(pixel, line, x, color1, Bpp);
                                pixel[line + x] = (byte) color1;
                                bicolor = true;
                                count++;
                            }
                            count--;
                            x++;
                        }

                        break;

                    case 0xd: /* White */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, 0xffffffff, Bpp);
                                pixel[line + x] = (byte) 0xff;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, 0xffffffff, Bpp);
                            pixel[line + x] = (byte) 0xff;
                            count--;
                            x++;
                        }
                        break;

                    case 0xe: /* Black */
                        while ( ((count & ~0x7) != 0) && ((x + 8) < width) )
                        {
                            for( int i = 0; i < 8; i++ )
                            {
                                // setli(pixel, line, x, 0, Bpp);
                                pixel[line + x] = (byte) 0x00;
                                count--;
                                x++;
                            }
                        }
                        while ( (count > 0) && (x < width) )
                        {
                            // setli(pixel, line, x, 0, Bpp);
                            pixel[line + x] = (byte) 0x00;
                            count--;
                            x++;
                        }

                        break;
                    default:
                        throw new RdesktopException(
                                "Unimplemented decompress opcode " + opcode );// ;
                }
            }
        }

        bmpCount++;

        return pixel;
    }

    private static int RLEDeltaDecode( int value, int prevValue )
    {
        if ( ((value >> 1) << 1) != value )
        {
            value = 255 - (value >> 1);
        }
        else
        {
            value >>= 1;
        }
        value = (value + prevValue) & 0xFF;
        return value;
    }

    /**
     * RDP 6.0 BMP RLE decompress function
     * 
     * @param src
     *            - archive byte array
     * @param offset
     *            - offset in source array
     * @param size
     *            - size of source array
     * @param width
     *            - width of image (scanline)
     * @param height
     *            - height of image
     * @param dst
     *            - result byte array
     * @return Next a RLE plane start position or an end of source array
     *         position
     * @throws RdesktopException
     */
    private static int RLEDecode( byte[] src, int offset, int size, int width,
            int height, int[] dst ) throws RdesktopException
    {
        int resCounter = 0;
        int readBytes = offset;
        boolean firstLine = true;
        int lastByte = 0;
        int maxSize = height * width;
        while ( readBytes < size && resCounter < maxSize )
        {
            int controlByte = src[readBytes++];

            if ( controlByte == 0 )
            {
                logger.error( "Incorrect control byte value in RLE archive" );
                throw new RdesktopException( "Error parsing server bitmap" );
            }

            int nRunLength = controlByte & 0x0F;

            int cRawBytes = (controlByte >> 4) & 0x0F;

            int copiesCount = cRawBytes > 0 ? 1 : 0;
            int arrSize = 1;
            if ( nRunLength == 1 )
            {
                copiesCount = 16 + cRawBytes;
                nRunLength = 0;
                cRawBytes = 0;
            }
            else if ( nRunLength == 2 )
            {
                copiesCount = 32 + cRawBytes;
                nRunLength = 0;
                cRawBytes = 0;
            }
            else
            {
                copiesCount += nRunLength;
                arrSize = cRawBytes > 0 ? cRawBytes : 1;
            }

            int[] arr = new int[arrSize];

            if ( cRawBytes == 0 )
            {
                arr[0] = lastByte;
            }
            else
            {
                for( int j = 0; j < cRawBytes; ++j )
                {
                    arr[j] = src[readBytes++] & 0xFF;
                }
                lastByte = arr[arr.length - 1];
            }

            boolean firstRun = true;
            for( int j = 0; j < copiesCount; ++j )
            {
                if ( firstRun )
                {
                    for( int k = 0; k < arr.length; ++k )
                    {
                        int val = arr[k];
                        if ( !firstLine )
                        {
                            val = RLEDeltaDecode( val, dst[resCounter - width] );
                        }
                        dst[resCounter++] = val;
                    }
                    firstRun = false;
                }
                else
                {
                    int val = lastByte;
                    if ( !firstLine )
                    {
                        val = RLEDeltaDecode( val, dst[resCounter - width] );
                    }
                    dst[resCounter++] = val;
                }
            }

            if ( resCounter % width == 0 )
            {
                lastByte = 0;
                if ( resCounter >= width )
                {
                    firstLine = false;
                }
            }
        }
        return readBytes;
    }

    /**
     * RDP 6.0 BMP decompress function
     * 
     * @param width
     *            - width of image
     * @param height
     *            - height of image
     * @param size
     *            - size of packed data
     * @param uncompressedSize
     *            - size of unpacked data (can be 0) (unused)
     * @param data
     *            - input buffer
     * @return Returns an array of integer pixels in the default ARGB color
     *         model
     * @throws RdesktopException
     */
    public static int[] RLEDecompress( int width, int height, int size,
            int uncompressedSize, RdpPacket_Localised data )
            throws RdesktopException
    {
        int[] pixel = new int[width * height];

        byte[] compressed = new byte[size];
        data.copyToByteArray( compressed, 0, data.getPosition(), size );
        data.incrementPosition( size );

        int FormatHeader = compressed[0];
        int CLL = FormatHeader & 0x07;
        int CS = ((FormatHeader >> 3) & 0x01);
        int RLE = ((FormatHeader >> 4) & 0x01);
        int NA = ((FormatHeader >> 5) & 0x01);

        if ( CLL != 0 || CS != 0 || NA != 0 )
        {
            logger.error( "Incompatible data: CLL=" + CLL + " CS=" + CS
                    + " NA=" + NA );
            throw new RdesktopException(
                    "Incompatible data in server bitmap message" );
        }

        if ( RLE == 0 )
        {
            int lineCount = (height - 1) * width;
            int counter = 0;
            for( int i = 1; i < compressed.length; )
            {
                byte A = compressed[i++];
                byte R = compressed[i++];
                byte G = compressed[i++];
                byte B = compressed[i++];

                pixel[lineCount + counter] = (A << 24) | (R << 16) | (G << 8)
                        | B;
                ++counter;
                if ( counter == width )
                {
                    counter = 0;
                    lineCount -= width;
                }
            }
        }
        else
        {
            int planeSize = width * height;
            int[] A = new int[planeSize];
            int[] R = new int[planeSize];
            int[] G = new int[planeSize];
            int[] B = new int[planeSize];

            int pos = RLEDecode( compressed, 1, size, width, height, A );
            pos = RLEDecode( compressed, pos, size, width, height, R );
            pos = RLEDecode( compressed, pos, size, width, height, G );
            pos = RLEDecode( compressed, pos, size, width, height, B );

            int lineCount = (height - 1) * width;
            int counter = 0;
            for( int i = 0; i < planeSize; ++i )
            {
                pixel[lineCount + counter] = ((A[i] & 0xFF) << 24)
                        | ((R[i] & 0xFF) << 16) | ((G[i] & 0xFF) << 8)
                        | (B[i] & 0xFF);
                ++counter;
                if ( counter == width )
                {
                    counter = 0;
                    lineCount -= width;
                }
            }
        }
        return pixel;
    }
}
