/* ISO.java
 * Component: ProperJavaRDP
 * 
 * Revision: $Revision: 1.7 $
 * Author: $Author: telliott $
 * Date: $Date: 2005/09/27 14:15:39 $
 *
 * Copyright (c) 2005 Propero Limited
 *
 * Purpose: Java 1.4 specific extension of ISO class
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * (See gpl.txt for details of the GNU General Public License.)
 * 
 */
// Created on 05-Aug-2003

package net.propero.rdp;

import java.io.*;
import java.net.*;
import java.nio.channels.SocketChannel;

import java.security.cert.*;
import java.util.*;

import javax.net.ssl.*;

import net.propero.rdp.crypto.CryptoException;
import android.app.Activity;

import com.provisionnetworks.common.SSLUtil;
import com.provisionnetworks.common.UICommon;
import com.provisionnetworks.rdp5.PNOptions;

public class ISO_Localised extends ISO 
{
    public ISO_Localised( Activity activity, Options options, PNOptions pnOptions )
    {
        super( activity, options, pnOptions );
    }

    protected void doSocketConnect( InetAddress host, int port) throws IOException{
        int timeout_ms = 3000; // timeout in milliseconds

        boolean useProxy = pnOptions.rdpSettings.useProxy;
        String proxyServer = pnOptions.rdpSettings.proxyServer;
        String proxyHost = null;
        int proxyPort = 0;

        // Determine if a proxy server is being used.
        if (useProxy)
        {
            if ((proxyServer == null) || (proxyServer.length() == 0))
            {
                logger.info("Detecting system-defined proxy server...");

                System.setProperty("java.net.useSystemProxies", "true");
                try
                {
                    List<Proxy> proxyList = ProxySelector.getDefault().select(new URI("http://" + host.getHostName()));
                    if ((proxyList != null) && (proxyList.size() > 0))
                    {
                        Proxy proxy = proxyList.get(0);
                        if (proxy.type() == Proxy.Type.HTTP)
                        {
                            InetSocketAddress addr = (InetSocketAddress)proxy.address();
                            if (addr != null)
                            {
                                proxyHost = addr.getHostName();
                                proxyPort = addr.getPort();
                            }
                        }
                    }
                } 
                catch (URISyntaxException e)
                {
                    e.printStackTrace();
                }

                if ((proxyHost == null) || (proxyHost.length() == 0))
                {
                    if ((proxyServer != null) && (proxyServer.length() > 0))
                    {
                        logger.info("Using Internet Explorer proxy server (" + proxyServer + ")...");
                        int pos = proxyServer.indexOf(';');
                        if (pos > 0)
                        {
                            proxyServer = proxyServer.substring(0, pos);
                        }
                        pos = proxyServer.indexOf(':');
                        if (pos > 0)
                        {
                            proxyHost = proxyServer.substring(0, pos);
                            proxyPort = Integer.parseInt(proxyServer.substring(pos + 1));
                        }
                        else
                        {
                            proxyHost = proxyServer;
                            proxyPort = 80;
                        }
                    }
                }
            }
            else
            {
                logger.info("Using user-defined proxy server (" + proxyServer + ")...");
                int pos = proxyServer.indexOf(':');
                if (pos > 0)
                {
                    proxyHost = proxyServer.substring(0, pos);
                    proxyPort = Integer.parseInt(proxyServer.substring(pos + 1));
                }
                else
                {
                    proxyHost = proxyServer;
                    proxyPort = 80;
                }
            }
        }

        // Connect to the proxy server.
        if ((proxyHost != null) && (proxyHost.length() > 0))
        {
            logger.info("Connecting to proxy server " + proxyHost + ":" + proxyPort);
            
            rdpsock = new Socket();
            
            rdpsock.connect(new InetSocketAddress(proxyHost, proxyPort), timeout_ms);
            doProxyHandshake(rdpsock, host, port);
        }

        String sslGateway = pnOptions.sslGateway;
        if ((sslGateway != null) && (sslGateway.length() > 0))
        {
        	try {
        		logger.info("SSL gateway certificate checking. Host: " + sslGateway);
        		//Check server certificate.
        		//If it not trusted show user warning message  
        		if ( !serverCertificateVerifier(host.getHostName(), port) ) {
        			logger.warn("The user does not trust the SSL certificate. The application will be stopped.");
        			//The server certificate checking is already disabled.
        	    	//So if user does not trust certificate then application is stopped 
        	    	//System.exit(0);
        			return;
        		}
			} catch (Exception e) {
				logger.warn("Method doSocketConnect(...): Error in SSL certificates checking or disabling: " + e.getMessage());
			}
			
			//Get socket factory from HttpsURLConnection 
			//for disabling server certificate checking
            SSLSocketFactory factory = HttpsURLConnection.getDefaultSSLSocketFactory();

            // Construct a connection string.
            String targetHost = pnOptions.targetHost;
            int targetPort = pnOptions.targetPort;
            String username = (options.username != null) && (options.username.length() > 0) ? options.username : "unknown";
            String connectString = "CONNECT: " + targetHost + ":" + targetPort + "\r\nCLIENT: " + username + "\r\n\r\n";
        	
            // Connect to the SSL gateway.
            logger.info("Connecting to SSL gateway " + host + ":" + port + "(target=" + targetHost + ":" + targetPort + ")");
            if (rdpsock != null)
            {
                // Tunnel through proxy server.
                SSLSocket sslsock = (SSLSocket)factory.createSocket(rdpsock, host.getHostAddress(), port, true);
                sslsock.startHandshake();
                rdpsock = sslsock;
            }
            else
            {
                // Connect directly to SSL gateway.
                rdpsock = factory.createSocket();
                rdpsock.connect(new InetSocketAddress(host,port), timeout_ms);
            }

            // Connect to target host.
            OutputStream os = rdpsock.getOutputStream();
            os.write(connectString.getBytes());
            os.flush();
        }

        // Connect directly to the Terminal Server.
        if (rdpsock == null)
        {
            rdpsock = new Socket();
            rdpsock.connect(new InetSocketAddress(host,port), timeout_ms);
        }
    }
    
    private boolean serverCertificateVerifier(String host, int port) {
    	if (SSLUtil.isServerCertificateTrusted(host, port)) {
    		SSLUtil.getCertificates(host, port);
    		return true;
    	}
    	
    	//The server certificates can not be got if certification checking is enabled 
    	SSLUtil.disableCertChecking();
    	Certificate[] certs = SSLUtil.getCertificates(host, port);
    	
    	//Only first certificate should be checked. 
    	//Other certificates can be ignored.
    	X509Certificate cert = (X509Certificate)certs[0];
    	
    	boolean res = UICommon.showConfirmDialog( activity, 
    			"The program tries to connect to a server " +
    			"that contains an untrusted certificate:\n" +
    			"    Certificate for: " + cert.getSubjectDN() + "\n" +
    			"    Certificate issued by: " + cert.getIssuerDN() + "\n" +
				"This can damage your computer.\n" + 
				"If you trust the certificate, you can continue working.\n" +
				"Otherwise, we recommend you to close a program.\n" +
				"Do you want to continue?", 
    			"vWorkspace Connector for Android");
    	
    	if ( res ) {
    		return true;
    	}
    	
    	return false;
    }
    
    public void connect(InetAddress host, int port) throws IOException, RdesktopException, OrderException, CryptoException {
        super.connect(host, port);
        //rdpsock.setTcpNoDelay(false);
        //rdpsock.setReceiveBufferSize(65536);
    }

    /*
     * Tell our tunnel where we want to CONNECT, and look for the
     * right reply.  Throw IOException if anything goes wrong.
     */
    private void doProxyHandshake(Socket tunnel, InetAddress host, int port) throws IOException
    {
	OutputStream out = tunnel.getOutputStream();
	String msg = "CONNECT " + host.getHostAddress() + ":" + port + " HTTP/1.0\r\n" +
		     "User-Agent: Quest/1.0\r\n\r\n";
	byte b[];
	try {
	    /*
	     * We really do want ASCII7 -- the http protocol doesn't change
	     * with locale.
	     */
	    b = msg.getBytes("ASCII7");
	} catch (UnsupportedEncodingException ignored) {
	    /*
	     * If ASCII7 isn't there, something serious is wrong, but
	     * Paranoia Is Good (tm)
	     */
	    b = msg.getBytes();
	}
	out.write(b);
	out.flush();

	/*
	 * We need to store the reply so we can create a detailed
	 * error message to the user.
	 */
	byte		reply[] = new byte[200];
	int		replyLen = 0;
	int		newlinesSeen = 0;
	boolean		headerDone = false;	/* Done on first newline */

	InputStream	in = tunnel.getInputStream();

	while (newlinesSeen < 2) {
	    int i = in.read();
	    if (i < 0) {
	        String errMsg = "Unexpected EOF from proxy";
	        logger.info(errMsg);
		throw new IOException(errMsg);
	    }
	    if (i == '\n') {
		headerDone = true;
		++newlinesSeen;
	    } else if (i != '\r') {
		newlinesSeen = 0;
		if (!headerDone && replyLen < reply.length) {
		    reply[replyLen++] = (byte) i;
		}
	    }
	}

	/*
	 * Converting the byte array to a string is slightly wasteful
	 * in the case where the connection was successful, but it's
	 * insignificant compared to the network overhead.
	 */
	String replyStr;
	try {
	    replyStr = new String(reply, 0, replyLen, "ASCII7");
	} catch (UnsupportedEncodingException ignored) {
	    replyStr = new String(reply, 0, replyLen);
	}

	/* We asked for HTTP/1.0, so we should get that back */
	if (!replyStr.startsWith("HTTP/") || (replyStr.indexOf("200") < 0)) {
	    String errMsg = "Unable to tunnel to " + host + ":" + port + ".  Proxy returns \"" + replyStr + "\"";
	    logger.info(errMsg);
	    throw new IOException(errMsg);
	}

	/* tunneling Handshake was successful! */
    }
}
