/* RasterOp.java
 * Component: ProperJavaRDP
 * 
 * Revision: $Revision: 1.7 $
 * Author: $Author: telliott $
 * Date: $Date: 2005/09/27 14:15:39 $
 *
 * Copyright (c) 2005 Propero Limited
 *
 * Purpose: Set of operations used in displaying raster graphics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * (See gpl.txt for details of the GNU General Public License.)
 * 
 */
// Created on 01-Jul-2003

package net.propero.rdp;


import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Rect;

import org.apache.log4j.Logger;

public class RasterOp 
{
    
	Logger logger = Logger.getLogger(RasterOp.class);

	private Options options = null;
	
	public RasterOp( Options options )
	{
	    this.options = options;
	}
	
	private void ropInvert(Bitmap biDst, int[] dest, int width, int x, int y, int cx, int cy,
			int Bpp) 
	{
		int pdest = (y * width + x);
		
		if ( biDst != null )
		{
		    int mask = 0xFFFFFFFF;
    		for ( int i = y; i < y + cy; ++i ) 
    		{
    			for ( int j = x; j < x + cx; ++j ) 
    			{
                    int c = biDst.getPixel( j, i );
                    biDst.setPixel( j, i, (~c & mask) | 0xFF000000 );
    				++pdest;
    			}
    			pdest += (width - cx);
    		}
		}
		else
		{
		    int mask = options.bpp_mask;
		    for ( int i = y; i < y + cy; ++i ) 
            {
                for ( int j = x; j < x + cx; ++j ) 
                {
                    dest[pdest] = (~dest[pdest]) & mask;
                    ++pdest;
                }
                pdest += (width - cx);
            }
		}
	}

	private void ropClear(Bitmap biDst, int width, int x, int y, int cx, int cy,
			int Bpp) 
	{
	    Canvas canvas = new Canvas( biDst );
	    Paint paint = new Paint();
	    paint.setColor( 0xFF000000 );
	    paint.setStyle(android.graphics.Paint.Style.FILL_AND_STROKE);
	    canvas.drawRect(new android.graphics.Rect(x, y, x + cx, y + cy), paint);
	    
//	    for ( int i = x; i < x+cx; i++ )
//	    {
//            for( int j = y; j < y+cy; j++ )
//            {
//                biDst.setPixel( i, j, 0xFF000000);
//            }
//	    }
	}

	private void ropSet(Bitmap biDst, int width, int x, int y, int cx, int cy,
			int Bpp) 
	{
        int mask = options.bpp_mask | 0xFF000000;
        
        Canvas canvas = new Canvas( biDst );
        Paint paint = new Paint();
        paint.setColor( mask );
        paint.setStyle(android.graphics.Paint.Style.FILL_AND_STROKE);
        canvas.drawRect(new android.graphics.Rect(x, y, x + cx, y + cy), paint);
        
//        for(int i = x; i < x+cx; i++)
//        {
//            for(int j = y; j < y+cy; j++)
//            {
//                biDst.setPixel(i,j,mask);
//            }
//        }
    }

	private void ropCopy(Bitmap biDst, int dstwidth, int x, int y, int cx, int cy,
			int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		Canvas canvas = new Canvas( biDst );
        
		if (src == null) 
		{
		    Bitmap chunk = Bitmap.createBitmap( biDst, srcx, srcy, cx, cy );
            Rect dstRect = new Rect(x, y, x + cx, y + cy);
            canvas.drawBitmap(chunk, dstRect.left, dstRect.top, null);
		} 
		else 
		{
            canvas.drawBitmap( src, srcx + srcy * srcwidth, srcwidth, x, y, cx, cy, false, null );
		}
	}

    /**
     * Perform an operation on a rectangular area of a Bitmap, using an integer array of colour values as
     * source if necessary
     * @param opcode Code defining operation to perform
     * @param biDst Destination image for operation
     * @param dstwidth Width of destination image
     * @param x X-offset of destination area within destination image
     * @param y Y-offset of destination area within destination image
     * @param cx Width of destination area
     * @param cy Height of destination area
     * @param src Source data, represented as an array of integer pixel values
     * @param srcwidth Width of source data
     * @param srcx X-offset of source area within source data
     * @param srcy Y-offset of source area within source data
     */
	public void do_array(int opcode, Bitmap biDst, int dstwidth, int x, int y,
			int cx, int cy, int[] src, int srcwidth, int srcx, int srcy) 
	{
	    logger.debug( "[Orders][ROP] do_array: opcode=" + opcode );
		int Bpp = options.Bpp;
        //int[] dst = null;
		// System.out.println("do_array: opcode = 0x" +
		// Integer.toHexString(opcode) );
		switch (opcode) 
		{
		case 0x0:
			ropClear(biDst, dstwidth, x, y, cx, cy, Bpp);
			break;
		case 0x1:
			ropNor(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0x2:
			ropAndInverted(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx,
					srcy, Bpp);
			break;
		case 0x3: // CopyInverted
			ropInvert(biDst, src, srcwidth, srcx, srcy, cx, cy, Bpp);
			ropCopy(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0x4: // AndReverse
			ropInvert(biDst, null, dstwidth, x, y, cx, cy, Bpp);
			ropAnd(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0x5:
			ropInvert(biDst, null, dstwidth, x, y, cx, cy, Bpp);
			break;
		case 0x6:
			ropXor(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0x7:
			ropNand(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0x8:
			ropAnd(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0x9:
			ropEquiv(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy,
					Bpp);
			break;
		case 0xa: // Noop
			break;
		case 0xb:
			ropOrInverted(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx,
					srcy, Bpp);
			break;
		case 0xc:
			ropCopy(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0xd: // OrReverse
			ropInvert(biDst, null, dstwidth, x, y, cx, cy, Bpp);
			ropOr(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0xe:
			ropOr(biDst, dstwidth, x, y, cx, cy, src, srcwidth, srcx, srcy, Bpp);
			break;
		case 0xf:
			ropSet(biDst, dstwidth, x, y, cx, cy, Bpp);
			break;
		default:
			logger.warn("do_array unsupported opcode: " + opcode);
		// rop_array(opcode,dst,dstwidth,x,y,cx,cy,src,srcwidth,srcx,srcy);
		}
	}
    
    /**
     * Perform an operation on a single pixel in a Bitmap
     * @param opcode Opcode defining operation to perform
     * @param dst Image on which to perform the operation
     * @param x X-coordinate of pixel to modify
     * @param y Y-coordinate of pixel to modify
     * @param color Colour to use in operation (unused for some operations)
     */
    public void do_pixel(int opcode, Bitmap dst, int x, int y, int color) 
    {
        logger.debug( "[Orders][ROP] do_pixel: opcode=" + opcode );
        int mask = options.bpp_mask;
        
        if(dst == null) return;
        
        int c = dst.getPixel(x,y);

        switch (opcode) 
        {
        case 0x0: c = 0x00000000; break;
        case 0x1: c = (~(c | color)) & mask; break;
        case 0x2: c = c & ((~color) & mask); break;
        case 0x3: c = ((~color) & mask); break;
        case 0x4: c = ((~c & color) * mask); break;
        case 0x5: c = ((~c) & mask);  break;
        case 0x6: c = (c ^ ((color) & mask)); break;
        case 0x7: c = ((~c & color) & mask); break;
        case 0x8: c = (c & ( color & mask )); break;
        case 0x9: c = (c ^ ( ~color & mask) ); break;
        case 0xa: /* Noop */ break;
        case 0xb: c = (c | ( ~color & mask )); break;
        case 0xc: c = (color); break;
        case 0xd: c = ((~c | color) & mask); break;
        case 0xe: c = (c | ( color & mask)); break;
        case 0xf: c = (mask); break;
        default:
            logger.warn("do_byte unsupported opcode: " + opcode);
        }
        
        c |= 0xFF000000;
        dst.setPixel( x, y, c );
    }

	private void ropNor(Bitmap biDst, int dstwidth, int x, int y, int cx, int cy,
			int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0x1
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);

        for (int row = y; row < y + cy; row++) 
        {
			for (int col = x; col < x + cx; col++) 
			{
                biDst.setPixel( col, row, ((~(biDst.getPixel(col,row) | src[psrc])) & mask) | 0xFF000000 );
                ++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropAndInverted(Bitmap biDst, int dstwidth, int x, int y, int cx,
			int cy, int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0x2
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		
		for ( int row = y; row < y + cy; ++row ) 
		{
			for (int col = x; col < x + cx; ++col ) 
			{
                int c = biDst.getPixel( col, row );
                biDst.setPixel( col, row, (c & ((~src[psrc]) & mask)) | 0xFF000000 );              
				++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropXor(Bitmap biDst, int dstwidth, int x, int y, int cx, int cy,
			int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0x6
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		for (int row = y; row < y + cy; ++row ) 
		{
			for (int col = x; col < x + cx; ++col )
			{
                int c = biDst.getPixel( col, row);
                biDst.setPixel( col, row, (c ^ ((src[psrc]) & mask)) | 0xFF000000 );
				++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropNand(Bitmap biDst, int dstwidth, int x, int y, int cx, int cy,
			int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0x7
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		for ( int row = y; row < y + cy; ++row ) 
		{
			for ( int col = x; col < x + cx; ++col ) 
			{
			    int c = biDst.getPixel( col, row );
                biDst.setPixel( col, row, ((~(c & src[psrc])) & mask) | 0xFF000000 );
				++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropAnd(Bitmap biDst, int dstwidth, int x, int y, int cx, int cy,
			int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0x8
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		
		for ( int row = y; row < y + cy; ++row ) 
		{
			for ( int col = x; col < x + cx; ++col ) 
			{
                int c = biDst.getPixel( col, row);
                biDst.setPixel( col, row, (c & ((src[psrc]) & mask)) | 0xFF000000 );
				++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropEquiv(Bitmap biDst, int dstwidth, int x, int y, int cx,
			int cy, int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0x9
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		
		for ( int row = y; row < y + cy; ++row ) 
		{
			for ( int col = x; col < x + cx; ++col ) 
			{
                int c = biDst.getPixel( col, row );
                biDst.setPixel( col, row, (c ^ ((~src[psrc]) & mask)) | 0xFF000000);
                ++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropOrInverted(Bitmap biDst, int dstwidth, int x, int y, int cx,
			int cy, int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0xb
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		
		for (int row = y; row < y + cy; ++row)
		{
			for ( int col = x; col < x + cx; ++col ) 
			{
                int c = biDst.getPixel( col, row );
                biDst.setPixel( col, row, c | ((~src[psrc]) & mask));
                ++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}

	private void ropOr(Bitmap biDst, int dstwidth, int x, int y, int cx, int cy,
			int[] src, int srcwidth, int srcx, int srcy, int Bpp) 
	{
		// opcode 0xe
		int mask = options.bpp_mask;
		int psrc = (srcy * srcwidth + srcx);
		
		for ( int row = y; row < y + cy; ++row ) 
		{
			for ( int col = x; col < x + cx; ++col ) 
			{
                int c = biDst.getPixel( col, row );
                biDst.setPixel( col, row, c | (src[psrc] & mask ) );
        		++psrc;
			}
			psrc += (srcwidth - cx);
		}
	}
}
