/* RdesktopCanvas_Localised.java
 * Component: ProperJavaRDP
 * 
 * Revision: $Revision: 1.7 $
 * Author: $Author: telliott $
 * Date: $Date: 2005/09/27 14:15:39 $
 *
 * Copyright (c) 2005 Propero Limited
 *
 * Purpose: Java 1.4 specific extension of RdesktopCanvas class
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * (See gpl.txt for details of the GNU General Public License.)
 * 
 */
package net.propero.rdp;

import quest.androidport.Cursor;
import quest.androidport.ScreenMouse;
import android.app.Activity;
import android.graphics.Matrix;
import android.graphics.Rect;
import android.graphics.RectF;
import android.widget.ImageView;

import net.propero.rdp.RdesktopCanvas;


public class RdesktopCanvas_Localised extends RdesktopCanvas 
{
    public void destroy()
    {
        mouse = null;
        super.destroy();
    }
    
	static final long serialVersionUID=0;
	
	boolean pointerEnabled = false;
	ScreenMouse mouse = null; 
	
    RdesktopCanvas_Localised(  ImageView parent, Activity activity, android.graphics.Bitmap bitmap, Options options, int width, int height)
    {
		super( parent, bitmap, options, width, height );

		mouse = new ScreenMouse( activity, new RectF( 0, 0, width, height ) );
	}
	
    public void setTouched( boolean touched )
    {
        mouse.setTouched( touched );
        repaintMouse();
    }
    
    public float getPointerX()
    {
        return mouse.getPointerClickPoint().x;
    }
    
    public float getPointerY()
    {
        return mouse.getPointerClickPoint().y;
    }
    
    public float getPointerPosX()
    {
        return mouse.getPosition().left;
    }
    
    public float getPointerPosY()
    {
        return mouse.getPosition().top;
    }
    
    public boolean isPointerTouched( float x, float y )
    {
        return mouse.isFingerArea( x, y );
    }
    
    public void setPointerEnabled( boolean enabled )
    {
        pointerEnabled = enabled;
        repaintMouse();
    }
    
    public void repaintMouse()
    {
        RectF newRect = mouse.getRepaintRectF();
        this.repaint( (int)newRect.left, (int)newRect.top, (int)newRect.width(), (int)newRect.height() );
    }
    
    public void movePointer( int x, int y )
    {
        mouse.setTouched( true );
        RectF oldRect = mouse.getRepaintRectF(); 
        mouse.setPosition( x, y );
        RectF newRect = mouse.getRepaintRectF();
        
        if ( pointerEnabled ) 
        {
            this.repaint( (int)oldRect.left, (int)oldRect.top, (int)oldRect.width(), (int)oldRect.height() );
            this.repaint( (int)newRect.left, (int)newRect.top, (int)newRect.width(), (int)newRect.height() );
        }
    }
    
    public void setCursor( Cursor cursor )
    {
        mouse.setMousePointer( cursor );
        RectF newRect = mouse.getRepaintRectF();
        
        if ( pointerEnabled ) 
        {
            this.repaint( (int)newRect.left, (int)newRect.top, (int)newRect.width(), (int)newRect.height() );
        }
    }

	public void update() 
	{
		Rect dstRect = this.getClipBounds();
	    logger.debug( "[Orders][Clip Area]: left=" + dstRect.left + " top=" + dstRect.top + " right=" + dstRect.right + " bottom=" + dstRect.bottom );
	    
	    this.drawBitmap( backstore, dstRect, dstRect, null );
	    
	    if ( pointerEnabled && mouse != null )
	    {
	        mouse.drawCursor( this );
	    }
	    
	    float values[] = new float[9]; 
	    parent.getImageMatrix().getValues( values );
	    float scale = values[Matrix.MSCALE_X];
	    int x = (int)values[Matrix.MTRANS_X];
	    int y = (int)values[Matrix.MTRANS_Y];
	    
	    int left = x + (int)(scale * dstRect.left);
	    int top = y + (int)(scale * dstRect.top);
	    int right = x + (int)(scale * dstRect.right);
	    int bottom = y + (int)(scale * dstRect.bottom);
	    
	    parent.postInvalidate( left,  top, right, bottom );
	}
}
