/* RdesktopFrame.java
 * Component: ProperJavaRDP
 * 
 * Revision: $Revision: 1.7 $
 * Author: $Author: telliott $
 * Date: $Date: 2005/09/27 14:15:39 $
 *
 * Copyright (c) 2005 Propero Limited
 *
 * Purpose: Window for RDP session
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * (See gpl.txt for details of the GNU General Public License.)
 * 
 */
package net.propero.rdp;

import org.apache.log4j.Logger;

import quest.android.appportal.R;

import com.provisionnetworks.common.OnRdpControlListener;

import android.app.Activity;
import android.app.Dialog;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Matrix;
import android.graphics.Rect;
import android.graphics.RectF;
import android.text.InputType;
import android.view.LayoutInflater;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import android.view.animation.Animation;
import android.view.animation.AnimationUtils;
import android.view.animation.Transformation;
import android.view.inputmethod.EditorInfo;
import android.view.inputmethod.InputMethodManager;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ImageView;
import android.widget.ImageView.ScaleType;
import android.widget.LinearLayout;
import android.widget.RelativeLayout;
import net.propero.rdp.Rdp;

public abstract class RdesktopFrame extends LinearLayout
{  
    
    public void destroy()
    {
        rdp = null;
        options = null;
        rdpControlListener = null;
         
        if ( input != null )
        {
            input.destroy();
            input = null;
        }
        
        if ( canvas != null )
        {
            canvas.destroy();
            canvas = null;
        }
        
        activity = null;
        rdpViewMainLayout = null;
        showConnBarButton = null;
        connBarLayout = null;
        connBar = null;
        connBarParams = null;
        desktopView = null;
        showConnBarArrow = null; 
        
        showKeyboardEditText = null;
        
        homeButton = null;
        keyboardButton = null;
        mouseButton = null;
        specialKeysButton = null;
        
        desktopViewBitmap = null;
    }
    
    Logger logger = Logger.getLogger(RdesktopFrame.class);

    private static final int ANIMATION_ITERATIONS = 8;
    
	private Bitmap desktopViewBitmap = null;
	public RdesktopCanvas_Localised canvas = null;
	
	public Rdp rdp = null;
	
	protected Options options = null;

	private Activity activity = null;
	
	private OnRdpControlListener rdpControlListener = null;
	private Input input = null; 
    RelativeLayout rdpViewMainLayout = null;
    View showConnBarButton = null;
    View connBarLayout = null;
    RelativeLayout connBar = null;
    RelativeLayout.LayoutParams connBarParams = null;
    ImageView desktopView = null;
    ImageView showConnBarArrow = null; 
    
    EditText showKeyboardEditText = null;
    
    View homeButton = null;
    View keyboardButton = null;
    View mouseButton = null;
    View specialKeysButton = null;
    
    ImageView highlight = null;
    
    int screenWidth = 0;
    int connectionBarWidth = 400;
    
    boolean isOnScreenKeyboardShowed = false;
    boolean savedOnScreenKeyboardShowed = false;
    
    /**
     * Create a new RdesktopFrame.
     * Size defined by Options.width and Options.height
     * Creates RdesktopCanvas occupying entire frame
     */
	public RdesktopFrame( Activity activity, Options options ) 
	{
		super(activity);
		this.activity = activity;
		this.options = options;
		
		desktopViewBitmap = Bitmap.createBitmap( options.width, options.height,  Bitmap.Config.ARGB_8888 );
		desktopViewBitmap.eraseColor( 0xFF000000 );
        
        init();
        
        try
        {
            canvas = new RdesktopCanvas_Localised( desktopView, activity, desktopViewBitmap, options, options.width, options.height);
        }
        catch(Error er)
        {
            logger.error(er.getMessage(), er.fillInStackTrace());
            er.fillInStackTrace().printStackTrace();
            er.printStackTrace();
        }
        
        initListeners();
	}
	
	public void setOnRdpControlListener( OnRdpControlListener listener )
	{
	    this.rdpControlListener = listener;
	    
	    if ( rdpControlListener != null )
	    {
	        rdpControlListener.OnRdpViewCreated( this, desktopViewBitmap );
	    }
	}
	
	private void init()
	{
	    screenWidth = activity.getWindowManager().getDefaultDisplay().getWidth();
	    
	    this.highlight = new ImageView( activity );
	    this.highlight.setScaleType( ScaleType.FIT_XY );
	    this.highlight.setImageResource( R.drawable.highlight );
	    
	    LayoutInflater inflater = LayoutInflater.from( this.getContext() );
	    inflater.inflate( R.layout.rdp_view, this );
	    
	    rdpViewMainLayout = (RelativeLayout) this.findViewById( R.id.RdpViewMainLayout );
	    this.showConnBarButton = this.findViewById( R.id.ShowConnBarButton );
	    this.showConnBarArrow = (ImageView) this.findViewById( R.id.ShowConnBarArrow );
	    this.showConnBarArrow.startAnimation( AnimationUtils.loadAnimation( activity, R.anim.rotate_to_180 ) );
        this.connBarLayout = this.findViewById( R.id.ConnectionBarLayout );
        
        this.connBar = (RelativeLayout) this.findViewById( R.id.ConnectionBar );
        connBarParams = (RelativeLayout.LayoutParams)this.connBar.getLayoutParams();
        if ( connectionBarWidth > screenWidth )
        {
            connectionBarWidth = screenWidth;
            connBarParams.leftMargin = 0;
        }
        else
        {
            int x = ((screenWidth - connectionBarWidth) >> 1 );
            connBarParams.leftMargin = x;
        }
        connBarParams.width = connectionBarWidth;
        this.connBar.setLayoutParams( connBarParams );
        
        this.homeButton = this.findViewById( R.id.HomeButton );
        this.keyboardButton = this.findViewById( R.id.KeyboardButton );
        this.mouseButton = this.findViewById( R.id.MouseButton );
        this.specialKeysButton = this.findViewById( R.id.SpecialKeysButton );
        
        this.desktopView = (ImageView) this.findViewById( R.id.DesktopView );
        this.desktopView.setImageBitmap( desktopViewBitmap );
	}
	
	protected void onMeasure( int widthMeasureSpec, int heightMeasureSpec )
	{
	    if ( activity != null )
	    {
            Rect rect = new Rect();
            activity.getWindow().getDecorView().getWindowVisibleDisplayFrame(rect);
            int statusBarHeight = rect.bottom;
            
            int screenHeight = activity.getWindowManager().getDefaultDisplay().getHeight();
            
            int diff = screenHeight - statusBarHeight;
            
            if ( diff < 128 ) 
            {
                isOnScreenKeyboardShowed = false;
            }
            else
            {
                isOnScreenKeyboardShowed = true;
            }
	    }
	    super.onMeasure( widthMeasureSpec, heightMeasureSpec );
	}
		
	private void showKeyboard()
	{
	    hideKeyboard();
	    
	    if ( showKeyboardEditText == null )
        {
            showKeyboardEditText = new EditText( activity );
            ((LinearLayout)rdpViewMainLayout.findViewById( R.id.TextContainer )).addView( showKeyboardEditText );
            showKeyboardEditText.setBackgroundColor( 0x00000000 );
            showKeyboardEditText.setTextColor( 0x00000000 );
            showKeyboardEditText.setInputType( InputType.TYPE_CLASS_TEXT | 
                    InputType.TYPE_TEXT_FLAG_IME_MULTI_LINE | 
                    InputType.TYPE_TEXT_FLAG_MULTI_LINE |
                    InputType.TYPE_TEXT_FLAG_NO_SUGGESTIONS );
            showKeyboardEditText.setImeOptions( EditorInfo.IME_ACTION_DONE | EditorInfo.IME_FLAG_NO_ENTER_ACTION );
            
            showKeyboardEditText.setOnKeyListener( input.getOnKeyListener() );
            showKeyboardEditText.addTextChangedListener( input.getTextWatcher() );
        }
	    
	    InputMethodManager imm = (InputMethodManager) activity.getSystemService( Context.INPUT_METHOD_SERVICE );
	    showKeyboardEditText.setText( "" );
        showKeyboardEditText.requestFocus();
        imm.showSoftInput( showKeyboardEditText, 0 );
	}
	
	private void hideKeyboard()
	{
	    if ( showKeyboardEditText != null )
	    {
	        InputMethodManager imm = (InputMethodManager) activity.getSystemService( Context.INPUT_METHOD_SERVICE );
	        
	        imm.hideSoftInputFromWindow( showKeyboardEditText.getWindowToken(), 0 );
            ((LinearLayout)rdpViewMainLayout.findViewById( R.id.TextContainer )).removeView( showKeyboardEditText );
            showKeyboardEditText = null;
	    }
	}
	
	private void showHighlight( View view )
	{
	    final int highlightSize = 100;
	    final int highlightHalfSize = 50;
	    final int highlightHalfSizeModified = 40;
	    
	    int xCenter = connBarLayout.getLeft() + (view.getRight() + view.getLeft()) / 2;
        int yCenter = connBarLayout.getTop() + (view.getTop() + view.getBottom()) / 2;
        
        RelativeLayout.LayoutParams params = new RelativeLayout.LayoutParams( highlightSize, highlightSize );
        params.leftMargin = xCenter - highlightHalfSize;
        params.topMargin = yCenter - highlightHalfSizeModified;
        params.rightMargin = connBar.getMeasuredWidth() - (params.leftMargin + highlightSize); 
        
        if ( highlight.getParent() != null )
        {
            hideHighlight();
        }
        connBar.addView( highlight, params );
	}
	
	private void hideHighlight()
	{
	    connBar.removeView( highlight );
	}
	
	private void initListeners()
    {
	    final View thisView = this;
	    
        this.showConnBarButton.setOnTouchListener( showConnBarButtonTouchListener );
        
        this.homeButton.setOnTouchListener( new OnTouchListener()
        {
            public boolean onTouch( View view, MotionEvent event )
            {
                try
                {
                    if ( event.getActionMasked() == MotionEvent.ACTION_DOWN )
                    {
                        showHighlight( view );
                    }
                    else if ( event.getActionMasked() == MotionEvent.ACTION_UP )
                    {
                        savedOnScreenKeyboardShowed = isOnScreenKeyboardShowed;
                        
                        if ( isOnScreenKeyboardShowed )
                        {
                            hideKeyboard();
                        }
                        
                        if ( rdpControlListener != null )
                        {
                            rdpControlListener.OnRdpViewBackClicked( thisView );
                        }
                        
                        hideHighlight();
                    }
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
                
                return true; 
            }
        });
        
        this.keyboardButton.setOnTouchListener( new OnTouchListener()
        {
            public boolean onTouch( View view, MotionEvent event )
            {
                try
                {
                    if ( event.getActionMasked() == MotionEvent.ACTION_DOWN )
                    {
                        showHighlight( view );
                    }
                    else if ( event.getActionMasked() == MotionEvent.ACTION_UP )
                    {
                        if ( isOnScreenKeyboardShowed )
                        {
                            hideKeyboard();
                        }
                        else
                        {
                            showKeyboard();
                        }
                        
                        hideHighlight();
                    }
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
                return true;
            }
        });
        
        this.mouseButton.setOnTouchListener( new OnTouchListener()
        {
            public boolean onTouch( View view, MotionEvent event )
            {
                try
                {
                    if ( event.getActionMasked() == MotionEvent.ACTION_DOWN )
                    {
                        showHighlight( view );
                    }
                    else if ( event.getActionMasked() == MotionEvent.ACTION_UP )
                    {
                        Matrix matrix = desktopView.getImageMatrix();
                        float[] values = new float[9];
                        matrix.getValues( values );
    
                        float scale = values[Matrix.MSCALE_X];
                        
                        float xPos = values[Matrix.MTRANS_X];
                        float yPos = values[Matrix.MTRANS_Y];
                        
                        float width = rdpViewMainLayout.getWidth();
                        float height = rdpViewMainLayout.getHeight();
                        
                        int imageWidth = desktopView.getDrawable().getIntrinsicWidth();
                        int imageHeight = desktopView.getDrawable().getIntrinsicHeight();
    
                        float scaledWidth = imageWidth * scale;
                        float scaledHeight = imageHeight * scale;
                        
                        if ( scaledWidth < width ) width = scaledWidth + 2*xPos;
                        if ( scaledHeight < height ) height = scaledHeight + 2*yPos;
                        
                        float x = ((width/2F) - xPos) / scale;
                        float y = ((height/2F) - yPos) / scale;
                        
                        input.setScreenMouseEnabled( !input.isScreenMouseEnabled(), x, y );
                        hideHighlight();
                    }
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
                return true;
            }
        });
        
        this.specialKeysButton.setOnTouchListener( new OnTouchListener()
        {
            public boolean onTouch( View view, MotionEvent event )
            {
                try
                {
                    if ( event.getActionMasked() == MotionEvent.ACTION_DOWN )
                    {
                        showHighlight( view );
                    }
                    else if ( event.getActionMasked() == MotionEvent.ACTION_UP )
                    {
                        showSpecialKeysDialog();
                        hideHighlight();
                    }
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
                return true;
            }
        });
    }
	
	private void showSpecialKeysDialog()
	{
	    if ( activity == null ) return;
	    
	    LinearLayout keys = new LinearLayout( activity );
        keys.setOrientation( LinearLayout.VERTICAL );
        
        final Dialog dialog = new Dialog( activity );
        dialog.setCancelable( true );
        dialog.setTitle( "Session Actions" );
        dialog.setContentView( keys, new LinearLayout.LayoutParams(300, ViewGroup.LayoutParams.WRAP_CONTENT) );
        
        Button sendCAD = new Button( activity );
        sendCAD.setWidth( ViewGroup.LayoutParams.MATCH_PARENT );
        sendCAD.setText("Send CTRL+ALT+DEL");
        sendCAD.setLines( 1 );
        sendCAD.setOnClickListener( new OnClickListener()
        {
            public void onClick( View view )
            {
                try
                {
                    dialog.dismiss();
                    input.sendCtrlAltDel();
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
            }
        });
        keys.addView( sendCAD );
        
        Button sendWSK = new Button( activity );
        sendWSK.setWidth( ViewGroup.LayoutParams.MATCH_PARENT );
        sendWSK.setText( "Send Windows Key" );
        sendWSK.setLines( 1 );
        sendWSK.setOnClickListener( new OnClickListener()
        {
            public void onClick( View view )
            {
                try
                {
                    dialog.dismiss();
                    input.sendWindowsKey();
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
            }
        });
        keys.addView( sendWSK );
        dialog.setCanceledOnTouchOutside( true );
        dialog.show();
	}
	
    private void touchShowConnBar()
    {
        if ( connBar.getAnimation() != null && connBar.getAnimation().hasStarted() )
        {
            return;
        }
        
        if ( connBarLayout.getVisibility() == View.GONE )
        {
            Animation animation = new Animation()
            {
                float pos = -connBarLayout.getHeight();
                float dpos = (float)connBarLayout.getHeight() / ANIMATION_ITERATIONS;
                
                protected void applyTransformation( float interpolatedTime, Transformation t )
                {
                    if ( pos < 0 )
                    {
                        connBarParams.topMargin = (int)pos;
                        pos += dpos;
                    }
                    else
                    {
                        connBarParams.topMargin = 0;
                    }
                    connBar.setLayoutParams( connBarParams );
                }
            };
            animation.setRepeatCount( ANIMATION_ITERATIONS + 1 );
            
            connBarLayout.setVisibility( View.VISIBLE );
            showConnBarArrow.startAnimation( AnimationUtils.loadAnimation( activity, R.anim.rotate_to_180 ) );
            connBar.startAnimation( animation );
        }
        else if ( connBarLayout.getVisibility() == View.VISIBLE )
        {
            Animation animation = new Animation()
            {
                float pos = 0;
                float dpos = (float)connBarLayout.getHeight() / ANIMATION_ITERATIONS;
                
                protected void applyTransformation( float interpolatedTime, Transformation t )
                {
                    if ( pos + connBarLayout.getHeight() > 0 )
                    {
                        connBarParams.topMargin = (int)pos;
                        pos -= dpos;
                    }
                    else
                    {
                        connBarLayout.setVisibility( View.GONE );
                        connBarParams.topMargin = -connBarLayout.getHeight();
                    }
                    
                    connBar.setLayoutParams( connBarParams );
                }
            };
            animation.setRepeatCount( ANIMATION_ITERATIONS + 1 );
            
            showConnBarArrow.startAnimation( AnimationUtils.loadAnimation( activity, R.anim.rotate_to_360 ) );
            connBar.startAnimation( animation );
        }
        else
        {
            connBarLayout.setVisibility( View.VISIBLE );
        }
    }
    
    OnTouchListener showConnBarButtonTouchListener = new OnTouchListener()
    {
        boolean isInMove = false;
        float innerX = 0;
        float innerY = 0;
        
        public boolean onTouch( View v, MotionEvent event )
        {
            try
            {
                if ( event.getAction() == MotionEvent.ACTION_MOVE &&
                     event.getEventTime() - event.getDownTime() > 1000 &&
                     innerX != 0 )
                {
                    screenWidth = activity.getWindowManager().getDefaultDisplay().getWidth();
                    
                    int x = (int) (event.getRawX() - innerX);
                    
                    if ( x < 0 ) x = 0;
                    if ( x + connectionBarWidth > screenWidth ) x = screenWidth - connectionBarWidth;
                    
                    connBarParams.leftMargin = x;
                    connBar.setLayoutParams( connBarParams );
                    
                    isInMove = true;
                    
                    return true;
                }
    
                RectF rect = new RectF( 75, 0, 326, 53 );
                
                switch( event.getActionMasked() )
                {
                    case MotionEvent.ACTION_DOWN:
                        innerX = event.getX();
                        innerY = event.getY();
                        if ( !rect.contains( event.getX(), event.getY() ) )
                        {
                            innerX = 0;
                        }
                        else
                        {
                            return true;
                        }
                        break;
                    case MotionEvent.ACTION_UP:
                        innerX = 0;
                        if ( isInMove )
                        {
                            isInMove = false;
                            return true;
                        }
                        else if ( rect.contains( event.getX(), event.getY() ) )
                        {
                            touchShowConnBar();
                            return true;
                        }
                        break;
                    case MotionEvent.ACTION_MOVE:
                        float x = event.getX() - innerX;
                        float y = event.getY() - innerY;
                        if ( (x*x + y*y) > 1 )
                        {
                            innerX = 0;
                        }
                        break;
                }
            }
            catch ( Exception e )
            {
                logger.error( e.getMessage(), e );
            }
            return false;
        }
    };

    /**
     * Retrieve the canvas contained within this frame
     * @return RdesktopCanvas object associated with this frame
     */
	public RdesktopCanvas getCanvas() 
	{
		return this.canvas;
	}
	
	private void initInput( )
	{
        input = new Input_Localised( activity, canvas, rdp, this );
        this.desktopView.setOnTouchListener( input );
	}
	
    /**
     * Register the RDP communications layer with this frame
     * @param rdp Rdp object encapsulating the RDP comms layer
     */
	public void registerCommLayer(Rdp rdp) 
	{
		this.rdp = rdp;
		if ( input != null )
		{
		    input.registerCommLayer( rdp );
		}
		else
		{
		    initInput( );
		}
	}
	
	protected void onAttachedToWindow()
	{
	    super.onAttachedToWindow();
	    
	    this.post( new Runnable()
        {
            public void run()
            {
                try
                {
                    if ( savedOnScreenKeyboardShowed )
                    {
                        showKeyboard();
                    }
                    else
                    {
                        hideKeyboard();
                    }
                }
                catch ( Exception e )
                {
                    logger.error( e.getMessage(), e );
                }
            }
        });
	}
}
