package net.propero.rdp.orders.emfplus;

import net.propero.rdp.orders.emfplus.enumerations.*;

import org.apache.log4j.Logger;

public class EmfPlusObjectsFactory {
	
	static Logger logger = Logger.getLogger(EmfPlusObjectsFactory.class);
	
	public static EmfPlusRecord getRecord(OrderBuffer buffer) {
		logger.info("Get Record");
		
		int pos = buffer.getPosition();
		int type = buffer.getLittleEndian16();
		buffer.setPosition(pos);
		
		switch (type) {
		case RecordTypeEnumeration.EmfPlusComment:
			return new EmfPlusComment(buffer);
		case RecordTypeEnumeration.EmfPlusObject:
			return getObject(buffer);
	    case RecordTypeEnumeration.EmfPlusDrawImage:
	    	return new EmfPlusDrawImage(buffer);
	    case RecordTypeEnumeration.EmfPlusSetClipRegion:
	    	return new EmfPlusSetClipRegion(buffer);
	    default:
	    	logger.error("This type of EMF+ record is not supported. Record type: " + Integer.toHexString(type));
	    	break;
		}
		
		return null;
	}
	
	private static EmfPlusRecord getObject(OrderBuffer buffer) {
		logger.info("Get Object");
		
		int pos = buffer.getPosition();
		EmfPlusObject record = new EmfPlusObject(buffer);
		buffer.setPosition(pos);
		
		switch(record.getObjectType()) {
		case ObjectTypeEnumeration.ObjectTypeRegion:
			return new EmfPlusRegion(buffer);
		case ObjectTypeEnumeration.ObjectTypeImage:
			return getImageObject(buffer);
		default:
			logger.error("This type of EMF+ object is not supported. Object type: " + Integer.toHexString(record.getObjectType()));
			break;
		}
		
		return null;
	}
	
	private static EmfPlusRecord getImageObject(OrderBuffer buffer) {
		logger.info("Get Image");
		
		int pos = buffer.getPosition();
		EmfPlusImage record = new EmfPlusImage(buffer);
		buffer.setPosition(pos);
		
		switch(record.getImageType()) {
		case ImageDataTypeEnumeration.ImageDataTypeBitmap:
			return new EmfPlusBitmap(buffer);
		default:
			logger.error("This type of EMF+ image type is not supported. Image type: " + Integer.toHexString(record.getImageType()));
			break;
		}
		
		return null;
	}
	
}
