package net.propero.rdp.orders.emfplus;

import org.apache.log4j.Logger;

import quest.androidport.Point2D;

public class EmfPlusPath extends EmfPlusObject {
	
	Logger logger = Logger.getLogger(EmfPlusPath.class);
	
    private int pathPointCount = 0;
    private int pathFlags = 0;
    private int reserved = 0;
        
    private Point2D.Float[] points = null;
    private IEmfPlusPathPointType[] pointTypes = null;
    
    private boolean pathPbit = false;
	private boolean pathRbit = false;
	private boolean pathCbit = false;
    
    private byte getInteger7(byte value)
	{
		byte result = 0;
		byte signFlag = (byte)((value & 0x40) << 1);
		result = (byte)((value & 0x7F) | signFlag);
		return result;
	}
	
	private short getInteger15(short value)
	{
		short result = 0;
		short signFlag = (short)((value & 0x4000) << 1);
		result = (short)((value & 0x7FFF) | signFlag);
		return result;
	}
	
	public EmfPlusPath(OrderBuffer buffer, boolean isNodeObject) {
		super();
		init(buffer, isNodeObject);
	}
	
	public EmfPlusPath(OrderBuffer buffer) {
    	super(buffer);
		init(buffer, false);
    }
	
	private void init(OrderBuffer buffer, boolean isNodeObject) {
		if(isNodeObject)
		{
			this.version = buffer.getLittleEndian32();
		}
		this.pathPointCount = buffer.getLittleEndian32();
		this.pathFlags = buffer.getLittleEndian16();
		
		this.pathPbit = (((this.pathFlags >> 11) & 1) == 1); 
		this.pathRbit = (((this.pathFlags >> 12) & 1) == 1);
		this.pathCbit = (((this.pathFlags >> 14) & 1) == 1);
		
		this.reserved = buffer.getLittleEndian16();
		this.points = new Point2D.Float[this.pathPointCount];
		for (int i = 0; i < this.pathPointCount; ++i)
		{
			this.points[i] = new Point2D.Float();
			if (this.getPbit())
			{
				//EmfPlusPointR
				int type = buffer.get8(buffer.getPosition());
				int flag = (type >> 7);
				if (flag == 0)
				{
					this.points[i].x = getInteger7((byte)buffer.get8());
				}
				else
				{
					this.points[i].x = getInteger15((short)buffer.getLittleEndian16());
				}
				
				type = buffer.get8(buffer.getPosition());
				flag = (type >> 7);
				if (flag == 0)
				{
					this.points[i].y = getInteger7((byte)buffer.get8());
				}
				else
				{
					this.points[i].y = getInteger15((short)buffer.getLittleEndian16());
				}
			}
			else if (this.getCbit())
			{
				//EmfPlusPoint
				this.points[i].x = buffer.getLittleEndian16();
				this.points[i].y = buffer.getLittleEndian16();
			}
			else
			{
				//EmfPlusPointF
				this.points[i].x = Float.intBitsToFloat(buffer.getLittleEndian32());
				this.points[i].y = Float.intBitsToFloat(buffer.getLittleEndian32());
			}
		}
		
		this.pointTypes = new IEmfPlusPathPointType[this.pathPointCount];
		
		for(int i = 0; i < this.pathPointCount; ++i)
		{
			if (this.getRbit())
			{
				this.pointTypes[i] = new EmfPlusPathPointTypeRLE(buffer);
			}
			else
			{
				this.pointTypes[i] = new EmfPlusPathPointType(buffer);
			}
		}
	}
	
    public boolean getPbit()
    {
    	return pathPbit;
    }
    
    public boolean getRbit()
    {
    	return pathRbit;
    }
    
    public boolean getCbit()
    {
    	return pathCbit;
    }
    
    public Point2D.Float[] getPoints() {
    	return points;
    }
    
    public IEmfPlusPathPointType[] getPointTypes() {
    	return pointTypes;
    }
}
