package quest.androidport;

import android.app.Activity;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.PointF;
import android.graphics.RectF;

public class ScreenMouse
{
    private static final int MAX_CURSOR_WIDTH = 32;
    private static final int MAX_CURSOR_HEIGHT = 32;
    
    private RectF viewSize = null;
    
    android.graphics.Bitmap fingerPointer = null;
    android.graphics.Bitmap fingerPointerTouched = null;
    RectF pointerRect = null;
    Cursor cursor = null;
    boolean touched = false;
    
    public ScreenMouse( Activity activity, RectF viewSize )
    {
        this.viewSize = viewSize;
        
        fingerPointer = BitmapFactory.decodeResource(activity.getResources(), quest.android.appportal.R.drawable.pointer);
        fingerPointerTouched = BitmapFactory.decodeResource(activity.getResources(), quest.android.appportal.R.drawable.pointer_alt);
        pointerRect = new RectF( 0, 0, fingerPointer.getWidth() + MAX_CURSOR_WIDTH, fingerPointer.getHeight() + MAX_CURSOR_HEIGHT );
    }
    
    public void setMousePointer( Cursor cursor )
    {
        this.cursor = cursor;
    }
    
    public void setTouched( boolean touched )
    {
        this.touched = touched;
    }
    
    public void setPosition( int x, int y )
    {
        setPosition( (float) x, (float) y );
    }
    
    public void setPosition( float x, float y )
    {
        if ( x < 0 ) x = 0;
        if ( y < 0 ) y = 0;
        if ( x + pointerRect.width() > viewSize.width() ) x = viewSize.width() - pointerRect.width(); 
        if ( y + pointerRect.height() > viewSize.height() ) y = viewSize.height() - pointerRect.height();
        
        pointerRect.offsetTo( x, y );
    }
    
    public RectF getPosition()
    {
        return pointerRect;
    }
    
    public RectF getRepaintRectF()
    {
        RectF res = new RectF();
        
        res.left = pointerRect.left - pointerRect.width();
        res.top = pointerRect.top - pointerRect.height();
        res.right = pointerRect.right + pointerRect.width();
        res.bottom = pointerRect.bottom + pointerRect.height();
        
        res.intersect( viewSize );
        
        return res;
    }
    
    public PointF getPointerClickPoint()
    {
        PointF dot = new PointF( pointerRect.left, pointerRect.top );
        
        if ( viewSize.right - pointerRect.right < pointerRect.width() )
        {
            dot.x += (pointerRect.width() - (viewSize.right - pointerRect.right));
        }
        if ( viewSize.bottom - pointerRect.bottom < pointerRect.height() )
        {
            dot.y += (pointerRect.height() - (viewSize.bottom - pointerRect.bottom));
        }
        
        return dot;
    }
    
    public boolean isFingerArea( float x, float y )
    {
        RectF area = new RectF();
        
        int cursorWidth = 0;
        int cursorHeight = 0;
        
        if ( cursor != null )
        {
            cursorWidth = cursor.getWidth();
            cursorHeight = cursor.getHeight();
        }
        
        area.set( pointerRect.left + cursorWidth, pointerRect.top + cursorHeight, 
                pointerRect.left + cursorWidth + fingerPointer.getWidth(), pointerRect.top + cursorHeight + fingerPointer.getHeight() );
        
        return area.contains( x, y );
    }
    
    public void drawCursor( Canvas canvas )
    {
        int cursorWidth = 0;
        int cursorHeight = 0;
        
        PointF point = getPointerClickPoint();
        
        if ( cursor != null )
        {
            cursorWidth = cursor.getWidth();
            cursorHeight = cursor.getHeight();
        }
        
        if ( touched )
        {
            canvas.drawBitmap( fingerPointerTouched, pointerRect.left + cursorWidth, pointerRect.top + cursorHeight, null );
        }
        else
        {
            canvas.drawBitmap( fingerPointer, pointerRect.left + cursorWidth, pointerRect.top + cursorHeight, null );
        }
        
        if ( cursor != null && cursor.getCursor() != null )
        {
            canvas.drawBitmap( cursor.getCursor(), point.x - cursor.getOffsetX(), point.y - cursor.getOffsetY(), null );
        }
    }
}
