package org.apache.log4j;

import java.io.File;
import java.io.FilePermission;
import java.io.IOException;
import java.security.AllPermission;
import java.security.PermissionCollection;
import java.security.Permissions;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Enumeration;

import org.apache.commons.logging.Log;

import android.os.Environment;

import com.google.code.microlog4android.appender.Appender;

public class Logger implements Enumeration<Appender>, Log 
{
	
    protected static Logger logger = null;
	protected static com.google.code.microlog4android.Logger log = null;
	
	protected String source = "";
	protected static String globalSource = "";
	
	public static void setEnabled( boolean enabled )
	{
	    if ( log == null ) return;
	        
	    if ( enabled )
	    {
	        log.setLevel( com.google.code.microlog4android.Level.DEBUG );
	    }
	    else
	    {
	        log.setLevel( com.google.code.microlog4android.Level.FATAL );
	    }
	}
	
	protected Logger(String name, com.google.code.microlog4android.repository.CommonLoggerRepository resourceBundleName) 
	{
		if ( log == null )
		{
			log = new com.google.code.microlog4android.Logger(name, resourceBundleName);
			log.removeAllAppenders();
			
		    SimpleDateFormat sdf = new SimpleDateFormat("HH-mm-dd-MM-yy");
            // create a File object for the parent directory
            File logsDir = new File( Environment.getExternalStorageDirectory(), "vWorkspace/" );
            
            // have the object build the directory structure, if needed.
            if ( !logsDir.exists() && !logsDir.mkdirs() )
            {
                logsDir = new File( Environment.getDataDirectory(), "vWorkspace");
                logsDir.mkdir();
            }
            
            File parent = Environment.getExternalStorageDirectory().getParentFile();
            String pathToRoot = "";
            while (parent != null )
            {
                pathToRoot += "/..";
                parent = parent.getParentFile();
            }
            
            pathToRoot = pathToRoot.substring( 1 );
            
            for ( int i = 0; (logsDir.listFiles() != null && i < logsDir.listFiles().length) ; ++i )
            {
                try
                {
                    logsDir.listFiles()[i].delete();
                }
                catch( Exception e )
                {
                    System.err.println( e + "\n" + logsDir.listFiles()[i]);
                }
            }
                
            String filePath = pathToRoot + logsDir.getAbsolutePath() + "/vWorkspace_" + sdf.format( Calendar.getInstance().getTime() ) + ".log";
            
			log.addAppender(new com.google.code.microlog4android.appender.LogCatAppender());
			
			try
            {
			    Runtime.getRuntime().exec("chmod 777 " + pathToRoot + logsDir.getAbsolutePath() );
                Runtime.getRuntime().exec("chmod 777 " + filePath);
            }
            catch( IOException e ) { }
			
			com.google.code.microlog4android.appender.FileAppender file =  
			new com.google.code.microlog4android.appender.FileAppender();
			file.setFileName( filePath );
			log.addAppender(file);
		}
		
		source = name;
	}
	
	public static Logger getLogger(Class<?> cl)
	{
		return new Logger(cl.getName(), null);
	}
	
	public static org.apache.log4j.Logger getLogger(String name)
	{
		return new Logger(name, null);
	}
	
	public Enumeration<Appender> getAllAppenders()
	{
		if ( log.getNumberOfAppenders() == 0 )
		{
			return null;
		}
		else
		{	
			appenderIndex = 0;
			Logger logger = this;
			return logger;
		}
	}
	
	private int appenderIndex = 0;
	
	public boolean hasMoreElements()
	{
		return ( appenderIndex < log.getNumberOfAppenders() );
	}
	
	public Appender nextElement()
	{
		++appenderIndex;
		return log.getAppender(appenderIndex - 1); 
	}
	
	public void debug(Object msg)
	{
	    msg = source + ": " + globalSource + " " + msg;
		this.log.debug(msg);
	}
	
	public void warn(Object msg)
	{
	    msg = source + ": " + globalSource + " " + msg;
		log.warn(msg);
	}
	
	public void error(Object msg)
	{
	    msg = source + ": " + globalSource + " " + msg;
		log.error(msg);
	}
	
	public void error(Object msg, Throwable thrown)
	{
	    msg = source + ": " + globalSource + " " + msg;
		log.error( msg, thrown);
	}
	
	public void info(Object msg)
	{
	    msg = source + ": " + globalSource + " " + msg;
		this.log.info(msg);
	}
	
	public void info(Object msg, Throwable thrown)
	{
		msg = source + ": " + globalSource + " " + msg;
		this.log.info(msg, thrown);
	}
	
	public void fatal(Object msg)
	{
		msg = source + ": " + msg;
		this.log.fatal(msg);
	}
	
	public void setLevel(Object level)
	{
		log.setLevel((com.google.code.microlog4android.Level)level);
	}
	
	public boolean isInfoEnabled()
	{
		return true;
	}

	public void debug(Object arg0, Throwable arg1) {
	    arg0 = source + ": " + globalSource + " " + arg0;
		log.debug(arg0, arg1);
	}

	public void fatal(Object arg0, Throwable arg1) {
	    arg0 = source + ": " + globalSource + " " + arg0;
		log.fatal(arg0, arg1);
	}

	public boolean isDebugEnabled() {
		return log.isDebugEnabled();
	}

	public boolean isErrorEnabled() {
		return true;
	}

	public boolean isFatalEnabled() {
		return true;
	}

	public boolean isTraceEnabled() {
		return log.isTraceEnabled();
	}

	public boolean isWarnEnabled() {
		return true;
	}

	public void trace(Object arg0) {
	    arg0 = source + ": " + globalSource + " " + arg0;
		log.trace(arg0);
	}

	public void trace(Object arg0, Throwable arg1) {
	    arg0 = source + ": " + globalSource + " " + arg0;
		log.trace(arg0, arg1);		
	}

	public void warn(Object arg0, Throwable arg1) {
		arg0 = source + ": " + globalSource + " " + arg0;
		log.warn(arg0, arg1);
	}
	
	public static void setSource( String string )
	{
	    globalSource = string;
	}
}
