package quest.android.appportal.common;

import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;
import java.util.UUID;

/**
 * Class contains only static methods.
 */
public class Common
{
    public static final int INVALID_VALUE = 0xFFFFFFFF;
    public static final String EMTPY_STRING = "";
    
    public static boolean isNullOrEmpty( String str )
    {
        return ( ( str == null ) || ( str.length() == 0 ) );
    }
    
    public static boolean isCorrectName( String str )
    {
        if ( isNullOrEmpty(str) )
        {
            return false;
        }
        
        String res = str.replace( " \n\r\0", "");
        
        if ( isNullOrEmpty(res) )
        {
            return false;
        }
        
        return true;
    }
    
    public static String generateUUID()
    {
        return UUID.randomUUID().toString();
    }
    
    public static int parseIntSafe( String string, int defaultValue )
    {
        try
        {
            return Integer.parseInt( string ); 
        }
        catch( Exception e ) { }
        
        return defaultValue;
    }
    
    public static int parseIntSafe( String string )
    {
        return parseIntSafe( string, 0 );
    }
    
    public static String getLocalIpAddress() 
    {
        try 
        {
            for ( Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces(); en.hasMoreElements(); ) 
            {
                NetworkInterface intf = en.nextElement();
                for ( Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses(); enumIpAddr.hasMoreElements(); )
                {
                    InetAddress inetAddress = enumIpAddr.nextElement();
                    
                    if ( !inetAddress.isLoopbackAddress() ) 
                    {
                        return inetAddress.getHostAddress().toString();
                    }
                }
            }
        } 
        catch (SocketException e) 
        {
            
        }
        return null;
    }
    
    public static String getLocalHostName() 
    {
        try 
        {
            for ( Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces(); en.hasMoreElements(); ) 
            {
                NetworkInterface intf = en.nextElement();
                for ( Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses(); enumIpAddr.hasMoreElements(); )
                {
                    InetAddress inetAddress = enumIpAddr.nextElement();
                    
                    if ( !inetAddress.isLoopbackAddress() ) 
                    {
                        return inetAddress.getHostName();
                    }
                }
            }
        } 
        catch (SocketException e) 
        {
        }
        return null;
    }
}
