
package quest.android.appportal.common;

import java.util.ArrayList;
import org.apache.log4j.Logger;

public class ICOConverter
{

    private Logger logger     = Logger.getLogger( ICOConverter.class );

    // Fields:
    public static short   PREF_ByPP  = 4;                                     // PREFix
                                                                               // BYte_Per_Pixel;
    public static short   PREF_VCAPS = 8;                                     // PREFix
                                                                               // Value
                                                                               // CAPasity
                                                                               // in
                                                                               // Symbols;
                                                                               // (PREF_ByPP*SYMB_PBYTE)
    public static short   SYMB_PBYTE = 2;                                     // SYMBols
                                                                               // Per_BYTE;

    // For Prefix:
    private int           PR_SIZE;                                            // PRefix
                                                                               // size:
                                                                               // (first+second
                                                                               // parts)
    private int           ICON_ByPP;                                          // ICON
                                                                               // Byte_Per_Pixel;

    private int           IHEIGHT, IWIDTH, capacity;                          // total
                                                                               // icon_byte_array
                                                                               // capasity
                                                                               // in
                                                                               // bytes

    private String        // alfaMask, mask for alfa
            redMask, // mask for RED
            greenMask, // mask for GREEN
            blueMask;                                                         // mask
                                                                               // for
                                                                               // BLUE
    private int           redBits, // stores number of bits for each color_part.
            greenBits, blueBits;
    private int           redShift, // stores shifting values, for each
                                    // color_part.
            greenShift, blueShift;
    private int           castBits;

    private int[]         rgbArray;                                           // Array
                                                                               // of
                                                                               // color_values

    // Methods:
    private void init()
    {
        PR_SIZE = 16 + 1; // will has full size after
        // dynamic calculation in getPrefixConfigs().
        ICON_ByPP = 0;

        IHEIGHT = 0;
        IWIDTH = 0;
        capacity = 0;

        //alfaMask = "";
        redMask = "";
        greenMask = "";
        blueMask = "";

        redBits = 0;
        greenBits = 0;
        blueBits = 0;

        redShift = 0;
        greenShift = 0;
        blueShift = 0;

        castBits = 0;
    }

    private int getExponent( int val )
    {
        int res = 0;
        while ( val > 1 )
        {
            val >>= 1;
            ++res;
        }
        return res;
    }

    private void getPrefixConfigs( String strPrefix )
    {

        ArrayList<String> prefix_byte_list = new ArrayList<String>();

        // Make a list of 4byted values. of prefix. for retrieving necessary
        // data.
        // Reads first part: 1. [0-15] PR_SIZE=16 and 2. [16] PR_SIZE=17
        for( int i = 0; i < PR_SIZE; i++ )
        { // Copy every 8 symbols string:
            prefix_byte_list.add( i, strPrefix.substring( i * 8, i * 8 + 8 ) );
        }
        // Retrieving data: 6th - height, 7th - width. This values refares to
        // alfa_array.
        IHEIGHT = Integer.decode( "#" + prefix_byte_list.get( 6 ) );
        IWIDTH = Integer.decode( "#" + prefix_byte_list.get( 7 ) );

        // Reads third part: 3. [17-(17+'width')].
        for( int i = PR_SIZE; i < (PR_SIZE + IWIDTH); i++ )
        {
            prefix_byte_list.add( i, strPrefix.substring( i * 8, i * 8 + 8 ) );
        }
        PR_SIZE += IWIDTH; // Set to start position of 3rd part.

        // Reads fourth part: 4. [(PR_SIZE)-(PR_SIZE+13)].
        for( int i = PR_SIZE; i < (PR_SIZE + 13); i++ )
        { // Copy every 8 symbols string:
            prefix_byte_list.add( i, strPrefix.substring( i * 8, i * 8 + 8 ) );
        }

        // Its not necessary but I think that this values refares exectly to
        // icon_array.
        IHEIGHT = Integer
                .decode( "#" + prefix_byte_list.get( PR_SIZE - 1 + 2 ) );
        IWIDTH = Integer.decode( "#" + prefix_byte_list.get( PR_SIZE - 1 + 3 ) );
        capacity = Integer
                .decode( "#" + prefix_byte_list.get( PR_SIZE - 1 + 6 ) );
        ICON_ByPP = (int) (capacity / (IHEIGHT * IWIDTH));
        PR_SIZE += 13; // Set to total prefix size.
        // Masks initialization:
        // alfaMask = prefix_byte_list.get( PR_SIZE - 4 ); // without "#". We dont
                                                        // use alfa for now.
        redMask = prefix_byte_list.get( PR_SIZE - 3 );
        greenMask = prefix_byte_list.get( PR_SIZE - 2 );
        blueMask = prefix_byte_list.get( PR_SIZE - 1 );

        // calculate number of bits for each color_part:
        int bMaxValue = Integer.decode( "#" + blueMask ) + 1;
        blueBits = getExponent( bMaxValue );
        int gMaxValue = (Integer.decode( "#" + greenMask ) >> blueBits) + 1;
        greenBits = getExponent( gMaxValue );
        int rMaxValue = ((Integer.decode( "#" + redMask )) >> (blueBits + greenBits)) + 1;
        redBits = getExponent( rMaxValue );

        // calculates shifting_value for each color_part:
        if ( ICON_ByPP > 2 )
        {
            // This we do course 4_bytes int has sing. and it courses
            // exceptions.
            castBits = 0;
            blueShift = 0;
            greenShift = blueShift + greenBits;
            redShift = greenShift + redBits;
        }
        else
        {
            castBits = 8;
            blueShift = blueBits;
            greenShift = blueShift + greenBits;
            redShift = greenShift + redBits;
        }

    }

    public int getWidth()
    {
        return IWIDTH;
    }

    public int getHeight()
    {
        return IHEIGHT;
    }

    public int[] convertIcon( String iconString )
    {
        logger.debug( "Starting convert icon" );
        
        if ( iconString == null || iconString.length() <= 0 )
        {
            return null;
        }
        init();

        // Declaration&Initialization.
        // Prefix:
        getPrefixConfigs( iconString );

        rgbArray = new int[IHEIGHT * IWIDTH];
        byte[] byte_codes; // Create main array in this program, byte_array of
                           // icon.
        // Reads after prefix.
        byte_codes = Hex.toByteArr( iconString
                .substring( PREF_VCAPS * PR_SIZE ) );

        // We cast all types to TYPE_4BYTE_ABGR - RGB with alfa.
        int alfa, R, G, B;
        int index = 0, hex = 0;
        String str = "#";
        // *Declaration&Initialization.

        // ----------------- Create icon --------------------
        for( int x = 0; x < IWIDTH; x++ )
        {
            for( int y = 0; y < IHEIGHT; y++ )
            {
                index = (x * IHEIGHT * ICON_ByPP) + (y * ICON_ByPP);
                str = "#";
                // Sach initialization course 4byted int use 4th bate for sign
                // and we reads 3 bytes of colors.
                for( int j = (ICON_ByPP < 4) ? (ICON_ByPP - 1)
                        : (ICON_ByPP - 2); j >= 0; j-- )
                {
                    str += Hex.toHex( byte_codes[index + j] );
                }
                hex = Integer.decode( str );
                B = ((hex & Integer.decode( "#" + blueMask )) << castBits >> blueShift) & 0xFF;
                G = ((hex & Integer.decode( "#" + greenMask )) << castBits >> greenShift) & 0xFF;
                R = ((hex & Integer.decode( "#" + redMask )) << castBits >> redShift) & 0xFF;

                alfa = (R == 0 && G == 0 && B == 0) ? (alfa = 0) : (alfa = 255);
                int color = ((alfa << 24) & 0xFF000000)
                        | ((R << 16) & 0x00FF0000) | ((G << 8) & 0x0000FF00)
                        | (B & 0x000000FF);
                rgbArray[(IWIDTH - x - 1) * IHEIGHT + y] = color;
            }
        }

        // *-------------- Create icon -------------------

        return rgbArray;
    }
}
