package quest.android.appportal.controller;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.provisionnetworks.common.OnRdpControlListener;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.DialogInterface.OnDismissListener;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.Point;
import android.view.GestureDetector;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnTouchListener;
import android.view.inputmethod.InputMethodManager;
import android.widget.AdapterView;
import android.widget.Gallery;
import android.widget.GridView;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.RelativeLayout;
import android.widget.TextView;
import android.widget.ViewAnimator;
import android.widget.AdapterView.OnItemClickListener;
import quest.android.appportal.R;
import quest.android.appportal.common.Common;
import quest.android.appportal.connection.PNApplication;
import quest.android.appportal.connection.PNException;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNLocation;
import quest.android.appportal.model.IModel;
import quest.android.appportal.model.RdpThread;
import quest.android.appportal.view.custom.ApplicationsGridAdapter;
import quest.android.appportal.view.custom.FarmListAdapter;
import quest.android.appportal.view.custom.RdpViewGalleryAdapter;
import quest.android.appportal.view.custom.LocationListAdapter;
import quest.android.appportal.view.custom.FarmListAdapter.FarmListItem;
import quest.android.appportal.view.custom.SwipeDetector;

public class DefaultController implements IController
{
    private Logger logger = Logger.getLogger( DefaultController.class );
    
    private Activity activity = null;
    private Resources resources = null;
    private IModel model = null;
    
    private DialogController dialogController = null;
    private ProgressDialog progressDialog = null;
    
    private Gallery rdpViewGallery = null;
    private Thread rdpViewGalleryInvalidator = null;
    
    TextView mainViewHeaderText = null;
    
    ViewAnimator mainViewContainer = null;
    GridView mainViewApplicationsGridView = null;
    View mainViewWallpaperHolder = null;
    TextView mainViewApplicationTitle = null;
    
    ImageButton mainGotoFarmListButton = null;
    ImageButton mainGlobalOptionsButton = null;
    ImageButton mainRefreshAppsButton = null;
    ImageButton mainHelpButton = null;
    
    ListView farmList = null;
    Point farmListTouchPoint = null;
    FarmListAdapter farmListAdapter = null;
    GestureDetector gestureDetector = null;
    
    ListView locationList = null;
    ImageView locationListArrow = null;
    
    public DefaultController( Activity activity, IModel model )
    {
        this.activity = activity;
        this.resources = activity.getResources();
        this.model = model;
        
        this.initControllers();
        this.initListeners();
        
        GlobalSettingsController.updateWallpaper( activity, model );
        
        updateOrientation( false );
    }

//    public void hideSystemUi( View view )
//    {
//        try
//        {
//            Integer[] input = { new Integer( 0 ) };
//            Class cl = View.class;
//            Class[] par = new Class[1];
//            par[0] = Integer.TYPE;
//            Method mthd = cl.getMethod( "setSystemUiVisibility", par );
//            mthd.invoke( view, input );
//        }
//        catch ( Throwable e )
//        {
//            logger.debug( "Android SDK Version < 11" );
//        }
//    }

    private void initControllers()
    {
//        hideSystemUi( this.activity.findViewById( R.id.Main ) );
        
        this.mainViewContainer = (ViewAnimator) this.activity.findViewById( R.id.MainViewContainer );
        this.mainViewHeaderText = (TextView) this.activity.findViewById( R.id.MainViewHeaderText );
        this.mainViewApplicationsGridView = (GridView) this.activity.findViewById( R.id.ApplicationsGridView );
        this.mainViewWallpaperHolder = this.activity.findViewById( R.id.WallpaperHolder );
        this.mainViewApplicationTitle = (TextView) this.activity.findViewById( R.id.MainViewCapitalText );
        
        this.mainGotoFarmListButton = (ImageButton) this.activity.findViewById( R.id.GotoFarmListButton );
        this.mainGlobalOptionsButton = (ImageButton) this.activity.findViewById( R.id.GlobalOptionsButton );
        this.mainRefreshAppsButton = (ImageButton) this.activity.findViewById( R.id.RefreshApplicationButton );
        this.mainHelpButton = (ImageButton) this.activity.findViewById( R.id.HelpButton );
        
        this.rdpViewGallery = (Gallery) this.activity.findViewById( R.id.RdpViewGallery );
        
        this.farmList = (ListView) this.activity.findViewById( R.id.FarmList );
        this.farmListTouchPoint = new Point( Common.INVALID_VALUE, Common.INVALID_VALUE );
        this.gestureDetector = new GestureDetector(new SwipeDetector());
        this.farmListAdapter = new FarmListAdapter( this.activity, this.model.getFarmList() );
        
        this.locationList = (ListView) this.activity.findViewById( R.id.LocationList );
        this.locationListArrow = (ImageView) this.activity.findViewById( R.id.LocationListViewArrow );
    }
    
    private void initListeners()
    {
        this.model.setOnApplicationsListListener( applicationsListLoadListener );
        
        this.model.setOnRdpControlListener( rdpControlListener );
        
        this.mainGlobalOptionsButton.setOnClickListener( this.mainGlobalOptionsButtonListener );
        this.mainGotoFarmListButton.setOnClickListener( this.mainGotoFarmListButtonListener );
        this.mainRefreshAppsButton.setOnClickListener( this.mainRefreshAppsButtonListener );
        this.mainHelpButton.setOnClickListener( this.mainHelpButtonListener );
        this.mainViewApplicationsGridView.setOnItemClickListener( applicationGridItemListener );
        
        this.farmList.setOnItemClickListener( this.farmListItemListener );
        this.farmList.setAdapter( this.farmListAdapter );
        this.farmList.setOnTouchListener( farmListOnTouchListener );
        
        this.rdpViewGallery.setOnItemClickListener( rdpGalleryItemListener );
    }
    
    public void onExit() throws Throwable
    {
        activity = null;
        resources = null;
        model = null;
        
        dialogController = null;
        
        if ( progressDialog != null )
        {
            progressDialog.dismiss();
            progressDialog = null;
        }
        
        stopRdpViewInvalidator();
    }
    
    private void updateRdpInvalidatorState()
    {
        if ( rdpViewGallery.getCount() == 0 )
        {
            stopRdpViewInvalidator();
        }
        else
        {
            startRdpViewInvalidator();
        }
    }
    
    private void startRdpViewInvalidator()
    {
        if ( rdpViewGalleryInvalidator == null )
        {
            rdpViewGalleryInvalidator = new Thread( rdpGalleryUpdater );
            rdpViewGalleryInvalidator.start();
        }
    }
    
    private void stopRdpViewInvalidator()
    {
        if ( rdpViewGalleryInvalidator != null )
        {
            rdpViewGalleryInvalidator.interrupt();
            rdpViewGalleryInvalidator = null;
        }
    }
    
    public void startSplashAnimation( final long delay )
    {
        updateSplash();
        
        this.mainViewContainer.postDelayed( new Runnable()
        {
            public void run()
            {
                mainViewContainer.setDisplayedChild( MainViews.MAIN_VIEW );
            }
        }, delay );
    }
    
    private Runnable rdpGalleryUpdater = new Runnable()
    {
        public void run()
        {
            try
            {
                while ( !Thread.interrupted() )
                {
                    rdpViewGallery.postInvalidate();
                    Thread.sleep( 500 );
                }
            }
            catch ( Exception e ) { }
        }
    };
    
    private OnItemClickListener rdpGalleryItemListener = new OnItemClickListener()
    {
        public void onItemClick( AdapterView<?> parent, View view, int position, long id )
        {
            List<RdpThread> threads = model.getRdpThreadList();
            
            mainViewContainer.removeViews( MainViews.RDP_VIEW, mainViewContainer.getChildCount() - MainViews.RDP_VIEW );
            
            for ( RdpThread thread : threads )
            {
                if ( thread.getId() == id )
                {
                    if ( threads.get( position ) != null )
                    {
                        mainViewContainer.addView( thread.getRdpView() );
                        mainViewContainer.setDisplayedChild( MainViews.RDP_VIEW );
                        stopRdpViewInvalidator();
                        break;
                    }
                }
            }
        }
    };
    
    private OnItemClickListener applicationGridItemListener = new OnItemClickListener()
    {
        public void onItemClick( AdapterView<?> parent, View view, int position,
                long id )
        {
            logger.debug( "Application Grid was clicked" );
            
            if ( progressDialog != null )
            {
                progressDialog.dismiss();
                progressDialog = null;
            }
            
            ApplicationsGridAdapter adapter = (ApplicationsGridAdapter) parent.getAdapter();
            PNFarm farm = model.getConnectedFarm();
            if ( farm != null )
            {
                logger.debug( "Application Grid: farm is not null: [" + farm.getName() + "]" );
                
                PNApplication application = (PNApplication) adapter.getItem( position );
                
                String progressText = resources.getString( R.string.additional_text_connecting );
                progressText += " [ " + application.getName() + " ].";
                progressText += "\n" + resources.getString( R.string.additional_text_please_wait );
                
                progressDialog = ProgressDialog.show( activity, Common.EMTPY_STRING, progressText, true);
                
                logger.debug( "Application Grid: Progress dialog is showed. Start RDP connection" );
                
                PNLocation location = farm.getLocationByUuid( farm.getLastLocation() );
                model.startRdpConnection( application, farm, location );
            }
        }
    };
    
    private OnRdpControlListener rdpControlListener =  new OnRdpControlListener()
    {
        public void OnRdpViewCreated( final View view, final Bitmap desktop )
        {
            mainViewContainer.post( new Runnable()
            {
                public void run()
                {
                    int index = mainViewContainer.getChildCount();
                    mainViewContainer.addView( view );
                    mainViewContainer.setDisplayedChild( index );
                    
                    if ( progressDialog != null )
                    {
                        progressDialog.dismiss();
                        progressDialog = null;
                    }
                    
                    updateRdpViewGallery();
                    stopRdpViewInvalidator();
                }
            });
        }
        
        public void OnRdpViewBackClicked( View view )
        {
            mainViewContainer.post( new Runnable()
            {
                public void run()
                {
                    InputMethodManager imm = (InputMethodManager) activity.getSystemService( Context.INPUT_METHOD_SERVICE );
                    imm.hideSoftInputFromWindow( mainViewContainer.getWindowToken(), 0 );
                    
                    mainViewContainer.setDisplayedChild( MainViews.MAIN_VIEW );
                    
                    mainViewContainer.removeViews( MainViews.RDP_VIEW, mainViewContainer.getChildCount() - MainViews.RDP_VIEW );
                    
                    updateRdpViewGallery( );
                    startRdpViewInvalidator();
                }
            });
        }

        public void OnRdpConnectionStopped( final View view, final String reason )
        {
            mainViewContainer.post( new Runnable()
            {
                public void run()
                {
                    boolean progressEnabled = false; 
                    
                    if ( progressDialog != null )
                    {
                        progressEnabled = true;
                        progressDialog.dismiss();
                        progressDialog = null;
                    }
                    
                    if ( (mainViewContainer.getCurrentView() == view || view == null || progressEnabled) && !Common.isNullOrEmpty( reason ) )
                    {
                        (new AlertDialog.Builder( activity ))
                        .setMessage( reason )
                        .setPositiveButton( R.string.button_text_ok, null )
                        .create()
                        .show();
                    }
                    
                    if ( mainViewContainer.getCurrentView() == view || view == null )
                    {
                        if ( view != null )
                        {
                            InputMethodManager imm = (InputMethodManager) activity.getSystemService( Context.INPUT_METHOD_SERVICE );
                            if ( imm.isActive() )
                            {
                                imm.hideSoftInputFromWindow( mainViewContainer.getWindowToken(), 0 );
                            }
                            
                            mainViewContainer.setDisplayedChild( MainViews.MAIN_VIEW );
                            mainViewContainer.removeView( view );
                        }
                    }
                    
                    updateRdpViewGallery();
                    updateRdpInvalidatorState();
                }
            });
        }
    };
    
    private OnClickListener closeRdpListener = new OnClickListener()
    {
        public void onClick( View view )
        {
            long id = (Long)view.getTag();
            
            for ( RdpThread thread : model.getRdpThreadList() )
            {
                if ( thread.getId() == id )
                {
                    mainViewContainer.removeView( thread.getRdpView() );
                    thread.stopConnection();
                    updateRdpViewGallery();
                    break;
                }
            }
            
            updateRdpInvalidatorState();
        }
    };
    
    OnTouchListener farmListOnTouchListener = new OnTouchListener()
    {
        int findViewAtPosition(int y) 
        {
            int childCount = farmList.getChildCount();
            if (childCount > 0) 
            {
                for (int i = 0; i < childCount; i++) 
                {
                    View v = farmList.getChildAt(i);
                    if (y <= v.getBottom()) 
                    {
                        return farmList.getFirstVisiblePosition() + i;
                    }
                }
                return farmList.getFirstVisiblePosition() + childCount - 1;
            }
            return Common.INVALID_VALUE;
        }
        
        public boolean onTouch( View v, MotionEvent event )
        {
            int childPos = findViewAtPosition( (int)event.getY() );
            
            View view = farmList.getChildAt( childPos - farmList.getFirstVisiblePosition() );
            
            if ( view != null )
            {
                farmListTouchPoint.x = farmList.getLeft() + view.getLeft();
                farmListTouchPoint.y = farmList.getTop() + view.getTop();
                
                if ( gestureDetector.onTouchEvent( event ) )
                {
                    if ( childPos >= 0 && childPos != (farmList.getCount() - 1) )
                    {
                        View currentItem = farmList.getChildAt( childPos - farmList.getFirstVisiblePosition() );
                        
                        if ( currentItem != null )
                        {
                            ViewAnimator switcher = (ViewAnimator) currentItem.findViewById( R.id.ButtonSwitcher );
                            if ( switcher != null )
                            {
                                switcher.showNext();
                                return true;
                            }
                        }
                    }
                }
            }
            return false;
        }
    };
    
    private OnClickListener mainGlobalOptionsButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            new GlobalSettingsController( activity, model ).showDialog();
        }
    };
    
    private OnClickListener mainHelpButtonListener = new OnClickListener()
    {
        public void onClick( View view )
        {
            dialogController = new AboutDialog( activity );
            dialogController.showDialog();
            dialogController.setOnDismissListener( dialogControllerDismissListener );
        }
    };
    
    private OnClickListener mainRefreshAppsButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            clearApplicationList();
            
            final PNFarm farm = model.getConnectedFarm();
            
            if ( progressDialog != null )
            {
                progressDialog.dismiss();
                progressDialog = null;
            }
            
            if ( farm == null )
            {
                (new AlertDialog.Builder( activity ))
                .setMessage( R.string.additional_text_refresh_error )
                .setNeutralButton( R.string.button_text_ok, null )
                .create().show();
                return;
            }
            
            progressDialog = ProgressDialog.show( activity, Common.EMTPY_STRING, getProgressText( farm ), true);
            
            PNLocation location = farm.getLocationByUuid( farm.getLastLocation() );
            model.getApplicationList( farm, location );
        }
    };
    
    private OnClickListener farmListSettingsButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            View farmListOutside = activity.findViewById( R.id.FarmListView );
            farmListOutside.setVisibility( View.GONE );
            farmListOutside.setOnClickListener( null );
            
            FarmListItem item = ((FarmListItem)v.getTag());
            PNFarm farm = model.getFarmByUuid( item.getFarmUuid() );
            dialogController = new FarmSettingsController( model, activity, farm );
            dialogController.showDialog();
            
            if ( farm.isConnected() )
            {
                dialogController.setOnDismissListener( new OnDismissListener()
                {
                    public void onDismiss( DialogInterface dialog )
                    {
                        FarmSettingsController farmSettings = (FarmSettingsController)dialogController;
                        if ( !farmSettings.getFarm().isConnected() )
                        {
                            clearApplicationList();
                        }
                    }
                });
            }
            else
            {
                dialogController.setOnDismissListener( dialogControllerDismissListener );
            }
        }
    };
    
    private OnClickListener farmListDeleteButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            final FarmListItem item = (FarmListItem)v.getTag(); 
            
            final PNFarm farm = model.getFarmByUuid( item.getFarmUuid() );
            
            String question = resources.getString( R.string.additional_text_farm_deleting_msg1 );
            question +=  " [" + farm.getName() + "].";
            question += "\n" + resources.getString( R.string.additional_text_farm_deleting_msg2 );
            question += "\n" + resources.getString( R.string.additional_text_farm_deleting_msg3 );
            
            AlertDialog.Builder alertBuilder = new AlertDialog.Builder( activity );
            alertBuilder.setMessage( question )
            .setPositiveButton( R.string.button_text_yes, new DialogInterface.OnClickListener()
            {
                public void onClick( DialogInterface dialog, int which )
                {
                    PNFarm currentFarmConnected = model.getConnectedFarm();
                    
                    if ( currentFarmConnected != null && currentFarmConnected.getUuid().equals( item.getFarmUuid() ) )
                    {
                        currentFarmConnected.setConnected( false );
                        clearApplicationList();
                    }
                    model.deleteFarm( farm.getUuid() );
                    updateFarmList();
                }
            })
            .setNegativeButton( R.string.button_text_no, null);
            
            AlertDialog alert = alertBuilder.create();
            alert.show();
        }
    };
    
    private OnItemClickListener farmListItemListener = new OnItemClickListener()
    {
        public void onItemClick( AdapterView<?> parent, View view, int position,
                long id )
        {
            if ( position == (farmList.getCount() - 1) )
            {
                View farmListOutside = activity.findViewById( R.id.FarmListView );
                farmListOutside.setVisibility( View.GONE );
                farmListOutside.setOnClickListener( null );
                
                dialogController = new FarmSettingsController( model, activity, null );
                dialogController.setOnDismissListener( dialogControllerDismissListener );
                ((FarmSettingsController)dialogController).setPositiveButton( new OnClickListener()
                {
                    public void onClick( View view )
                    {
                        FarmSettingsController farmSettings = (FarmSettingsController) dialogController;
                        
                        PNFarm farm = farmSettings.getFarm();
                        if ( farmSettings.isNewFarm() && farm != null && farm.getLocationList().size() > 0 )
                        {
                            connectFarm( farm, farm.getLocation( 0 ) );
                        }
                    }
                });
                dialogController.showDialog();
            }
            else
            {
                final PNFarm farm = model.getFarmList().get( position );
                connectFarm( farm, null );
            }
        }
    };
    
    private void connectFarm( final PNFarm farm, PNLocation location )
    {
        if ( progressDialog != null )
        {
            progressDialog.dismiss();
            progressDialog = null;
        }
        
        if ( location == null && farm.getLocationsCount() <= 0 )
        {
            (new AlertDialog.Builder( activity ))
            .setMessage( R.string.additional_text_farm_no_locations )
            .setNeutralButton( R.string.button_text_ok, null )
            .create().show();
            return;
        }
        
        if ( location == null && Common.isNullOrEmpty( farm.getDefaultLocation() ) && farm.getLocationsCount() > 1 )
        {
            final int arrowPostitionOffset = 5;
            final int locationListOffset = -10;
            
            locationListArrow.setVisibility( View.VISIBLE );
            
            RelativeLayout.LayoutParams locationListArrowParams = (RelativeLayout.LayoutParams) locationListArrow.getLayoutParams();
            locationListArrowParams.topMargin = farmListTouchPoint.y + arrowPostitionOffset;
            locationListArrow.setLayoutParams( locationListArrowParams );
            
            LinearLayout.LayoutParams locationListParams = (LinearLayout.LayoutParams) locationList.getLayoutParams();
            locationListParams.topMargin = farmListTouchPoint.y + locationListOffset;
            locationList.setLayoutParams( locationListParams );
            
            final View locationListOutside = activity.findViewById( R.id.LocationListView );
            
            locationList.setAdapter( new LocationListAdapter( activity, farm.getLocationList() ) );
            
            locationList.setOnItemClickListener( new OnItemClickListener()
            {
                public void onItemClick( AdapterView<?> parent, View view,
                        int position, long id )
                {
                    progressDialog = ProgressDialog.show( activity, Common.EMTPY_STRING, getProgressText( farm ), true);
                    
                    clearApplicationList();
                    
                    locationListOutside.setVisibility(View.GONE);
                    locationListOutside.setOnClickListener( null );
                    
                    View farmListOutside = activity.findViewById( R.id.FarmListView );
                    farmListOutside.setVisibility( View.GONE );
                    farmListOutside.setOnClickListener( null );
                    
                    locationListArrow.setVisibility( View.GONE );
                    
                    PNLocation location = farm.getLocation( position ); 
                    model.getApplicationList( farm, location );
                }
            });
            
            locationListOutside.setOnClickListener( new OnClickListener()
            {
                public void onClick( View v )
                {
                    locationListArrow.setVisibility( View.GONE );
                    locationListOutside.setVisibility(View.GONE);
                    locationListOutside.setOnClickListener( null );
                }
            });
            locationListOutside.setVisibility( View.VISIBLE );
        }
        else 
        {
            progressDialog = ProgressDialog.show( activity, Common.EMTPY_STRING, getProgressText( farm ), true);
            
            // Clear application list
            clearApplicationList();
            
            View farmListOutside = activity.findViewById( R.id.FarmListView );
            farmListOutside.setVisibility( View.GONE );
            farmListOutside.setOnClickListener( null );
            
            if ( location == null )
            {
                location = farm.getLocationByUuid( farm.getDefaultLocation() );
            }
            
            if ( location == null )
            {
                location = farm.getLocation( 0 );
            }
            
            model.getApplicationList( farm, location );
        }
    }
    
    private OnClickListener mainGotoFarmListButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            updateFarmList();
            
            final View farmListOutside = activity.findViewById( R.id.FarmListView );
            
            farmListOutside.setOnClickListener( new OnClickListener()
            {
                public void onClick( View v )
                {
                    farmListOutside.setOnClickListener( null );
                    farmListOutside.setVisibility( View.GONE );
                }
            });
            farmListOutside.setVisibility( View.VISIBLE );
        }
    };
    
    private OnApplicationsListLoadListener applicationsListLoadListener = new OnApplicationsListLoadListener()
    {
        public void OnApplicationsListLoaded( final OnApplicationsListLoadListener.Arguments arg )
        {
            mainViewApplicationsGridView.post( new Runnable()
            {
                public void run()
                {
                    List<PNApplication> applicationsList = arg.applications;
                    
                    PNFarm farm = arg.farm;
                    PNLocation location = arg.location;
                    
                    String appListMessage = arg.message;
                    int errorCode = arg.errorCode;
                    
                    if ( progressDialog != null )
                    {
                        progressDialog.dismiss();
                        progressDialog = null;
                    }
                    
                    if ( errorCode == PNException.ERROR_BAD_USERNAME_OR_PASSWORD ||
                         errorCode == PNException.ERROR_INCORRECT_PARAMETER )
                    {
                        dialogController = new CredentialController( activity, farm, location, appListMessage );
                        ((CredentialController)dialogController).setPositiveButton( new OnClickListener()
                        {
                            public void onClick( View view )
                            {
                                PNFarm credsFarm = ((CredentialController)dialogController).getFarm();
                                PNLocation location = ((CredentialController)dialogController).getLocation();
                                connectFarm( credsFarm, location );
                            }
                        });
                        
                        ((CredentialController)dialogController).setNegativeButton( new OnClickListener()
                        {
                            public void onClick( View view )
                            {
                                CredentialController creds = (CredentialController) dialogController;
                                if ( Common.isNullOrEmpty( creds.getFarm().getName() ) )
                                {
                                    model.deleteFarm( creds.getFarm().getUuid() );
                                }
                            }
                        });
                        
                        dialogController.setOnDismissListener( dialogControllerDismissListener );
                        
                        dialogController.showDialog();
                        
                        return;
                    }
                    
                    if ( appListMessage != null )
                    {
                        if ( Common.isNullOrEmpty( farm.getName() ) )
                        {
                            model.deleteFarm( farm.getUuid() );
                        }
                        
                        (new AlertDialog.Builder(activity))
                        .setMessage(appListMessage)
                        .setNeutralButton( R.string.button_text_ok, null)
                        .create()
                        .show();
                        return;
                    }
            
                    updateApplicationList( applicationsList, farm.getName(), location.getName() );
                }
            });
        }
    };
    
    private OnDismissListener dialogControllerDismissListener = new OnDismissListener()
    {
        public void onDismiss( DialogInterface dialog )
        {
            dialogController = null;
        }
    };
    
    private void updateRdpViewGallery()
    {
        List<RdpThread> rdpList = model.getRdpThreadList();

        ArrayList<RdpViewGalleryAdapter.RdpViewGalleryItem> list = new ArrayList<RdpViewGalleryAdapter.RdpViewGalleryItem>();

        for( RdpThread thread : rdpList )
        {
            if ( !thread.isInterrupted() && thread.getApplicationName() != null )
            {
                RdpViewGalleryAdapter.RdpViewGalleryItem newItem = new RdpViewGalleryAdapter.RdpViewGalleryItem(
                        thread.getId(), thread.getApplicationName(), thread.getRdpDesktop() );
                list.add( newItem );
            }
        }

        int pos = rdpViewGallery.getSelectedItemPosition();
        
        rdpViewGallery.setAdapter( new RdpViewGalleryAdapter( activity, list, closeRdpListener ) );
        
        if ( pos >= 0 && pos < rdpViewGallery.getCount() )
        {
            rdpViewGallery.setSelection( pos );
        }
    }

    private void updateFarmList()
    {
        this.farmListAdapter = new FarmListAdapter( this.activity, this.model.getFarmList() );
        this.farmListAdapter.setOnDeleteClickListener( farmListDeleteButtonListener );
        this.farmListAdapter.setOnSettingsClickListener( farmListSettingsButtonListener );
        this.farmList.setAdapter( this.farmListAdapter );
    }
    
    public void updateOrientation( boolean isLandscape )
    {
        updateSplash();
        
        RelativeLayout.LayoutParams headerParams = new RelativeLayout.LayoutParams( RelativeLayout.LayoutParams.WRAP_CONTENT, RelativeLayout.LayoutParams.MATCH_PARENT );
        
        int screenWidth = activity.getWindowManager().getDefaultDisplay().getWidth(); 
        
        if ( screenWidth < resources.getDimension( R.dimen.small_screen_width ) )
        {
            headerParams.addRule( RelativeLayout.RIGHT_OF, R.id.GotoFarmListButton );
            headerParams.leftMargin = 25;
        }
        else
        {
            headerParams.addRule( RelativeLayout.CENTER_IN_PARENT, RelativeLayout.TRUE );
        }
        
        mainViewHeaderText.setLayoutParams( headerParams );
        
        if ( dialogController != null && dialogController.isShowed() )
        {
            dialogController.updateOrientation( isLandscape );
        }
        
        List<View> rdpViewsList = model.getRdpViewList();
        if ( rdpViewsList != null && rdpViewsList.size() > 0 )
        {
            for ( View view : rdpViewsList )
            {
                if ( view == null ) continue;
                
                RelativeLayout connBar = (RelativeLayout) view.findViewById( R.id.ConnectionBar );
                RelativeLayout.LayoutParams params = (RelativeLayout.LayoutParams)connBar.getLayoutParams();
                
                int connectionBarWidth = (int)resources.getDimension( R.dimen.connection_bar_width );
                if ( connectionBarWidth > screenWidth )
                {
                    connectionBarWidth = screenWidth;
                    params.leftMargin = 0;
                }
                else
                {
                    int x = ((screenWidth - connectionBarWidth) >> 1 );
                    params.leftMargin = x;
                }
                params.width = connectionBarWidth;
                connBar.setLayoutParams( params );
            }
        }
        
        GlobalSettingsController.updateWallpaper( activity, model );
        
        updateListViews();
    }
    
    private void updateListViews()
    {
        LinearLayout.LayoutParams farmListParams = (LinearLayout.LayoutParams)farmList.getLayoutParams();
        LinearLayout.LayoutParams locationListParams = (LinearLayout.LayoutParams)locationList.getLayoutParams();
        RelativeLayout.LayoutParams locationListArrowParams = (RelativeLayout.LayoutParams) locationListArrow.getLayoutParams();
        
        if ( activity.getWindowManager().getDefaultDisplay().getWidth() < resources.getDimension( R.dimen.small_screen_width ) )
        {
            farmListParams.width = (int) resources.getDimension( R.dimen.farm_list_small_width );
            
            locationListParams.width = (int) resources.getDimension( R.dimen.location_list_small_width );
            locationListParams.leftMargin = (int) resources.getDimension( R.dimen.location_list_small_left_margin );
            
            locationListArrowParams.leftMargin = (int) resources.getDimension( R.dimen.location_arrow_small_left_margin );
        }
        else
        {
            farmListParams.width = (int) resources.getDimension( R.dimen.farm_list_default_width );
            
            locationListParams.width = (int) resources.getDimension( R.dimen.location_list_default_width );
            locationListParams.leftMargin = (int) resources.getDimension( R.dimen.location_list_default_left_margin );
            
            locationListArrowParams.leftMargin = (int) resources.getDimension( R.dimen.location_arrow_default_left_margin );
        }
        
        farmList.setLayoutParams( farmListParams );
        locationList.setLayoutParams( locationListParams );
        locationListArrow.setLayoutParams( locationListArrowParams );
    }
    
    private String getProgressText( PNFarm farm )
    {
        
        String progressText = resources.getString( R.string.additional_text_connecting );
        
        if ( Common.isNullOrEmpty( farm.getName() ) ) 
        {
            progressText += " " + resources.getString( R.string.additional_text_new_farm ); 
        }
        else
        {
            progressText += " [ " + farm.getName() + " ].";
        }
        
        progressText += "\n" + resources.getString( R.string.additional_text_please_wait );
        
        return progressText;
    }
    
    private void updateApplicationList( List<PNApplication> appList, String farm, String location )
    {
        this.mainViewApplicationsGridView.setAdapter( new ApplicationsGridAdapter( activity, appList ) );
        this.mainViewApplicationTitle.setText( resources.getString( R.string.additional_text_applications) + " [ " + farm + " : " + location + " ]" );
    }
    
    private void updateSplash()
    {
        if ( this.mainViewContainer.getCurrentView() == activity.findViewById( R.id.SplashView ))
        {
            ImageView splash = (ImageView) activity.findViewById( R.id.SplashView ); 
        
            int width = activity.getWindowManager().getDefaultDisplay().getWidth();
            int height = activity.getWindowManager().getDefaultDisplay().getHeight();
            
            if ( width == 480 && height == 800 )
            {
                splash.setImageResource( R.drawable.splash_480x800 );
            }
            else if ( width == 800 && height == 480 )
            {
                splash.setImageResource( R.drawable.splash_800x480 );
            }
            else if ( width == 1024 && height == 600 )
            {
                splash.setImageResource( R.drawable.splash_1024x600 );
            }
            else if ( width == 600 && height == 1024 )
            {
                splash.setImageResource( R.drawable.splash_600x1024 );
            }
            else if ( width == 1280 && height == 800 )
            {
                splash.setImageResource( R.drawable.splash_1280x800 );
            }
            else if ( width == 800 && height == 1280 )
            {
                splash.setImageResource( R.drawable.splash_800x1280 );
            }
            
            splash.postInvalidate();
        }
    }
    
    private void clearApplicationList()
    {
        this.mainViewApplicationsGridView.setAdapter( new ApplicationsGridAdapter( activity, null ) );
        this.mainViewApplicationTitle.setText( R.string.additional_text_applications );
    }
    
    public static class MainViews
    {
        public static final int SPLASH_VIEW                 = 0;
        public static final int MAIN_VIEW                   = 1;
        public static final int RDP_VIEW                    = 2;
    }
}
