package quest.android.appportal.controller;

import quest.android.appportal.R;
import android.app.Activity;
import android.app.Dialog;
import android.content.DialogInterface.OnDismissListener;
import android.content.res.Resources;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.view.Window;
import android.widget.FrameLayout;

public abstract class DialogController
{
    protected Activity activity = null;
    protected Resources resources = null;
    protected Dialog dialog = null;
    protected OnDismissListener dismissListener = null;
    
    OnClickListener positiveListener = null;
    OnClickListener negativeListener = null;
    
    protected DialogController( Activity activity)
    {
        this.activity = activity;
        this.resources = activity.getResources();
        
        dialog = new Dialog( activity );        
        dialog.setCancelable( true );
        dialog.requestWindowFeature( Window.FEATURE_NO_TITLE );
    }
    
    protected void setContentView( int layoutResID )
    {
        dialog.setContentView( layoutResID );
        dialog.findViewById( R.id.DialogMain ).setLayoutParams( new FrameLayout.LayoutParams( 
                (int) resources.getDimension( R.dimen.dialog_default_width ), 
                (int) resources.getDimension( R.dimen.dialog_default_height ) ) );
    }
    
    public abstract void showDialog();
    protected abstract void updateDialog();
    
    protected void updateSettings() { };
    
    public void setOnDismissListener( OnDismissListener listener ) 
    {
        dismissListener = listener;
        if ( dialog != null )
        {
            dialog.setOnDismissListener( listener );
        }
    }
    
    public void updateOrientation( boolean isLandscape )
    {
        View view = dialog.findViewById( R.id.DialogMain );
        ViewGroup.LayoutParams viewParams = view.getLayoutParams();

        if ( activity.getWindowManager().getDefaultDisplay().getWidth() <= 440 )
        {
            viewParams.width = ViewGroup.LayoutParams.MATCH_PARENT;
        }
        else if ( activity.getWindowManager().getDefaultDisplay().getWidth() <= resources.getDimension( R.dimen.small_screen_width ) )
        {
            viewParams.width = (int) resources.getDimension( R.dimen.dialog_small_width );
        }
        else
        {
            viewParams.width = (int) resources.getDimension( R.dimen.dialog_default_width );
        }
        
        if ( activity.getWindowManager().getDefaultDisplay().getHeight() <= resources.getDimension( R.dimen.small_screen_height ) )
        {
            viewParams.height = ViewGroup.LayoutParams.MATCH_PARENT;
        }
        else
        {
            viewParams.height = (int) resources.getDimension( R.dimen.dialog_default_height );
        }
        
        view.setLayoutParams( viewParams );
        view.invalidate();
    }
    
    public boolean isShowed( )
    {
        return (dialog != null);
    }
    
    public void setPositiveButton( OnClickListener listener )
    {
        positiveListener = listener;
    }
    
    public void setNegativeButton( OnClickListener listener )
    {
        negativeListener = listener;
    }
}
