package quest.android.appportal.controller;

import org.apache.log4j.Logger;

import quest.android.appportal.R;
import quest.android.appportal.common.Common;
import quest.android.appportal.farm.PNFarm;
import android.app.Activity;
import android.content.DialogInterface.OnDismissListener;
import android.view.View;
import android.view.ViewGroup;
import android.view.View.OnClickListener;
import android.widget.ArrayAdapter;
import android.widget.ListView;

public class DisplaySettingsController extends DialogController
{
    private final Logger logger = Logger.getLogger( DisplaySettingsController.class );
    
    protected PNFarm farm = null;
    
    public DisplaySettingsController( PNFarm farm, Activity activity )
    {
        super( activity );
        this.farm = farm;
    }
    
    public void showDialog()
    {
        this.setContentView( R.layout.display_settings );
        
        dialog.findViewById( R.id.DisplaySettingsSaveButton ).setOnClickListener( saveButtonListener );
        dialog.findViewById( R.id.DisplaySettingsCancelButton ).setOnClickListener( cancelButtonListener );
        
        updateDialog();
        
        this.updateOrientation( true );
        
        dialog.show();
    }
    
    protected void updateDialog()
    {
        ListView colorList = (ListView)dialog.findViewById( R.id.ColorList );
        colorList.setChoiceMode( ListView.CHOICE_MODE_SINGLE );
        colorList.setFocusable( false );
        colorList.setAdapter( new ArrayAdapter<String>( activity, R.layout.checked_list_item, 
               PNFarm.COLOR_DEPTH_LIST  ) );
        
        int rows = PNFarm.COLOR_DEPTH_LIST.length;
        ViewGroup.LayoutParams colorListParams = colorList.getLayoutParams();
        colorListParams.height = rows * 46 // 35px is items height
                                    + (rows - 1) * colorList.getDividerHeight() 
                                    + colorList.getPaddingTop() 
                                    + colorList.getPaddingBottom();
        colorList.setLayoutParams( colorListParams );
        
        int colorDepth = Integer.parseInt( farm.getColorDepth() );
        switch(colorDepth)
        {
            case 8:
            case 256:
                colorList.setItemChecked( 0, true );
                break;
            case 15:
                colorList.setItemChecked( 1, true );
                break;
            case 16:
                colorList.setItemChecked( 2, true );
                break;
            case 24:
                colorList.setItemChecked( 3, true );
                break;
            case 32:
                colorList.setItemChecked( 4, true );
                break;
        }
        
        ListView resolutionList = (ListView)dialog.findViewById( R.id.ResolutionList );
        resolutionList.setChoiceMode( ListView.CHOICE_MODE_SINGLE );
        resolutionList.setFocusable( false );
        resolutionList.setAdapter( new ArrayAdapter<String>( activity, R.layout.checked_list_item, 
                PNFarm.DISPLAY_RESOLUTIONS ) );
        
        rows = PNFarm.DISPLAY_RESOLUTIONS.length;
        ViewGroup.LayoutParams resolutionListParams = resolutionList.getLayoutParams();
        resolutionListParams.height = rows * 46 // 46px is items height
                                    + (rows - 1) * resolutionList.getDividerHeight() 
                                    + resolutionList.getPaddingTop() 
                                    + resolutionList.getPaddingBottom();
        resolutionList.setLayoutParams( resolutionListParams );
        
        resolutionList.setItemChecked( Common.parseIntSafe( farm.getResolution() ), true );
    }
    
    protected void updateSettings()
    {
        ListView colorList = (ListView)dialog.findViewById( R.id.ColorList );
        int pos = colorList.getCheckedItemPosition();
        switch(pos)
        {
            case 0:
                farm.setColorDepth( PNFarm.COLOR_DEPTH_8BIT );
                break;
            case 1:
                farm.setColorDepth( PNFarm.COLOR_DEPTH_15BIT );
                break;
            case 2:
                farm.setColorDepth( PNFarm.COLOR_DEPTH_16BIT );
                break;
            case 3:
                farm.setColorDepth( PNFarm.COLOR_DEPTH_24BIT );
                break;
            case 4:
                farm.setColorDepth( PNFarm.COLOR_DEPTH_32BIT );
                break;
        }
        
        ListView resolutionList = (ListView)dialog.findViewById( R.id.ResolutionList );
        pos = resolutionList.getCheckedItemPosition();
        
        int width = activity.getWindowManager().getDefaultDisplay().getWidth();
        int height = activity.getWindowManager().getDefaultDisplay().getHeight();
        // HOTFIX: honeycomb has another screen resolution then Window Manager reports.
        int barHeight = height - activity.findViewById( R.id.Main ).getBottom();
        
        if ( width < height )
        {
            int tmp = width;
            width = height;
            height = tmp;
        }
        
        farm.setResolution( "" + pos );
        
        if (pos == 0 )
        {
            logger.debug("Screen Resolution: " + PNFarm.DISPLAY_RESOLUTIONS[pos] );
            
            farm.setDesktopWidth( "" + height );
            farm.setDesktopHeight( "" + (width - barHeight) );
        }
        else if ( pos == 1 )
        {
            logger.debug("Screen Resolution: " + PNFarm.DISPLAY_RESOLUTIONS[pos] );
            
            farm.setDesktopWidth( "" + width );
            farm.setDesktopHeight( "" + (height - barHeight) );
        }
        else if ( pos >= 2 && pos < PNFarm.DISPLAY_RESOLUTIONS.length )
        {
            String[] displaySizes = PNFarm.DISPLAY_RESOLUTIONS[pos].split( " " );
        
            logger.debug("Screen Resolution: " + PNFarm.DISPLAY_RESOLUTIONS[pos] + " selected position: " + pos + " width=" + displaySizes[0] + " height=" + displaySizes[2]);
        
            farm.setDesktopWidth( displaySizes[0] );
            farm.setDesktopHeight( displaySizes[2] );
        }
    }
    
    private OnClickListener saveButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            updateSettings();
            dialog.dismiss();
        }
    };
    
    private OnClickListener cancelButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            dialog.dismiss();
        }
    }; 
    
    public void setOnDismissListener( OnDismissListener listener )
    {
        if ( dialog != null )
        {
            dialog.setOnDismissListener( listener );
        }
    }
}
