package quest.android.appportal.controller;

import quest.android.appportal.R;
import quest.android.appportal.farm.PNFarm;
import android.app.Activity;
import android.util.SparseBooleanArray;
import android.view.View;
import android.view.ViewGroup;
import android.view.View.OnClickListener;
import android.widget.ArrayAdapter;
import android.widget.ListView;

public class ExperienceSettingsController extends DialogController
{
    private PNFarm farm = null;
    
    public ExperienceSettingsController( PNFarm farm, Activity activity )
    {
        super( activity );
        this.farm = farm;
    }
    
    public void showDialog()
    {
        this.setContentView( R.layout.experience_settings );
        
        dialog.findViewById( R.id.UserSettingsSaveButton ).setOnClickListener( saveButtonListener );
        dialog.findViewById( R.id.UserSettingsCancelButton ).setOnClickListener( cancelButtonListener );
        
        updateDialog();
        this.updateOrientation( true );
        
        dialog.show();
    }
    
    protected void updateDialog()
    {
        ListView performanceList = (ListView)dialog.findViewById( R.id.PerformanceList );
        performanceList.setChoiceMode( ListView.CHOICE_MODE_MULTIPLE );
        performanceList.setFocusable( false );
        performanceList.setAdapter( new ArrayAdapter<String>( activity, R.layout.checked_list_item, 
               resources.getStringArray( R.array.PerformanceOptions ) ) );
        
        int rows = performanceList.getCount();
        ViewGroup.LayoutParams colorListParams = performanceList.getLayoutParams();
        colorListParams.height = rows * 46 // 46px is items height
                                    + (rows - 1) * performanceList.getDividerHeight() 
                                    + performanceList.getPaddingTop() 
                                    + performanceList.getPaddingBottom();
        performanceList.setLayoutParams( colorListParams );
        
        
        ListView eopList = (ListView)dialog.findViewById( R.id.EopList );
        eopList.setChoiceMode( ListView.CHOICE_MODE_MULTIPLE );
        eopList.setFocusable( false );
        eopList.setAdapter( new ArrayAdapter<String>( activity, R.layout.checked_list_item, 
                resources.getStringArray( R.array.EopOptions ) ) );
        
        rows = eopList.getCount();
        ViewGroup.LayoutParams eopListParams = eopList.getLayoutParams();
        eopListParams.height = rows * 46 // 46px is items height
                                    + (rows - 1) * eopList.getDividerHeight() 
                                    + eopList.getPaddingTop() 
                                    + eopList.getPaddingBottom();
        eopList.setLayoutParams( eopListParams );
        
        
        performanceList.setItemChecked( 0, farm.isWallpaperEnabled() );
        performanceList.setItemChecked( 1, farm.isFontSmoothingEnabled() );
        performanceList.setItemChecked( 2, farm.isThemesEnabled() );
        performanceList.setItemChecked( 3, farm.isFullWindowDragEnabled() );
        performanceList.setItemChecked( 4, farm.isAnimationEnabled() );
        performanceList.setItemChecked( 5, farm.isBitmapCachingEnabled() );
        
        eopList.setItemChecked( 0, farm.isBitmapAccelerationEnabled() );
    }
    
    protected void updateSettings()
    {
        ListView performanceList = (ListView)dialog.findViewById( R.id.PerformanceList );
        ListView eopList = (ListView)dialog.findViewById( R.id.EopList );
        
        SparseBooleanArray performanceArray = performanceList.getCheckedItemPositions();
        farm.setWallpaperEnabled( performanceArray.valueAt( 0 ) );
        farm.setFontSmoothingEnabled( performanceArray.valueAt( 1 ) );
        farm.setThemesEnabled( performanceArray.valueAt( 2 ) );
        farm.setFullWindowDragEnabled( performanceArray.valueAt( 3 ) );
        farm.setAnimationEnabled( performanceArray.valueAt( 4 ) );
        farm.setBitmapCachingEnabled( performanceArray.valueAt( 5 ) );
        
        SparseBooleanArray eopArray = eopList.getCheckedItemPositions();
        farm.setBitmapAccelerationEnabled( eopArray.valueAt( 0 ) );
    }
    
    private OnClickListener saveButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            updateSettings();
            dialog.dismiss();
        }
    };
    
    private OnClickListener cancelButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            dialog.dismiss();
        }
    };
}
