package quest.android.appportal.controller;

import java.util.ArrayList;

import org.apache.log4j.Logger;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.DialogInterface.OnDismissListener;
import android.graphics.Typeface;
import android.view.View;
import android.view.ViewGroup;
import android.view.View.OnClickListener;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.AdapterView.OnItemClickListener;
import android.widget.TextView;
import quest.android.appportal.R;
import quest.android.appportal.common.Common;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNFarmException;
import quest.android.appportal.farm.PNLocation;
import quest.android.appportal.model.IModel;

public class FarmSettingsController extends DialogController
{
    private Logger logger = Logger.getLogger( FarmSettingsController.class );
    
    private IModel model = null;
    private PNFarm farm = null;
    
    private boolean isNewFarm = false;
    private boolean advancedMode = false;
    
    private DialogController additionalDlg = null;
    
    public FarmSettingsController( IModel model, Activity activity, PNFarm farm )
    {
        super( activity );
        this.model = model;
        
        this.isNewFarm = (farm == null); 
        
        if ( isNewFarm )
        {
            this.farm = new PNFarm( Common.EMTPY_STRING );
            if ( isNewFarm ) setNewFarmDefaults( this.farm );
        }
        else
        {
            this.farm = farm.clone();
        }
        
        this.setContentView( R.layout.farm_settings );
        
    }
    
    public void showDialog()
    {
        logger.debug( "Farm settings showing." );
        
        dialog.findViewById( R.id.FarmSettingsSaveButton ).setOnClickListener( saveButtonListener );
        dialog.findViewById( R.id.FarmSettingsCancelButton ).setOnClickListener( cancelButtonListener );
        
        updateDialog();
        this.updateOrientation( true );
        
        dialog.show();
    }
    
    protected void updateDialog()
    {
        View basicView = dialog.findViewById( R.id.FarmSettingsBasicView );
        View advancedView1 = dialog.findViewById( R.id.FarmSettingsAdvancedView1 );
        View advancedView2 = dialog.findViewById( R.id.FarmSettingsAdvancedView2 );
        Button advancedButton = (Button) dialog.findViewById( R.id.FarmSettingsModeButton );
        
        TextView header = (TextView) dialog.findViewById( R.id.FarmSettingsHeaderText );
        View deleteButton = dialog.findViewById( R.id.FarmSettingsDeleteButton );
        Button saveButton = (Button) dialog.findViewById( R.id.FarmSettingsSaveButton );
        EditText farmName   = (EditText)dialog.findViewById( R.id.FarmSettingsFarmName );
        EditText username   = (EditText)dialog.findViewById( R.id.FarmSettingsUsername );
        EditText password   = (EditText)dialog.findViewById( R.id.FarmSettingsPassword );
        EditText domain     = (EditText)dialog.findViewById( R.id.FarmSettingsDomain );
        CheckBox saveCreds  = (CheckBox)dialog.findViewById( R.id.FarmSettingsSaveCreds );
        ListView locations  = (ListView)dialog.findViewById( R.id.FarmSettingsLocationList );
        ListView audioList  = (ListView)dialog.findViewById( R.id.FarmSettingsAudioList );
        Button displaySettingsButton  = (Button) dialog.findViewById( R.id.FarmSettingsDisplaySettingsButton );
        Button userSettingsButton  = (Button) dialog.findViewById( R.id.FarmSettingsExperienceSettingsButton );
        
        if ( isNewFarm )
        {
            header.setText( R.string.farm_settings_dialog_header_add );
            advancedButton.setVisibility( View.VISIBLE );
            deleteButton.setVisibility( View.GONE );
            
            if ( advancedMode )
            {
                advancedButton.setText( R.string.farm_settings_dialog_basic );
                
                basicView.setVisibility( View.GONE );
                advancedView1.setVisibility( View.VISIBLE );
                advancedView2.setVisibility( View.VISIBLE );
            }
            else
            {
                advancedButton.setText( R.string.farm_settings_dialog_advanced );
                
                basicView.setVisibility( View.VISIBLE );
                advancedView1.setVisibility( View.GONE );
                advancedView2.setVisibility( View.GONE );
            }
            
            saveButton.setText( R.string.button_text_connect );
            
            advancedButton.setOnClickListener( new OnClickListener()
            {
                public void onClick( View view )
                {
                    advancedMode = !advancedMode;
                    
                    View basicView = dialog.findViewById( R.id.FarmSettingsBasicView );
                    View advancedView1 = dialog.findViewById( R.id.FarmSettingsAdvancedView1 );
                    View advancedView2 = dialog.findViewById( R.id.FarmSettingsAdvancedView2 );
                    Button advancedButton = (Button) dialog.findViewById( R.id.FarmSettingsModeButton );
                    
                    if ( advancedMode )
                    {
                        advancedButton.setText( R.string.farm_settings_dialog_basic );
                        
                        basicView.setVisibility( View.GONE );
                        advancedView1.setVisibility( View.VISIBLE );
                        advancedView2.setVisibility( View.VISIBLE );
                        
                        EditText farmName   = (EditText)dialog.findViewById( R.id.FarmSettingsFarmName );
                        farmName.requestFocus();
                    }
                    else
                    {
                        advancedButton.setText( R.string.farm_settings_dialog_advanced );
                        
                        basicView.setVisibility( View.VISIBLE );
                        advancedView1.setVisibility( View.GONE );
                        advancedView2.setVisibility( View.GONE );
                        
                        EditText brokerName   = (EditText)dialog.findViewById( R.id.FarmSettingsBrokerName );
                        brokerName.requestFocus();
                    }
                }
            });
        }
        
        farmName.setText( farm.getName() );
        username.setText( farm.getUsername() );
        password.setText( farm.getPassword() );
        password.setTypeface( Typeface.DEFAULT );
        domain.setText( farm.getDomain() );
        saveCreds.setChecked( farm.isSaveCredentials() );
        
        ArrayList<String> list = new ArrayList<String>();
        
        for( int i = 0; i < farm.getLocationsCount(); ++i )
        {
            list.add( farm.getLocation( i ).getName() );
        }
        list.add( resources.getString( R.string.farm_settings_dialog_add_location ) );
        
        locations.setAdapter( new ArrayAdapter<String>(activity, R.layout.farm_settings_location_list_item, list) );
        
        int rows = list.size();
        ViewGroup.LayoutParams locationListParams = locations.getLayoutParams();
        locationListParams.height = rows * 35 // 35px is items height
                                    + (rows - 1) * locations.getDividerHeight() 
                                    + locations.getPaddingTop() 
                                    + locations.getPaddingBottom();
        locations.setLayoutParams( locationListParams );
        locations.setOnItemClickListener( new OnItemClickListener()
        {
            public void onItemClick( AdapterView<?> parent, View view, int position,
                    long id )
            {
                updateSettings();
                
                PNLocation location = null;
                if ( position < farm.getLocationsCount() )
                {
                    location = farm.getLocation( position );
                }
                
                additionalDlg = new LocationSettingsController( farm, activity, location );
                additionalDlg.showDialog();
                
                additionalDlg.setOnDismissListener( new OnDismissListener()
                {
                    public void onDismiss( DialogInterface dialog )
                    {
                        updateDialog();
                        additionalDlg = null;
                    }
                });
            }
        });
        
        displaySettingsButton.setOnClickListener( new OnClickListener()
        {
            public void onClick( View v )
            {
                updateSettings();
                
                additionalDlg = new DisplaySettingsController( farm, activity );
                additionalDlg.showDialog();
                
                additionalDlg.setOnDismissListener( new OnDismissListener()
                {
                    public void onDismiss( DialogInterface dialog )
                    {
                        updateDialog();
                        additionalDlg = null;
                    }
                } );
            }
        });
        
        String  displayButtonText = "";
        displayButtonText += PNFarm.DISPLAY_RESOLUTIONS[ Common.parseIntSafe( farm.getResolution() ) ];
        displayButtonText += ", ";
        displayButtonText += farm.getColorDepth() + " bit color";
        
        displaySettingsButton.setText( displayButtonText );
        
        userSettingsButton.setOnClickListener( new OnClickListener()
        {
            public void onClick( View v )
            {
                updateSettings();
                additionalDlg = new ExperienceSettingsController( farm, activity );
                additionalDlg.showDialog();
                
                additionalDlg.setOnDismissListener( new OnDismissListener()
                {
                    public void onDismiss( DialogInterface dialog )
                    {
                        additionalDlg = null;
                    }
                } );
            }
        });
        
        audioList.setFocusable( false );
        audioList.setChoiceMode( ListView.CHOICE_MODE_SINGLE );
        audioList.setAdapter( new ArrayAdapter<String>( activity, R.layout.checked_list_item, 
                resources.getStringArray( R.array.AudioOptions ) ) );
        rows = audioList.getCount();
        ViewGroup.LayoutParams audioListParams = audioList.getLayoutParams();
        audioListParams.height = rows * 46 // 46px is items height
                                    + (rows - 1) * audioList.getDividerHeight() 
                                    + audioList.getPaddingTop() 
                                    + audioList.getPaddingBottom();
        audioList.setLayoutParams( audioListParams );
        audioList.setItemChecked( Integer.parseInt( farm.getRedirectAudio() ), true );
        
        deleteButton.setOnClickListener( new OnClickListener()
        {
            public void onClick( View v )
            {
                model.deleteFarm( farm.getUuid() );
                dialog.dismiss();
            }
        });
    }
    
    protected void updateSettings()
    {
        EditText farmName   = (EditText)dialog.findViewById( R.id.FarmSettingsFarmName );
        EditText username   = (EditText)dialog.findViewById( R.id.FarmSettingsUsername );
        EditText password   = (EditText)dialog.findViewById( R.id.FarmSettingsPassword );
        EditText domain     = (EditText)dialog.findViewById( R.id.FarmSettingsDomain );
        CheckBox saveCreds  = (CheckBox)dialog.findViewById( R.id.FarmSettingsSaveCreds );
        ListView audioList  = (ListView)dialog.findViewById( R.id.FarmSettingsAudioList );
        
        farm.setName( farmName.getText().toString() );
        farm.setUsername( username.getText().toString() );
        farm.setPassword( password.getText().toString() );
        farm.setDomain( domain.getText().toString() );
        farm.setSaveCredentialsEnabled( saveCreds.isChecked() );
        if ( audioList.getCheckedItemPosition() != ListView.INVALID_POSITION )
        {
            farm.setRedirectAudio( "" + audioList.getCheckedItemPosition() );
        }
        else
        {
            farm.setRedirectAudio( "0" );
        }
    }

    private OnClickListener saveButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            if ( dialog == null ) return;
            
            if ( !isNewFarm || advancedMode )
            {
                EditText farmSettingsName = ((EditText)dialog.findViewById( R.id.FarmSettingsFarmName )); 
                String farmName = farmSettingsName.getText().toString();
                
                if ( !Common.isCorrectName( farmName ) )
                {
                    (new AlertDialog.Builder( activity ))
                    .setMessage( R.string.farm_settings_type_farm_name )
                    .setNeutralButton( R.string.button_text_ok, null )
                    .create().show();
                    
                    farmSettingsName.requestFocus();
                    
                    return;
                }
                
                if ( farm.getLocationsCount() == 0 )
                {
                    (new AlertDialog.Builder( activity ))
                    .setMessage( R.string.farm_settings_ask_add_location )
                    .setNeutralButton( R.string.button_text_ok, null )
                    .create().show();
                    dialog.findViewById( R.id.FarmSettingsLocationList ).requestFocus();
                    return;
                }
                
                updateSettings();
                
                if ( !isNewFarm )
                {
                    model.deleteFarm( farm.getUuid() );
                }
            }
            else
            {
                PNFarm farmNew = new PNFarm( Common.EMTPY_STRING );
                
                EditText brokerName = (EditText) dialog.findViewById( R.id.FarmSettingsBrokerName );
                EditText username   = (EditText)dialog.findViewById( R.id.FarmSettingsUsername );
                EditText password   = (EditText)dialog.findViewById( R.id.FarmSettingsPassword );
                EditText domain     = (EditText)dialog.findViewById( R.id.FarmSettingsDomain );
                CheckBox saveCreds  = (CheckBox)dialog.findViewById( R.id.FarmSettingsSaveCreds );
                
                farmNew.setUsername( username.getText().toString() );
                farmNew.setPassword( password.getText().toString() );
                farmNew.setDomain( domain.getText().toString() );
                farmNew.setSaveCredentialsEnabled( saveCreds.isChecked() );
                
                PNLocation location = new PNLocation( "Location 1" );
                
                String server = brokerName.getText().toString();
                
                if ( server.toUpperCase().startsWith( PNLocation.HTTPS_PROTO ) )
                {
                    location.setHttpsProtocol( true );
                    location.setTcpPort( "443" );
                    server = server.substring( 8 );
                }
                else if ( server.toUpperCase().startsWith( PNLocation.HTTP_PROTO ) )
                {
                    location.setHttpsProtocol( false );
                    location.setTcpPort( "8080" );
                    server = server.substring( 7 );
                }
                
                if ( server.contains( ":" ) )
                {
                    String[] splitted = server.split( ":" );
                    if ( splitted.length > 0)
                    {
                        server = splitted[0];
                        location.setTcpPort( splitted[1] );
                    }
                }   
                
                location.addServer( server );
                
                try
                {
                    farmNew.addLocation( location );
                }
                catch( PNFarmException e )
                {
                    logger.error( "Cannot add new location: " + e.getMessage(), e );
                }
                
                setNewFarmDefaults( farmNew );
                
                farm = farmNew;
            }
            
            model.addFarm( farm );
            model.sortFarmsByName();
            dialog.dismiss();
            
            if ( positiveListener != null )
            {
                positiveListener.onClick( v );
            }
        }
    };
    
    private OnClickListener cancelButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            dialog.dismiss();
        }
    };
    
    public void updateOrientation( boolean isLandscape )
    {
        super.updateOrientation( isLandscape );
        
        if ( additionalDlg != null && additionalDlg.isShowed() )
        {
            additionalDlg.updateOrientation( isLandscape );
        }
    }
    
    public PNFarm getFarm()
    {
        return farm;
    }
    
    public boolean isNewFarm()
    {
        return this.isNewFarm;
    }
    
    private void setNewFarmDefaults( PNFarm farmDefault )
    {
        int width = activity.getWindowManager().getDefaultDisplay().getWidth();
        int height = activity.getWindowManager().getDefaultDisplay().getHeight();
        // HOTFIX: honeycomb has another screen resolution then Window Manager reports.
        int barHeight = height - activity.findViewById( R.id.Main ).getBottom();
        
        if ( height > width )
        {
            int tmp = width;
            width = height;
            height = tmp;
        }
        
        int colorDepth = 24;
        
        farmDefault.setDesktopWidth( "" + width );
        farmDefault.setDesktopHeight( "" + (height - barHeight) );
        farmDefault.setColorDepth( "" + colorDepth );
    }
    
}
