package quest.android.appportal.controller;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import quest.android.appportal.R;
import quest.android.appportal.common.Common;
import quest.android.appportal.model.IModel;
import quest.android.appportal.view.custom.WallpaperGalleryAdapter;
import android.app.Activity;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.res.AssetManager;
import android.database.Cursor;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.provider.MediaStore;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnTouchListener;
import android.widget.CheckBox;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.LinearLayout.LayoutParams;

public class GlobalSettingsController
{
    private static Logger logger = Logger.getLogger( GlobalSettingsController.class );
    
    private static final int THUMBNAIL_WIDTH = 100;
    private static final int THUMBNAIL_HEIGHT = 100;
    
    Activity activity = null;
    IModel model = null;
    
    public GlobalSettingsController( Activity activity, IModel model )
    {
        this.activity = activity;
        this.model = model;
    }
    
    public void showDialog()
    {
        updateDialog();
        
        View globalOptsView = activity.findViewById( R.id.GlobalOptionsView );
        globalOptsView.setVisibility( View.VISIBLE );
    }
    
    protected void updateDialog()
    {
        final CheckBox enableLogging  = (CheckBox) activity.findViewById( R.id.GlobalEnableLogging );
        enableLogging.setChecked( model.isLoggingEnabled() );
        
        View loggerOption = activity.findViewById( R.id.GlobalOptionsLoggerOption );
        
        LinearLayout.LayoutParams loggerLayoutParams = (LayoutParams) loggerOption.getLayoutParams();
        int width = activity.getWindowManager().getDefaultDisplay().getWidth();
        int height = activity.getWindowManager().getDefaultDisplay().getHeight();
        width = height > width ? width : height;
        width -= 30;
        loggerLayoutParams.width = width; 
        loggerOption.setLayoutParams( loggerLayoutParams );
        
        loggerOption.setOnTouchListener( new OnTouchListener()
        {
            public boolean onTouch( View view, MotionEvent event )
            {
                enableLogging.onTouchEvent( event );
                return true;
            }
        });
        
        final View globalOptsView = this.activity.findViewById( R.id.GlobalOptionsView );
        View hideArea = this.activity.findViewById( R.id.GlobalOptionsHideArea );
        hideArea.setOnClickListener( new OnClickListener()
        {
            public void onClick( View view )
            {
                globalOptsView.setVisibility( View.GONE );
            }
        });
        
        View saveGlobalOptionsButton = this.activity.findViewById( R.id.SaveGlobalOptionsButton );
        saveGlobalOptionsButton.setOnClickListener( saveGlobalOptionsButtonListener );
        
        LinearLayout wallpaperGallery = (LinearLayout) this.activity.findViewById( R.id.WallpaperGallery );
        
        if ( wallpaperGallery.getChildCount() <= 1 ) loadImages( false );
        
        View externalWallpapers = activity.findViewById( R.id.GetExternalWallpaperImageButton );
        externalWallpapers.setOnClickListener( new OnClickListener()
        {
            public void onClick( View view )
            {
                loadImages( true );
            }
        });
    }
    
    OnClickListener saveGlobalOptionsButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            // Logger global options
            CheckBox enableLogging  = (CheckBox) activity.findViewById( R.id.GlobalEnableLogging );
            model.setLoggingEnabled( enableLogging.isChecked() );
            
            View globalOptsView = activity.findViewById( R.id.GlobalOptionsView );
            globalOptsView.setVisibility( View.GONE );
        }
    };
    
    private void setWallpaper( int position )
    {
        ImageView imageView = (ImageView) activity.findViewById( R.id.WallpaperHolder );
        String image = null;
        
        AssetManager assets = activity.getAssets();
        String[] thumbnails = new String[0];
        
        try
        {
            thumbnails = assets.list( "thumbnails" );
        }
        catch( IOException e ) { } 
        
        if ( position < thumbnails.length )
        {
            if ( position == 0 )
            {
                image = getDefaultWallpaperPath( activity );
            }
            else
            {
                image = "wallpapers/wallpaper_" + position + ".jpg";
            }
            
            try
            {
                imageView.setImageBitmap( BitmapFactory.decodeStream( assets.open( image ) ) );
            }
            catch ( OutOfMemoryError outOfMemoryError )
            {
                (new AlertDialog.Builder( activity ))
                .setMessage( "Memory is full. Please try select standard wallpaper and then additional wallpaper." )
                .setPositiveButton( R.string.button_text_ok, null )
                .create().show();
            }
            catch( Throwable e )
            {
                logger.warn( "Cannot load wallpaper: " + e.getMessage(), e );
            }
        }
        else
        {
            position -= thumbnails.length;
            List<String> paths = getWallpaperPathList();
            
            image = paths.get( position );
                
            try
            {
                imageView.setImageBitmap( BitmapFactory.decodeFile( image ) );
            }
            catch ( OutOfMemoryError outOfMemoryError )
            {
                (new AlertDialog.Builder( activity ))
                .setMessage( "Memory is full. Please try to select any standard wallpaper and then additional wallpaper." )
                .setPositiveButton( R.string.button_text_ok, null )
                .create().show();
            }
            catch( Throwable e )
            {
                logger.warn( "Cannot load wallpaper: " + e.getMessage(), e );
            }
        }
        
        model.setPathToDefaultWallpaper( image );
    }
    
    private List<Object> getThumbnailList()
    {
        ArrayList<Object> wallpapers = new ArrayList<Object>();
        
        try
        {
            AssetManager assets = activity.getAssets();
            String[] thumbnails = assets.list( "thumbnails" );
            int count = thumbnails.length;
            for ( int i = 0; i < count; ++i )
            {
                try
                {
                    wallpapers.add( BitmapFactory.decodeStream( assets.open( "thumbnails/thumbnail_" + i + ".png" ) ) );
                }
                catch( Throwable e )
                {
                    logger.warn( "Cannot load wallpaper: " + e.getMessage(), e );
                }
            }
        }
        catch( Throwable e )
        {
            logger.error( "Cannot load wallpaper thumbnails: " + e.getMessage(), e );
        }
        return wallpapers;
    }
    
    public static String getDefaultWallpaperPath( Activity activity )
    {
        String path = null;
        int width = activity.getWindowManager().getDefaultDisplay().getWidth();
        int height = activity.getWindowManager().getDefaultDisplay().getHeight();
        if ( width > height && width <= 800 && height <= 480 )
        {
            path = "wallpapers/wallpaper_0_800x480.jpg";
        }
        else if ( width < height && width <= 480 && height <= 800 )
        {
            path = "wallpapers/wallpaper_0_480x800.jpg";
        }
        else if ( width > height && width <= 1024 && height <= 600 )
        {
            path = "wallpapers/wallpaper_0_1024x600.jpg";
        }
        else if ( width < height && width <= 600 && height <= 1024 )
        {
            path = "wallpapers/wallpaper_0_600x1024.jpg";
        }
        else if ( width > height && width <= 1280 && height <= 800 )
        {
            path = "wallpapers/wallpaper_0_1280x800.jpg";
        }
        else if ( width < height && width <= 800 && height <= 1280 )
        {
            path = "wallpapers/wallpaper_0_800x1280.jpg";
        }
        else
        {
            path = "wallpapers/wallpaper_0.jpg";
        }
        return path;
    }
    
    public static Bitmap getDefaultWallpaper( Activity activity )
    {
        try
        {
            return BitmapFactory.decodeStream( activity.getAssets().open( getDefaultWallpaperPath( activity ) ) );
        }
        catch( Throwable e )
        {
            logger.error( "Cannot load wallpaper: " + e.getMessage(), e );
        }
        return null;
    }
    
    public static Bitmap getWallpaper( Activity activity, Object wallpaper )
    {
        if ( wallpaper instanceof String )
        {
            String path = (String)wallpaper;
            try
            {
                if ( !path.startsWith( "/" ) )
                {
                    if ( path.contains( "wallpaper_0" ))
                    {
                        path = getDefaultWallpaperPath( activity );
                    }

                    return BitmapFactory.decodeStream( activity.getAssets().open( path ) );
                }
                else
                {
                    return BitmapFactory.decodeFile( path );
                }
            }
            catch( Throwable e )
            {
                logger.error( "Cannot load wallpaper: " + e.getMessage(), e );
            }
        }
        return null;
    }
    
    public static void updateWallpaper( Activity activity, IModel model )
    {
        ImageView imageView = (ImageView) activity.findViewById( R.id.WallpaperHolder );
        
        try
        {
            String path = model.getPathToDefaultWallpaper();
            
            if ( Common.isNullOrEmpty( path ) )
            {
                imageView.setImageBitmap( getDefaultWallpaper( activity ) );
            }
            else
            {
                imageView.setImageBitmap( getWallpaper( activity, path ) );
            }
        }
        catch( Throwable e )
        {
            logger.warn( "Cannot set wallpaper: " + e.getMessage(), e );
        }
    }
    
    private void updateThumbnails( List<Object> additionalThumbnails )
    {
        List<Object> thumbnails = this.getThumbnailList();
        
        if ( additionalThumbnails != null )
        {
            thumbnails.addAll( additionalThumbnails );
        }
        
        LinearLayout wallpaperGallery = (LinearLayout) this.activity.findViewById( R.id.WallpaperGallery );
        wallpaperGallery.removeAllViews();
        
        WallpaperGalleryAdapter adapter = new WallpaperGalleryAdapter( activity, thumbnails );
        LinearLayout.LayoutParams params = new LinearLayout.LayoutParams( THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT );
        params.leftMargin = 5;
        params.rightMargin = 5;
        
        OnClickListener clickListener = new OnClickListener()
        {
            public void onClick( View view )
            {
                Integer position = (Integer) view.getTag();
                setWallpaper( position );
            }
        };
        
        for ( int i = 0; i < adapter.getCount(); ++i )
        {
            View view = adapter.getView( i, null, null );
            view.setLayoutParams( params );
            view.setTag( new Integer(i) );
            view.setOnClickListener( clickListener );
            wallpaperGallery.addView( view );
        }
    }
    
    private void getImagesList( Cursor cursor, List<String> list )
    {
        if ( cursor != null && cursor.getCount() > 0 )
        {
            cursor.moveToFirst();
            
            final int PATH_COLUMN = cursor.getColumnIndex( MediaStore.Images.Media.DATA );
            
            do
            {
                String path = cursor.getString( PATH_COLUMN );
                list.add( path );
            }
            while ( cursor.moveToNext() );
        }
    }
    
    private List<String> getWallpaperPathList()
    {
        ArrayList<String> paths = new ArrayList<String>();
        String[] columns = new String[]{ MediaStore.Images.Media.DATA };
        
        Cursor cursor = activity.getContentResolver().query( android.provider.MediaStore.Images.Media.INTERNAL_CONTENT_URI, columns, null, null, null);
        getImagesList( cursor, paths );
        
        cursor = activity.getContentResolver().query( android.provider.MediaStore.Images.Media.EXTERNAL_CONTENT_URI, columns, null, null, null);
        getImagesList( cursor, paths );
        
        return paths;
    }
    
    private void loadImages( final boolean addExternal )
    {
        final ProgressDialog progress = ProgressDialog.show( activity, null, "Loading wallpaper thumbnails...", true );
        
        new Thread( new Runnable()
        {
            public void run()
            {
                try
                {
                    LinearLayout wallpaperGallery = (LinearLayout) activity.findViewById( R.id.WallpaperGallery );
                    ArrayList<Object> externalViews = null;
                    
                    if ( addExternal )
                    {
                        List<String> externalPaths = getWallpaperPathList();
                        
                        externalViews = new ArrayList<Object>();
                        
                        for ( String path : externalPaths )
                        {
                            try
                            {
                                Bitmap bitmap = Bitmap.createScaledBitmap( BitmapFactory.decodeFile( path ), THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT, true );
                                externalViews.add( bitmap );
                            }
                            catch( Throwable e )
                            {
                                logger.error( "Error loading [" + path + "]: " + e.getMessage(), e );
                            }
                        }
                    }
                    
                    final ArrayList<Object> views = externalViews;
                    
                    wallpaperGallery.post( new Runnable()
                    {
                        public void run()
                        {
                            updateThumbnails( views );
                        }
                    });
                }
                catch( Throwable e )
                {
                    logger.error( "Error loading wallpapers: " + e.getMessage(), e );
                }
                
                progress.dismiss();
            }
        }).start();
    }
}
