package quest.android.appportal.controller;

import org.apache.log4j.Logger;

import quest.android.appportal.R;
import quest.android.appportal.common.Common;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNFarmException;
import quest.android.appportal.farm.PNLocation;
import quest.android.appportal.farm.PNLocationException;
import android.app.Activity;
import android.app.AlertDialog;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.CheckBox;
import android.widget.CompoundButton;
import android.widget.CompoundButton.OnCheckedChangeListener;
import android.widget.EditText;
import android.widget.TextView;

public class LocationSettingsController extends DialogController
{
    private Logger logger = Logger.getLogger( LocationSettingsController.class );
    
    private PNFarm farm = null;
    private PNLocation location = null;
    
    private boolean isNewLocation = false;
    
    public LocationSettingsController( PNFarm farm, Activity activity, PNLocation location )
    {
        super( activity );
        this.farm = farm;

        if ( location == null )
        {
            isNewLocation = true;
            this.location = new PNLocation( Common.EMTPY_STRING );
        }
        else
        {
            this.location = location;
        }
    }
    
    public void showDialog()
    {
        this.setContentView( R.layout.location_settings );
        
        dialog.findViewById( R.id.LocationSettingsSaveButton ).setOnClickListener( saveButtonListener );
        dialog.findViewById( R.id.LocationSettingsCancelButton ).setOnClickListener( cancelButtonListener );
        
        updateDialog();
        this.updateOrientation( true );
        
        dialog.show();
    }
    
    protected void updateDialog()
    {
        TextView header = (TextView) dialog.findViewById( R.id.LocationSettingsHeaderText );
        View deleteButton = dialog.findViewById( R.id.LocationSettingsDeleteButton );
        
        if ( isNewLocation ) 
        {
            header.setText( R.string.location_settings_dialog_header_add );
            deleteButton.setVisibility( View.GONE );
        }
        
        CompoundButton httpButton = (CompoundButton) dialog.findViewById( R.id.LocationSettingsHttpButton );
        CompoundButton httpsButton = (CompoundButton) dialog.findViewById( R.id.LocationSettingsHttpsButton );
        EditText name = (EditText) dialog.findViewById( R.id.LocationSettingsName );
        CheckBox defLocation = (CheckBox) dialog.findViewById( R.id.LocationSettingsDefaultLocation );
        EditText brokerName = (EditText) dialog.findViewById( R.id.LocationSettingsBrokerName );
        final EditText port = (EditText) dialog.findViewById( R.id.LocationSettingsTcpPort );
        CheckBox nat = (CheckBox) dialog.findViewById( R.id.LocationSettingsNat );
        CheckBox rdpOverSsl = (CheckBox) dialog.findViewById( R.id.LocationSettingsRdpOverSsl );
        EditText gateway = (EditText) dialog.findViewById( R.id.LocationSettingsSecureGateway );
        
        httpButton.setOnCheckedChangeListener( new OnCheckedChangeListener()
        {
            public void onCheckedChanged( CompoundButton view, boolean checked )
            {
                if ( checked )
                {
                    if ( port.getText().toString().equals( "443" ) )
                    {
                        port.setText( "8080" );
                    }
                }
                else
                {
                    if ( port.getText().toString().equals( "8080" ) )
                    {
                        port.setText( "443" );
                    }
                }
            }
        });
        
        httpButton.setChecked( !location.isHttpsProtocol() );
        httpsButton.setChecked( location.isHttpsProtocol() );
        
        name.setText( location.getName() );
        defLocation.setChecked( location.getUuid().equals( farm.getDefaultLocation() ) );
        brokerName.setText( location.getServersList().size() > 0 ? location.getServersList().get( 0 ) : Common.EMTPY_STRING );
        if ( !Common.isNullOrEmpty( location.getTcpPort() ) ) 
        {
            port.setText( location.getTcpPort() );
        }
        nat.setChecked( location.isNATEnabled() );
        rdpOverSsl.setChecked( location.isRdpOverSslEnabled() );
        gateway.setText( location.getSslGateway() );
        
        deleteButton.setOnClickListener( deleteButtonListener );
    }
    
    protected void updateSettings()
    {
        CompoundButton httpsButton = (CompoundButton) dialog.findViewById( R.id.LocationSettingsHttpsButton );
        EditText name = (EditText) dialog.findViewById( R.id.LocationSettingsName );
        CheckBox defLocation = (CheckBox) dialog.findViewById( R.id.LocationSettingsDefaultLocation );
        EditText brokerName = (EditText) dialog.findViewById( R.id.LocationSettingsBrokerName );
        EditText port = (EditText) dialog.findViewById( R.id.LocationSettingsTcpPort );
        CheckBox nat = (CheckBox) dialog.findViewById( R.id.LocationSettingsNat );
        CheckBox rdpOverSsl = (CheckBox) dialog.findViewById( R.id.LocationSettingsRdpOverSsl );
        EditText gateway = (EditText) dialog.findViewById( R.id.LocationSettingsSecureGateway );
        
        location.setHttpsProtocol( httpsButton.isChecked() );
        
        location.setName( name.getText().toString() );
        farm.setDefaultLocation( defLocation.isChecked() ? location.getUuid() : Common.EMTPY_STRING );
        
        try { location.removeServer( 0 ); } catch( PNLocationException e1 ) { }
        location.addServer( brokerName.getText().toString() );
        
        location.setTcpPort( port.getText().toString() );
        location.setNatEnabled( nat.isChecked() );
        location.setRdpOverSslEnabled( rdpOverSsl.isChecked() );
        location.setSslGateway( gateway.getText().toString() );
        
        if ( isNewLocation )
        {
            try
            {
                farm.addLocation( location );
            }
            catch( PNFarmException e )
            {
                logger.error( e.getMessage(), e.fillInStackTrace() );
            }
        }
    }
    
    private OnClickListener deleteButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            String uuid = location.getUuid();
            if ( farm.getLastLocation().equals( uuid ) )
            {
                farm.setConnected( false );
            }
            farm.getLocationList().remove( location );
            
            dismiss();
        }
    };
    
    private OnClickListener saveButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            String locationName = ((EditText) dialog.findViewById( R.id.LocationSettingsName )).getText().toString();
            
            if ( !Common.isCorrectName( locationName ) )
            {
                (new AlertDialog.Builder( activity ))
                .setMessage( R.string.location_settings_please_type_a_location_name )
                .setNeutralButton( R.string.button_text_ok, null )
                .create().show();
                return;
            }
            
            updateSettings();
            dismiss();
        }
    };
    
    private OnClickListener cancelButtonListener = new OnClickListener()
    {
        public void onClick( View v )
        {
            dismiss();
        }
    };
    
    public void dismiss()
    {
        dialog.dismiss();
    }
    
}
