
package quest.android.appportal.farm;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.provisionnetworks.common.PNCipher;

import quest.android.appportal.common.PropertyDefinitions;
import quest.android.appportal.common.Common;

public class PNFarm implements PropertyDefinitions
{
    private Logger logger = Logger.getLogger( PNFarm.class.getName() );
    
    // **********************************************************************************
    // Static members
    // **********************************************************************************
    
    public static final String REDIRECT_AUDIO_TYPE_TO_REMOTE    = "0";
    public static final String REDIRECT_AUDIO_TYPE_NOT_PLAY     = "1";
    public static final String REDIRECT_AUDIO_TYPE_LEAVE        = "2";
    
    public static final String KEYBOARD_HOOK_TYPE_LOCAL         = "0";
    public static final String KEYBOARD_HOOK_TYPE_REMOTE        = "1";
    public static final String KEYBOARD_HOOK_TYPE_FULL_SCREEN   = "2";
    
    public static final String COLOR_DEPTH_8BIT      = "8";
    public static final String COLOR_DEPTH_256       = "256";
    public static final String COLOR_DEPTH_15BIT     = "15";
    public static final String COLOR_DEPTH_16BIT     = "16";
    public static final String COLOR_DEPTH_24BIT     = "24";
    public static final String COLOR_DEPTH_32BIT     = "32";
    
    public static final String[] COLOR_DEPTH_LIST = new String[]{"256 Colors", 
        "High Color (15 Bit)", "High Color (16 Bit)", "True Color (24 Bit)", "True Color (32 Bit)" };
    
    public static final String[] DISPLAY_RESOLUTIONS = new String[]{ 
        "Portrait (full screen)", 
        "Landscape (full screen)",
        "640 x 480",
        "800 x 600",
        "1024 x 768",
        "1200 x 800",
        "1280 x 1024"        
        };
    
    public static final String Username                             = "username";
    public static final String Password                             = "password";
    public static final String Domain                               = "domain";
    public static final String AllowPasswordManagement              = "allowpasswordmanagement";
    public static final String AutoReconnect                        = "autoreconnect";
    public static final String ClientShortcutLocations              = "clientshortcutlocations";
    public static final String ClientShortcutLocationsLast          = "clientshortcutlocationslast";
    public static final String ColorDepth                           = "colordepth";
    public static final String ConfigFile                           = "configfile";
    public static final String CredentialsType                      = "credentialstype";
    public static final String DefaultLocation                      = "defaultlocation";
    public static final String Resolution                           = "resolution";
    public static final String DesktopHeight                        = "desktopheight";
    public static final String DesktopWidth                         = "desktopwidth";
    public static final String SaveCredentials                      = "savecredentials";
    public static final String DisplayConnectionBar                 = "displayconnectionbar";
    public static final String EnableAnimation                      = "enableanimation";
    public static final String EnableBitmapAcceleration             = "enablebitmapacceleration";
    public static final String EnableBitmapAccelerationPolicy       = "enablebitmapaccelerationpolicy";
    public static final String EnableBitmapCaching                  = "enablebitmapcaching";
    public static final String EnableDesktopComposition             = "enabledesktopcomposition";
    public static final String EnableFlashRedirection               = "enableflashredirection";
    public static final String EnableFlashRedirectionPolicy         = "enableflashredirectionpolicy";
    public static final String EnableFontSmoothing                  = "enablefontsmoothing";
    public static final String EnableFullWindowDrag                 = "enablefullwindowdrag";
    public static final String EnableLocalTextEcho                  = "enablelocaltextecho";
    public static final String EnableLocalTextEchoPolicy            = "enablelocaltextechopolicy";
    public static final String EnableMultimediaRedirection          = "enablemultimediaredirection";
    public static final String EnableMultimediaRedirectionPolicy    = "enablemultimediaredirectionpolicy";
    public static final String EnableSmartSizing                    = "enablesmartsizing";
    public static final String EnableThemes                         = "enablethemes";
    public static final String EnableWallpaper                      = "enablewallpaper";
    public static final String FarmUuid                             = "farmuuid";
    public static final String FullScreen                           = "fullscreen";
    public static final String HideSettings                         = "hidesettings";
    public static final String KerberosMode                         = "kerberosmode";
    public static final String KeyboardHook                         = "keyboardhook";
    public static final String LastLocation                         = "lastlocation";
    public static final String LoadFromServerProtocol               = "loadfromserverprotocol";
    public static final String Name                                 = "name";
    public static final String OverrideFarmName                     = "overridefarmname";
    public static final String PasswordManagementPort               = "passwordnanagementport";
    public static final String PasswordManagementServer             = "passwordnanagementserver";
    public static final String PinConnectionBar                     = "pinconnectionbar";
    public static final String RedirectAudio                        = "redirectaudio";
    public static final String RedirectClipboard                    = "redirectclipboard";
    public static final String RedirectComPorts                     = "redirectcomports";
    public static final String RedirectDrives                       = "redirectdrives";
    public static final String RedirectHandhelds                    = "redirecthandhelds";
    public static final String RedirectMicrophone                   = "redirectmicrophone";
    public static final String RedirectPrinters                     = "redirectprinters";
    public static final String RedirectSmartCards                   = "redirectsmartcards";
    public static final String RedirectUniversalPrinters            = "redirectuniversalprinters";
    public static final String SeamlessMode                         = "seamlessmode";
    public static final String SpanMonitors                         = "spanmonitors";
    public static final String Version                              = "version";
 
    // **********************************************************************************
    // Private members
    // **********************************************************************************
    
    private Properties farmProps = new Properties();
    private List<PNLocation> locations = new ArrayList<PNLocation>();
    private String name = "";
    private boolean connected = false;
    private boolean needReconnect = false;

    // **********************************************************************************
    // Constructors
    // **********************************************************************************

    public PNFarm(PNFarm farm)
    {
        logger.debug( "Farm copy constructor: " + farm.name );
        init();
        
        this.farmProps = new Properties( farm.farmProps );
        this.locations = new ArrayList<PNLocation>( farm.locations );
        this.setName( farm.name );
        this.setUuid( Common.generateUUID() );
        this.connected = false;
    }
    
    public PNFarm(String name)
    {
        logger.debug( "Create new farm: " + name );
        init();
        
        this.setName( name );
        this.setUuid( Common.generateUUID() );
        this.connected = false;
    }
    
    private void init()
    {
        
        this.farmProps.put( Username                         , "" );
        this.farmProps.put( Password                         , "" );
        this.farmProps.put( Domain                           , "" );
        this.farmProps.put( AllowPasswordManagement          , "0" );
        this.farmProps.put( AutoReconnect                    , "0" );
        this.farmProps.put( ClientShortcutLocations          , "" );
        this.farmProps.put( ClientShortcutLocationsLast      , "" );
        this.farmProps.put( ColorDepth                       , "24" );
        this.farmProps.put( ConfigFile                       , "config.xml" );
        this.farmProps.put( CredentialsType                  , "1" );
        this.farmProps.put( DefaultLocation                  , "" );
        this.farmProps.put( Resolution                       , "1" );
        this.farmProps.put( DesktopWidth                     , "1024" );
        this.farmProps.put( DesktopHeight                    , "600" );
        this.farmProps.put( SaveCredentials                  , "1" );
        this.farmProps.put( DisplayConnectionBar             , "0" );
        this.farmProps.put( EnableAnimation                  , "0" );
        this.farmProps.put( EnableBitmapAcceleration         , "0" );
        this.farmProps.put( EnableBitmapAccelerationPolicy   , "0" );
        this.farmProps.put( EnableBitmapCaching              , "0" );
        this.farmProps.put( EnableDesktopComposition         , "0" );
        this.farmProps.put( EnableFlashRedirection           , "0" );
        this.farmProps.put( EnableFlashRedirectionPolicy     , "0" );
        this.farmProps.put( EnableFontSmoothing              , "0" );
        this.farmProps.put( EnableFullWindowDrag             , "0" );
        this.farmProps.put( EnableLocalTextEcho              , "0" );
        this.farmProps.put( EnableLocalTextEchoPolicy        , "0" );
        this.farmProps.put( EnableMultimediaRedirection      , "0" );
        this.farmProps.put( EnableMultimediaRedirectionPolicy, "0" );
        this.farmProps.put( EnableSmartSizing                , "0" );
        this.farmProps.put( EnableThemes                     , "0" );
        this.farmProps.put( EnableWallpaper                  , "0" );
        this.farmProps.put( FarmUuid                         , "" );
        this.farmProps.put( FullScreen                       , "1" );
        this.farmProps.put( HideSettings                     , "0" );
        this.farmProps.put( KerberosMode                     , "0" );
        this.farmProps.put( KeyboardHook                     , "0" );
        this.farmProps.put( LastLocation                     , "" );
        this.farmProps.put( LoadFromServerProtocol           , "" );
        this.farmProps.put( Name                             , this.name );
        this.farmProps.put( OverrideFarmName                 , "0" );
        this.farmProps.put( PasswordManagementPort           , "" );
        this.farmProps.put( PasswordManagementServer         , "" );
        this.farmProps.put( PinConnectionBar                 , "0" );
        this.farmProps.put( RedirectAudio                    , "0" );
        this.farmProps.put( RedirectClipboard                , "0" );
        this.farmProps.put( RedirectComPorts                 , "0" );
        this.farmProps.put( RedirectDrives                   , "0" );
        this.farmProps.put( RedirectHandhelds                , "0" );
        this.farmProps.put( RedirectMicrophone               , "0" );
        this.farmProps.put( RedirectPrinters                 , "0" );
        this.farmProps.put( RedirectSmartCards               , "0" );
        this.farmProps.put( RedirectUniversalPrinters        , "0" );
        this.farmProps.put( SeamlessMode                     , "1" );
        this.farmProps.put( SpanMonitors                     , "0" );
        this.farmProps.put( Version                          , "6.0" );

    }
    
    // **********************************************************************************
    // Getters and Setters for Farm parameters
    // **********************************************************************************
    
    public String getUuid()
    {
        return getStringVal( FarmUuid );
    }
    
    public boolean isConnected()
    {
        return connected;
    }
    
    public void setConnected( boolean connected )
    {
        this.connected = connected;
    }
    
    public boolean isNeedReconnect()
    {
        return needReconnect;
    }
    
    public void setReconnect( boolean needReconnect )
    {
        this.needReconnect = (needReconnect && connected);
    }
    
    public void addLocation( PNLocation location ) throws PNFarmException
    {
        logger.debug( "Add new Location: " + location.getName());
        
        if ( this.getLocationByName( location.getName() ) == null )
        {
            this.locations.add( location );
        }
        else
        {
            throw new PNFarmException( "Location with same name exsists: " + name);
        }
    }
    
    public void addLocation( String name ) throws PNLocationException, PNFarmException
    {
        logger.debug( "Add new Location: " + name);
        
        if ( this.getLocationByName( name ) == null )
        {
            this.locations.add( new PNLocation( name ) );
        }
        else
        {
            throw new PNFarmException("Location with same name exsists: " + name);
        }
    }
    
    public String getUsername()
    {
        return getStringVal( Username );
    }
    
    public String getPassword()
    {
        return getStringVal( Password );
    }
    
    public String getDomain()
    {
        return getStringVal( Domain );
    }
    
    public String getClientShortcutLocations()
    {
        return getStringVal( ClientShortcutLocations );
    }
    
    public String getClientShortcutLocationsLast()
    {
        return getStringVal( ClientShortcutLocationsLast );
    }
    
    public String getColorDepth()
    {
        return getStringVal( ColorDepth );
    }
    
    public String getConfigFile()
    {
        return getStringVal( ConfigFile );
    }
    
    public String getCredentialsType()
    {
        return getStringVal( CredentialsType );
    }
    
    public String getDefaultLocation()
    {
        return getStringVal( DefaultLocation );
    }
    
    public String getResolution( )
    {
        return getStringVal( Resolution );
    }
    
    public String getDesktopHeight()
    {
        return getStringVal( DesktopHeight );
    }
    
    public String getDesktopWidth()
    {
        return getStringVal( DesktopWidth );
    }
        
    public String getLastLocation()
    {
        return getStringVal( LastLocation );
    }
    
    public int getLocationsCount( )
    {
        return locations.size();
    }
    
    public PNLocation getLocation( int index )
    {
        return locations.get( index );
    }
    
    public List<PNLocation> getLocationList()
    {
        return locations;
    }
    
    public PNLocation getLocationByName( String name )
    {
        Iterator<PNLocation> iter = locations.iterator();
        
        while ( iter.hasNext() )
        {
            PNLocation loc = iter.next();
            if ( loc.getName().equals( name ) )
            {
                return loc;
            }
        }
        
        return null;
    }
    
    public PNLocation getLocationByUuid( String uuid )
    {
        Iterator<PNLocation> iter = locations.iterator();
        
        while ( iter.hasNext() )
        {
            PNLocation loc = iter.next();
            if ( loc.getUuid().equals( uuid ) )
            {
                return loc;
            }
        }
        
        return null;
    }
    
    public String getName()
    {
        return name;
    }
    
    public boolean isAllowPasswordManagement()
    {
        return getBoolVal( AllowPasswordManagement );
    }
    
    public boolean isAnimationEnabled()
    {
        return getBoolVal( EnableAnimation );
    }
    
    public boolean isAutoReconnect()
    {
        return getBoolVal( AutoReconnect );
    }
    
    public boolean isBitmapAccelerationEnabled()
    {
        return getBoolVal( EnableBitmapAcceleration );
    }
    
    public boolean isBitmapCachingEnabled()
    {
        return getBoolVal( EnableBitmapCaching );
    }
    
    public boolean isDesktopCompositionEnabled()
    {
        return getBoolVal( EnableDesktopComposition );
    }
    
    public boolean isSaveCredentials()
    {
        return getBoolVal( SaveCredentials );
    }
    
    public boolean isDisplayConnectionBar()
    {
        return getBoolVal( DisplayConnectionBar );
    }
    
    public boolean isFlashRedirectionEnabled()
    {
        return getBoolVal( EnableFlashRedirection );
    }
    
    public boolean isFlashRedirectionPolicyEnabled()
    {
        return getBoolVal( EnableFlashRedirectionPolicy );
    }
    
    public boolean isFontSmoothingEnabled()
    {
        return getBoolVal( EnableFontSmoothing );
    }
    
    public boolean isFullScreenEnabled()
    {
        return getBoolVal( FullScreen );
    }
    
    public boolean isFullWindowDragEnabled()
    {
        return getBoolVal( EnableFullWindowDrag );
    }
    
    public boolean isHideSettingsEnabled()
    {
        return getBoolVal( HideSettings );
    }
    
    public boolean isKerberosModeEnabled()
    {
        return getBoolVal( KerberosMode );
    }
    
    public String getKeyboardHook()
    {
        return getStringVal( KeyboardHook );
    }
    
    public boolean isLocalTextEchoEnabled()
    {
        return getBoolVal( EnableLocalTextEcho );
    }
    
    public boolean isLocalTextEchoPolicyEnabled()
    {
        return getBoolVal( EnableLocalTextEchoPolicy );
    }
    
    public boolean isMultimediaRedirectionEnabled()
    {
        return getBoolVal( EnableMultimediaRedirection );
    }
    
    public boolean isMultimediaRedirectionPolicyEnabled()
    {
        return getBoolVal( EnableMultimediaRedirectionPolicy );
    }

    public boolean isSmartSizingEnabled()
    {
        return getBoolVal( EnableSmartSizing );
    }

    public boolean isThemesEnabled()
    {
        return getBoolVal( EnableThemes );
    }
    
    public boolean isWallpaperEnabled()
    {
        return getBoolVal( EnableWallpaper );
    }
    
    public boolean isLoadFromServerProtocol()
    {
        return getBoolVal( LoadFromServerProtocol );
    }
    
    public boolean isOverrideFarmName()
    {
        return getBoolVal( OverrideFarmName );
    }
    
    public String getPasswordManagementPort()
    {
        return getStringVal( PasswordManagementPort );
    }
    
    public String getPasswordManagementServer()
    {
        return getStringVal( PasswordManagementServer );
    }
    
    public boolean isPinConnectionBarEnabled()
    {
        return getBoolVal( PinConnectionBar );
    }
    
    public String getRedirectAudio()
    {
        return getStringVal( RedirectAudio );
    }
    
    public boolean isRedirectClipboard()
    {
        return getBoolVal( RedirectClipboard );
    }
    
    public boolean isRedirectComPorts()
    {
        return getBoolVal( RedirectComPorts );
    }
    
    public boolean isRedirectDrives()
    {
        return getBoolVal( RedirectDrives );
    }
    
    public boolean isRedirectHandhelds()
    {
        return getBoolVal( RedirectHandhelds );
    }
    
    public boolean isRedirectMicrophone()
    {
        return getBoolVal( RedirectMicrophone );
    }
    
    public boolean isRedirectPrinters()
    {
        return getBoolVal( RedirectPrinters );
    }
    
    public boolean isRedirectSmartCards()
    {
        return getBoolVal( RedirectSmartCards );
    }
    
    public boolean isRedirectUniversalPrinters()
    {
        return getBoolVal( RedirectUniversalPrinters );
    }
    
    public boolean isSeamlessMode()
    {
        return getBoolVal( SeamlessMode );
    }
    
    public boolean isSpanMonitors()
    {
        return getBoolVal( SpanMonitors );
    }
    
    public String getVersion()
    {
        return getStringVal( Version );
    }
    
    public void setName( String name )
    {
        logger.debug( "Set Farm name: " + name);
        this.setStringVal( Name, name );
        this.name = name;
    }
    
    public void setUsername( String username )
    {
        setStringVal( Username, username );
    }
    
    public void setPassword( String password )
    {
        setStringVal( Password, password );
    }
    
    public void setDomain( String domain )
    {
        setStringVal( Domain, domain );
    }
    
    public void setClientShortcutLocations( String val )
    {
        setStringVal( ClientShortcutLocations, val );
    }
    
    public void setClientShortcutLocationsLast( String val )
    {
        setStringVal( ClientShortcutLocationsLast, val );
    }
    
    public void setColorDepth( String val )
    {
        setStringVal( ColorDepth, val );
    }
    
    public void setConfigFile( String val )
    {
        setStringVal( ConfigFile, val );
    }
    
    public void setCredentialsType( String val )
    {
        setStringVal( CredentialsType, val );
    }
    
    public void setDefaultLocation( String val )
    {
        setStringVal( DefaultLocation, val );
    }
    
    public void setResolution( String val )
    {
        setStringVal( Resolution, val );
    }
    
    public void setDesktopHeight( String val )
    {
        setStringVal( DesktopHeight, val );
    }
    
    public void setDesktopWidth( String val )
    {
        setStringVal( DesktopWidth, val );
    }
    
    public void setUuid( String val )
    {
        setStringVal( FarmUuid, val );
    }
    
    public void setLastLocation( String val )
    {
        setStringVal( LastLocation, val );
    }
    
    public void setAllowPasswordManagement( boolean val )
    {
        setBoolVal( AllowPasswordManagement, val );
    }
    
    public void setAnimationEnabled( boolean val )
    {
        setBoolVal( EnableAnimation, val );
    }
    
    public void setAutoReconnect( boolean val )
    {
        setBoolVal( AutoReconnect, val );
    }
    
    public void setBitmapAccelerationEnabled( boolean val )
    {
        setBoolVal( EnableBitmapAcceleration, val );
    }
    
    public void setBitmapCachingEnabled( boolean val )
    {
        setBoolVal( EnableBitmapCaching, val );
    }
    
    public void setDesktopCompositionEnabled( boolean val )
    {
        setBoolVal( EnableDesktopComposition, val );
    }
    
    public void setSaveCredentialsEnabled( boolean val )
    {
        setBoolVal( SaveCredentials, val );
    }
    
    public void setDisplayConnectionBar( boolean val )
    {
        setBoolVal( DisplayConnectionBar, val );
    }
    
    public void setFlashRedirectionEnabled( boolean val )
    {
        setBoolVal( EnableFlashRedirection, val );
    }
    
    public void setFlashRedirectionPolicyEnabled( boolean val )
    {
        setBoolVal( EnableFlashRedirectionPolicy, val );
    }
    
    public void setFontSmoothingEnabled( boolean val )
    {
        setBoolVal( EnableFontSmoothing, val );
    }
    
    public void setFullScreenEnabled( boolean val )
    {
        setBoolVal( FullScreen, val );
    }
    
    public void setFullWindowDragEnabled( boolean val )
    {
        setBoolVal( EnableFullWindowDrag, val );
    }
    
    public void setHideSettingsEnabled( boolean val )
    {
        setBoolVal( HideSettings, val );
    }
    
    public void setKerberosModeEnabled( boolean val )
    {
        setBoolVal( KerberosMode, val );
    }
    
    public void setKeyboardHook( String val )
    {
        setStringVal( KeyboardHook, val );
    }
    
    public void setLocalTextEchoEnabled( boolean val )
    {
        setBoolVal( EnableLocalTextEcho, val );
    }
    
    public void setLocalTextEchoPolicyEnabled( boolean val )
    {
        setBoolVal( EnableLocalTextEchoPolicy, val );
    }
    
    public void setMultimediaRedirectionEnabled( boolean val )
    {
        setBoolVal( EnableMultimediaRedirection, val );
    }
    
    public void setMultimediaRedirectionPolicyEnabled( boolean val )
    {
        setBoolVal( EnableMultimediaRedirectionPolicy, val );
    }

    public void setSmartSizingEnabled( boolean val )
    {
        setBoolVal( EnableSmartSizing, val );
    }

    public void setThemesEnabled( boolean val )
    {
        setBoolVal( EnableThemes, val );
    }
    
    public void setWallpaperEnabled( boolean val )
    {
        setBoolVal( EnableWallpaper, val );
    }
    
    public void setLoadFromServerProtocol( boolean val )
    {
        setBoolVal( LoadFromServerProtocol, val );
    }
    
    public void setOverrideFarmName( boolean val )
    {
        setBoolVal( OverrideFarmName, val );
    }
    
    public void setPasswordManagementPort( String val )
    {
        setStringVal( PasswordManagementPort, val );
    }
    
    public void setPasswordManagementServer( String val )
    {
        setStringVal( PasswordManagementServer, val );
    }
    
    public void setPinConnectionBarEnabled( boolean val )
    {
        setBoolVal( PinConnectionBar, val );
    }
    
    public void setRedirectAudio( String val )
    {
        setStringVal( RedirectAudio, val );
    }
    
    public void setRedirectClipboard( boolean val )
    {
        setBoolVal( RedirectClipboard, val );
    }
    
    public void setRedirectComPorts( boolean val )
    {
        setBoolVal( RedirectComPorts, val );
    }
    
    public void setRedirectDrives( boolean val )
    {
        setBoolVal( RedirectDrives, val );
    }
    
    public void setRedirectHandhelds( boolean val )
    {
        setBoolVal( RedirectHandhelds, val );
    }
    
    public void setRedirectMicrophone( boolean val )
    {
        setBoolVal( RedirectMicrophone, val );
    }
    
    public void setRedirectPrinters( boolean val )
    {
        setBoolVal( RedirectPrinters, val );
    }
    
    public void setRedirectSmartCards( boolean val )
    {
        setBoolVal( RedirectSmartCards, val );
    }
    
    public void setRedirectUniversalPrinters( boolean val )
    {
        setBoolVal( RedirectUniversalPrinters, val );
    }
    
    public void setSeamlessMode( boolean val )
    {
        setBoolVal( SeamlessMode, val );
    }
    
    public void setSpanMonitors( boolean val )
    {
        setBoolVal( SpanMonitors, val );
    }
    
    public void setVersion( String val )
    {
        setStringVal( Version, val );
    }

    // **********************************************************************************
    // Utility functions
    // **********************************************************************************

    private void setBoolVal( String key, boolean value )
    {
        this.farmProps.put( key, value ? TRUE : FALSE );
    }
    
    private void setStringVal( String key, String value )
    {
        this.farmProps.put( key, value );
    }
    
    private boolean getBoolVal( String key )
    {
        String val = this.farmProps.getProperty( key, FALSE );
        return val != null ? ((String)val).equals( TRUE ) : false;
    }
    
    private String getStringVal( String key )
    {
        return this.farmProps.getProperty( key, "" );
    }
    
    public PNFarm clone()
    {
        logger.debug( "Start farm clonning: " + this.name );
        
        PNFarm clone = null;
        clone = new PNFarm(this.name);
        clone.farmProps = new Properties( this.farmProps );
        clone.locations = new Vector<PNLocation>( this.locations );
        clone.connected = this.connected; 
            
        return clone;
    }
    
    public void saveFarm( Document doc, Node root )
    {
        Enumeration<?> keys = this.farmProps.propertyNames();
        String param = null;
        String value = null;
        
        Element farmElement = doc.createElement( "Farm" );
        
        while ( keys.hasMoreElements() )
        {
            param = (String)keys.nextElement();
            value = this.farmProps.getProperty( param );
            
            if ( (param.equals( Password ) || param.equals( Username ) || param.equals( Domain ) ) && !this.isSaveCredentials() )
            {
                value = "";
            }
            
            if ( param.equals( Password  ) && value != null && value.length() > 0 )
            {
                byte[] bin = PNCipher.encode256( value );
                PNCipher.encrypt( bin );
                String encryptedPassword = PNCipher.binToHex( bin );
                
                value = encryptedPassword;
                logger.debug( "Saved: key=" + param );
            }
            else
            {
                logger.debug( "Saved: key=" + param + " value=" + value );
            }

            farmElement.setAttribute( param, value );
        }
        
        for ( PNLocation location : this.locations )
        {
            location.saveLocation( doc, farmElement );
        }
        
        root.appendChild( farmElement );
    }

    public static PNFarm loadFarm( Node root )
    {
        if ( !root.getNodeName().equals( "Farm" ) )
        {
            return null;
        }
        
        String name = ((Element) root).getAttribute( Name );
        
        if ( name == null )
        {
            return null;
        }
        
        PNFarm farm = null;
        
        try
        {
            farm = new PNFarm( name );
            farm.locations.clear();            
            
            NamedNodeMap attrs = root.getAttributes();
            
            for ( int i = 0; i < attrs.getLength(); ++i )
            {
                String key = attrs.item( i ).getNodeName();
                String value = attrs.item( i ).getNodeValue();
                
                if ( key.equals( Password ) && value != null && value.length() > 0 )
                {
                    byte[] bin = PNCipher.hexToBin( value ); 
                    PNCipher.decrypt( bin );
                    String decryptedPassword = PNCipher.decode256( bin );
                    
                    value = decryptedPassword;
                }
                
                farm.farmProps.put( key, value );
            }
            
            NodeList nodes = root.getChildNodes();
            
            for ( int i = 0; i < nodes.getLength(); ++i )
            {
                Node node = nodes.item( i );
                
                if ( node == null )
                {
                    break;
                }
                
                PNLocation loc = PNLocation.loadLocation( (Element)node );
                
                if ( loc != null )
                {
                    farm.addLocation( loc );
                }
            }
        }
        catch( PNFarmException e )
        {
        }
        
        return farm;
    }
    
}
