package quest.android.appportal.farm;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import android.content.Context;


public class PNFarmKeeper
{
    private Logger logger = Logger.getLogger( PNFarmKeeper.class.getName() ); 
    
    private static final String DEFAULT_FARM_FILE_NAME = "farms.xml";
    private static final String DEFAULT_ROOT_NODE_NAME = "Farms";
    
    private Vector<PNFarm> farms = new Vector<PNFarm>();
    
    public PNFarm createNewFarm( String name ) throws PNFarmException
    {
        if ( this.getFarmByName( name ) != null )
        {
            throw new PNFarmException( "Farm with this name already exists: " + name ); 
        }
        else
        {
            PNFarm farm = new PNFarm( name ); 
            this.farms.add( farm );
            return farm;
        }
    }
    
    public boolean addFarm( PNFarm farm )
    {
        if ( getFarmByUuid( farm.getUuid() ) == null )
        {
            return farms.add( farm );
        }
        
        return false;
    }
    
    public int getFarmCount()
    {
        return farms.size();
    }
    
    public PNFarm getFarmByName( String name )
    {
        for ( PNFarm farm : farms )
        {
            if ( farm.getName().equals( name ) )
            {
                return farm;
            }
        }
        
        return null;
    }
    
    public PNFarm getFarmByIndex( int index )
    {
        return farms.get( index ); 
    }
    
    public PNFarm getFarmByUuid( String uuid )
    {
        for ( PNFarm farm : farms )
        {
            if ( farm.getUuid().equals( uuid ) )
            {
                return farm;
            }
        }
        
        return null;
    }
    
    public void sortFarmsByName()
    {
        Collections.sort( farms, new Comparator<PNFarm>()
        {
            public int compare( PNFarm object1, PNFarm object2 )
            {
                return object1.getName().compareTo( object2.getName() );
            }
        });
    }
    
    public PNFarm getConnectedFarm()
    {
        for ( PNFarm farm : farms )
        {
            if ( farm.isConnected() )
            {
                return farm;
            }
        }
        
        return null;
    }
    
    public void saveFarms( Document doc, Node root )
    {
        for ( PNFarm farm : farms )
        {
            farm.saveFarm( doc, root );
        }
    }
    
    public void loadFarms( Node root )
    {
        NodeList nodes = root.getChildNodes();
        
        for ( int i = 0; i < nodes.getLength(); ++i )
        {
            PNFarm farm = PNFarm.loadFarm( nodes.item( i ) );
            
            if ( farm != null )
            {
                this.farms.add( farm );
            }
        }
        
    }
    
    public void saveFarmsToFile( Context context )
    {
        try
        {
            FileOutputStream fOut = context.openFileOutput( DEFAULT_FARM_FILE_NAME, Context.MODE_PRIVATE );
            
            DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
            Document doc = builderFactory.newDocumentBuilder().newDocument();
            Element root = doc.createElement( DEFAULT_ROOT_NODE_NAME );
            doc.appendChild( root );
            
            this.saveFarms( doc, root );
            
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            
            doc.normalize();
            javax.xml.transform.TransformerFactory.newInstance().
            newTransformer().
            transform(
                new javax.xml.transform.dom.DOMSource( doc ),
                new javax.xml.transform.stream.StreamResult( os ));
            
            
            fOut.write( os.toByteArray() );
            
            fOut.close();
        }
        catch( Exception e )
        {
            logger.error( e.getMessage(), e );
        }
    }
    
    public void loadFarmsFromFile( Context context )
    {
        try
        {
            FileInputStream fIn = context.openFileInput( DEFAULT_FARM_FILE_NAME );
            
            DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
            Document doc = builderFactory.newDocumentBuilder().parse( fIn );
            Element root = doc.getDocumentElement();
            
            this.loadFarms( root );
            
            fIn.close();
        }
        catch( Exception e )
        {
            logger.error( e.getMessage(), e );
        }
    }
    
    public void deleteFarmFile( Context context )
    {
        context.deleteFile( DEFAULT_FARM_FILE_NAME );
    }
    
    public void deletFarm( String uuid )
    {
        this.farms.remove( this.getFarmByUuid( uuid ) );
    }
}
