package quest.android.appportal.model;

import java.util.ArrayList;

import org.apache.log4j.Logger;

import android.app.Activity;

import quest.android.appportal.R;
import quest.android.appportal.common.Common;
import quest.android.appportal.connection.PNApplication;
import quest.android.appportal.connection.PNClient;
import quest.android.appportal.connection.PNException;
import quest.android.appportal.connection.PNFarmProperties;
import quest.android.appportal.controller.OnApplicationsListLoadListener;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNLocation;

public class ApplicationListLoaderThread extends Thread
{
    static private Logger logger = Logger.getLogger( ApplicationListLoaderThread.class );
    
    Activity activity = null;
    
    PNClient client = null;
    PNFarm farm = null;
    PNLocation location = null;
    OnApplicationsListLoadListener applicationsListListener = null;
    
    public ApplicationListLoaderThread( Activity activity, PNClient client, PNFarm farm, PNLocation location )
    {
        super();
        this.activity = activity;
        
        this.client = client;
        this.farm = farm;
        this.location = location;
    }
    
    public void setApplicationsListListener(
            OnApplicationsListLoadListener applicationsListListener )
    {
        this.applicationsListListener = applicationsListListener;
    }

    public void run()
    {
        String message = null;
        int errorCode = 0;
        ArrayList<PNApplication> applications = null;
        
        farm.setLastLocation( location.getUuid() );
        
        try
        {
            if ( Common.isNullOrEmpty( farm.getName() ) )
            {
                PNFarmProperties props = client.getFarmProperties();
                farm.setName( props.getFarmName() );
            }
            
            applications = client.getApplicationList();
            farm.setConnected( true );
        }
        catch ( PNException e )
        {
            logger.debug( message, e.fillInStackTrace() );
                                    
            message = e.getMessage();
            errorCode = e.getErrorCode();
            if ( message == null )
            {
                message = "";
            }
        }
        catch ( Exception e )
        {
            logger.debug( message, e.fillInStackTrace() );
        }
        
        if ( applications == null && message == null )
        {
            message = activity.getString( R.string.broker_error_cannot_connect_to_the_farm );
        }
        
        if ( applicationsListListener != null )
        {
            applicationsListListener.OnApplicationsListLoaded( new OnApplicationsListLoadListener.Arguments( farm, location, applications, message, errorCode ) );
        }
    }
}