package quest.android.appportal.model;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.provisionnetworks.common.OnRdpControlListener;

import android.app.Activity;
import android.graphics.Bitmap;
import android.view.View;
import quest.android.appportal.connection.PNApplication;
import quest.android.appportal.controller.OnApplicationsListLoadListener;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNFarmException;
import quest.android.appportal.farm.PNFarmKeeper;
import quest.android.appportal.farm.PNLocation;

public class DefaultModel implements IModel
{
    protected void finalize() throws Throwable
    {
        super.finalize();
    }
    
    private Logger logger = Logger.getLogger( DefaultModel.class );
    
    private Activity activity = null;
    private PNFarmKeeper farmKeeper = null;
    private RDPConnectionManager rdpConnMngr = null; 
    
    private GlobalOptions globalOptions = null;
    
    public DefaultModel( Activity activity )
    {
        this.activity = activity;
        this.farmKeeper = new PNFarmKeeper();
        this.rdpConnMngr = new RDPConnectionManager( activity, this );
        
        this.farmKeeper.loadFarmsFromFile( this.activity );
        this.globalOptions = new GlobalOptions();
        this.globalOptions.loadGlobalOptions( this.activity );
    }
    
    public void reset( Activity activity )
    {
        this.activity = activity;
    }
    
    public void save()
    {
        this.farmKeeper.saveFarmsToFile( this.activity );
        this.globalOptions.saveGlobalOptions( this.activity );
    }
    
    public void onExit() throws Throwable
    {
        if ( this.rdpConnMngr != null )
        {
            this.rdpConnMngr.onExit();
        }
        
        this.activity = null;
        this.farmKeeper = null;
        this.rdpConnMngr = null; 
    }

    public List<PNFarm> getFarmList()
    {
        List<PNFarm> farms = new ArrayList<PNFarm>();
        int count = this.farmKeeper.getFarmCount();
        for ( int i = 0; i < count; ++i )
        {
            farms.add( this.farmKeeper.getFarmByIndex( i ) );
        }
        return farms;
    }
    
    public PNFarm getFarm( String name )
    {
        return this.farmKeeper.getFarmByName( name );
    }
    
    public PNFarm getFarmByUuid( String uuid )
    {
        return this.farmKeeper.getFarmByUuid( uuid );
    }
    
    public PNFarm createNewFarm( String name )
    {
        try
        {
            return this.farmKeeper.createNewFarm( name );
        }
        catch( PNFarmException e )
        {
            logger.debug( e.getMessage(), e );
        }
        return null;
    }
    
    public void deleteSavedFarms()
    {
        this.farmKeeper.deleteFarmFile( activity );
    }
    
    public void deleteFarm( String uuid )
    {
        this.farmKeeper.deletFarm( uuid );
    }
    
    public PNFarm getConnectedFarm()
    {
        return this.farmKeeper.getConnectedFarm();
    }
    
    public void setOnApplicationsListListener( OnApplicationsListLoadListener listener )
    {
        this.rdpConnMngr.setOnApplicationsListListener( listener );
    }
    
    public void getApplicationList( PNFarm farm, PNLocation location )
    {
        this.rdpConnMngr.openConnection( farm, location );
    }
    
    public String testConnection( String protocol, List<String> servers, int port, String username, String password, String domain )
    {
        return rdpConnMngr.testConnection( protocol, servers, port, username, password, domain );
    }
    
    public String getPathToDefaultWallpaper()
    {
        return this.globalOptions.getWallpaperPath();
    }
    
    public String getPasscode()
    {
        return this.globalOptions.getPasscode();
    }
    
    public boolean isLoggingEnabled()
    {
        return this.globalOptions.isLoggingEnabled();
    }
    
    public void setPathToDefaultWallpaper( String path )
    {
        this.globalOptions.setWallpaperPath( path );
    }
    
    public void setPasscode( String passcode )
    {
        this.globalOptions.setPasscode( passcode );
    }
    
    public void setLoggingEnabled( boolean enabled )
    {
        this.globalOptions.setLoggingEnabled( enabled );
    }
    
    public void startRdpConnection( PNApplication application, PNFarm farm, PNLocation location )
    {
        logger.debug( "Model gets statring RDP connection" );
        
        this.rdpConnMngr.startRdpConnection( application, farm, location );
    }
    
    public boolean addFarm( PNFarm farm )
    {
        return farmKeeper.addFarm( farm );
    }

    public void sortFarmsByName()
    {
        this.farmKeeper.sortFarmsByName();
    }
    
    public void setOnRdpControlListener( OnRdpControlListener listener )
    {
        if ( this.rdpConnMngr != null )
        {
            this.rdpConnMngr.setOnRdpControlListener( listener );
        }
    }

    public List<Bitmap> getRdpDesktopsList()
    {
        return this.rdpConnMngr.getRdpDesktopsList();
    }
    
    public List<View> getRdpViewList()
    {
        return this.rdpConnMngr.getRdpViewList();
    }
    
    public List<RdpThread> getRdpThreadList()
    {
        return this.rdpConnMngr.getRdpThreadList();
    }
}
