package quest.android.appportal.model;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import android.content.Context;

public class GlobalOptions
{
    
    private Logger logger = Logger.getLogger( GlobalOptions.class.getName() );
    
    private final static String DEFAULT_GLOBAL_OPTIONS_FILE_NAME = "globaloptions.xml"; 
    private final static String DEFAULT_GLOBAL_OPTIONS_ROOT_NODE = "globaloptions.xml";
    
    private final static String WALLPAPER_PATH_NODE_NAME = "WallpaperPath";
    private final static String WALLPAPER_PASSCODE_NODE_NAME = "Passcode";
    private final static String ENABLE_LOGGING_NODE_NAME = "Logging";
    
    private String wallpaperPath = "";
    private String passcode = "";
    private Boolean loggingEnabled = false; 
    
    Context context = null;

    void saveGlobalOptions( Context context )
    {
        try
        {
            FileOutputStream fOut = context.openFileOutput( DEFAULT_GLOBAL_OPTIONS_FILE_NAME, Context.MODE_PRIVATE );
            
            DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
            Document doc = builderFactory.newDocumentBuilder().newDocument();
            Element root = doc.createElement( DEFAULT_GLOBAL_OPTIONS_ROOT_NODE );
            doc.appendChild( root );
            
            root.setAttribute( WALLPAPER_PATH_NODE_NAME, this.wallpaperPath );
            root.setAttribute( WALLPAPER_PASSCODE_NODE_NAME, this.passcode );
            root.setAttribute( ENABLE_LOGGING_NODE_NAME, this.loggingEnabled.toString() );
            
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            
            doc.normalize();
            javax.xml.transform.TransformerFactory.newInstance().
            newTransformer().
            transform(
                new javax.xml.transform.dom.DOMSource( doc ),
                new javax.xml.transform.stream.StreamResult( os ));
            
            
            fOut.write( os.toByteArray() );
            
            fOut.close();
        }
        catch( Exception e )
        {
            logger.error( e.getMessage(), e.fillInStackTrace() );
        }
    }
    
    public void loadGlobalOptions( Context context )
    {
        try
        {
            FileInputStream fIn = context.openFileInput( DEFAULT_GLOBAL_OPTIONS_FILE_NAME );
            
            DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
            Document doc = builderFactory.newDocumentBuilder().parse( fIn );
            Element root = doc.getDocumentElement();
            
            this.wallpaperPath = root.getAttribute( WALLPAPER_PATH_NODE_NAME );
            this.passcode = root.getAttribute( WALLPAPER_PASSCODE_NODE_NAME );
            this.loggingEnabled = Boolean.parseBoolean( root.getAttribute( ENABLE_LOGGING_NODE_NAME ) );
            
            Logger.setEnabled( loggingEnabled );
            
            fIn.close();
        }
        catch( Exception e )
        {
            logger.error( e.getMessage(), e.fillInStackTrace() );
        }
        
    }

    public String getWallpaperPath()
    {
        return wallpaperPath;
    }

    public void setWallpaperPath( String wallpaperPath )
    {
        this.wallpaperPath = wallpaperPath;
    }

    public String getPasscode()
    {
        return passcode;
    }

    public void setPasscode( String passcode )
    {
        this.passcode = passcode;
    }    
    
    public void setLoggingEnabled( boolean enabled )
    {
        this.loggingEnabled = enabled;
        Logger.setEnabled( enabled );
    }
    
    public boolean isLoggingEnabled()
    {
        return this.loggingEnabled;
    }
}
