package quest.android.appportal.model;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.provisionnetworks.common.OnRdpControlListener;

import quest.android.appportal.common.Common;
import quest.android.appportal.connection.PNApplication;
import quest.android.appportal.connection.PNClient;
import quest.android.appportal.connection.PNException;
import quest.android.appportal.controller.OnApplicationsListLoadListener;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNLocation;

import android.app.Activity;
import android.graphics.Bitmap;
import android.view.View;

public class RDPConnectionManager implements OnThreadStopListener
{
    private Logger logger = Logger.getLogger( RDPConnectionManager.class );
    
    private IModel model = null;
    private Activity activity = null;
    
    private PNClient activeClient = null;
    private ArrayList<RdpThread> rdpConnections = null;
    
    // Listeners
    OnApplicationsListLoadListener applicationsListListener = null;
    OnRdpControlListener rdpControlListener = null;
    
    public RDPConnectionManager( Activity activity, IModel model )
    {
        this.activity = activity;
        this.model = model;
        
        rdpConnections = new ArrayList<RdpThread>();
    }
    
    public void onExit()
    {
        for ( RdpThread thread : rdpConnections )
        {
            thread.stopConnection();
        }
        model = null;
        activity = null;
        
        activeClient = null;
        rdpConnections = null;
        
        // Listeners
        applicationsListListener = null;
        rdpControlListener = null;
    }
    
    public void openConnection( PNFarm farm, PNLocation location )
    {
        if ( model.getConnectedFarm() != null )
        {
            model.getConnectedFarm().setConnected( false );
        }
        
        String username = farm.getUsername();
        String password = farm.getPassword();
        String domain = farm.getDomain();
        
        int port = Common.parseIntSafe( location.getTcpPort() );
        String protocol = location.getProtocol();
        
        List<String> servers = location.getServersList();
        
        try
        {
            activeClient = new PNClient( activity, protocol, servers, port, username, password, domain );
        }
        catch( Exception e )
        {
            logger.error( e.getMessage(), e.fillInStackTrace() );
        }
        
        ApplicationListLoaderThread thread = new ApplicationListLoaderThread( activity, activeClient, farm, location );
        thread.setApplicationsListListener( applicationsListListener );
        thread.start();
    }
    
    public String testConnection( String protocol, List<String> servers, int port, String username, String password, String domain )
    {
        String message = null;
        try
        {
            PNClient client = new PNClient( activity, protocol, servers, port, username, password, domain );
            client.authenticate();
        }
        catch ( PNException e )
        {
            logger.debug( message, e.fillInStackTrace() );
            
            message = e.getMessage();
            if ( message == null )
            {
                message = "";
            }
        }
        catch ( Exception e )
        {
            if ( message == null )
            {
                message = e.getMessage();
                if ( message == null )
                {
                    message = "";
                }
            }
            
            logger.debug( message, e.fillInStackTrace() );
        }
        
        return message;
    }
    
    public void setOnApplicationsListListener( OnApplicationsListLoadListener listener )
    {
        applicationsListListener = listener;
    }
    
    public void startRdpConnection( PNApplication application, PNFarm farm, PNLocation location )
    {   
        RdpThread thread = new RdpThread( activity, activeClient, farm, location, application );
        logger.debug( "RDP thread created" );
        thread.setOnStopListener( this );
        thread.setOnRdpControlListener( rdpControlListener );
        this.rdpConnections.add( thread );
        thread.start();
        logger.debug( "RDP thread started" );
    }
    
    public void onThreadStop( Thread thread )
    {
        if ( rdpConnections != null && !this.rdpConnections.remove( thread ) )
        {
            logger.error( "Cannot find thread in the thread pool: " + thread.getId() );
        }
    }
    
    public int getRdpConnectionsCount()
    {
        return rdpConnections.size();
    }
    
    public RdpThread getRdpThread( int index )
    {
        return rdpConnections.get( index );
    }
    
    public List<Bitmap> getRdpDesktopsList()
    {
        ArrayList<Bitmap> views = new ArrayList<Bitmap>();
        for ( int i = 0; i < rdpConnections.size(); ++i )
        {
            views.add( rdpConnections.get( i ).getRdpDesktop() );
        }
        return views;
    }
    
    public List<View> getRdpViewList()
    {
        ArrayList<View> views = new ArrayList<View>();
        for ( int i = 0; i < rdpConnections.size(); ++i )
        {
            views.add( rdpConnections.get( i ).getRdpView() );
        }
        return views;
    }
    
    public void setOnRdpControlListener( OnRdpControlListener listener )
    {
        rdpControlListener = listener;
    }
    
    public List<RdpThread> getRdpThreadList()
    {
        return this.rdpConnections;
    }
}