package quest.android.appportal.model;

import java.io.IOException;

import org.apache.log4j.Logger;

import quest.android.appportal.common.Common;
import quest.android.appportal.connection.PNApplication;
import quest.android.appportal.connection.PNClient;
import quest.android.appportal.connection.PNException;
import quest.android.appportal.connection.PNSessionInfo;
import quest.android.appportal.farm.PNFarm;
import quest.android.appportal.farm.PNLocation;
import android.app.Activity;
import android.graphics.Bitmap;
import android.view.View;

import com.provisionnetworks.common.OnRdpControlListener;
import com.provisionnetworks.common.PNCipher;
import com.provisionnetworks.pntsc.PNTSC;

public class RdpThread extends Thread implements OnRdpControlListener
{
    private Logger logger = Logger.getLogger( RdpThread.class );
    
    private static long idCounter = 0;
    
    private long id = 0;
    
    private Activity activity = null;
    private PNClient client = null;
    private PNFarm farm = null;
    private PNLocation location = null;
    private PNApplication application = null;
    
    private View rdpView = null;
    private Bitmap rdpDesktop = null;
    
    private OnThreadStopListener stopListener = null;
    private OnRdpControlListener rdpControlListener = null; 

    private PNTSC pntsc = null;
    
    public RdpThread( Activity activity, PNClient client, PNFarm farm,
            PNLocation location, PNApplication application )
    {
        super();
        this.activity = activity;
        this.client = client;
        this.farm = farm;
        this.location = location;
        this.application = application;
        this.id = idCounter++;
    }
    
    public void setOnStopListener( OnThreadStopListener stopListener )
    {
        this.stopListener = stopListener;
    }
    
    public void setOnRdpControlListener( OnRdpControlListener listener )
    {
        this.rdpControlListener = listener;
    }
    
    private String encrypt256(String value)
    {
        try
        {
            byte[] bin = PNCipher.encode256(value);
            PNCipher.encrypt(bin);
            return PNCipher.binToHex(bin);
        }
        catch (Exception e) 
        { 
            return Common.EMTPY_STRING;
        }
    }
    
    private String encrypt1024(String value)
    {
        try
        {
            byte[] bin = PNCipher.encode1024(value);
            PNCipher.encrypt(bin);
            return PNCipher.binToHex(bin);
        }
        catch (Exception e) 
        { 
            return Common.EMTPY_STRING; 
        }
    }
    
    public void run()
    {
        logger.debug( "RDP thread is ran" );
        
        String reason = null;
        
        try
        {
//            final int desktopImageSize = Common.parseIntSafe( farm.getDesktopWidth() ) * Common.parseIntSafe( farm.getDesktopHeight() );
//            final int cacheSize = 4 * 921600;
//            final int heapNeeded = desktopImageSize + cacheSize;
//            Common.allocateHeap( heapNeeded );
            
            String commandString = null;
            PNSessionInfo info = this.client.launchApplication( application.getName() );
            
            String port = Common.EMTPY_STRING;
            if ( !Common.isNullOrEmpty( info.getDnaProxy() ) )
            {
                int pos = info.getDnaProxy().indexOf( ':' );
                if ( info.getDnaProxy().length() > pos + 1 )
                {
                    port = ":" + info.getDnaProxy().substring( pos + 1 );
                }
            }
            
            commandString = "\nservername:s:" + info.getServerName()
                + "\nserveraddr:s:" + info.getServerAddr() + port
                + "\nconnectingtext:s:" + info.getName()
                + "\nfarmname:s:" + farm.getName()
                + "\ndomain:s:" + farm.getDomain()
                + "\nusername:s:" + farm.getUsername()
                + "\npassword:s:" + encrypt256( farm.getPassword() )
                + "\nenablesso:i:0"
                + "\nenablekerberos:i:0"
                + "\nsslgateway:s:" + location.getSslGateway()
                + "\nproxyserver:s:" + location.getProxyServer()
                + "\nalternate shell:s:" + encrypt1024( info.getPath() )
                + "\nprogramworkingdir:s:" + info.getWorkingDir()
                + "\nshell working directory:s:" + info.getWorkingDir()
                + "\napptype:i:0"
                + "\nscreen mode id:i:1"
                + "\ndesktopwidth:i:" + farm.getDesktopWidth()
                + "\ndesktopheight:i:" + farm.getDesktopHeight()
                + "\nseamless:i:0"
                + "\nsession bpp:i:" + farm.getColorDepth()
                + "\nsession sharing:i:" + info.getSessionSharing()
                + "\ndisplayconnectionbar:i:0"
                + "\npinconnectionbar:i:0"
                + "\nsmartsizing:i:0"
                + "\naudiomode:i:" + farm.getRedirectAudio()
                + "\nkeyboardhook:i:" + farm.getKeyboardHook()
                + "\nredirectdrives:i:1"
                + "\nredirectprinters:i:0"
                + "\nredirectcomports:i:0"
                + "\nredirectsmartcards:i:0"
                + "\nredirecthandhelds:i:0"
                + "\nredirectuniversalprinters:i:0"
                + "\nredirectclipboard:i:0"
                + "\ndisable wallpaper:i:" + (farm.isWallpaperEnabled() ? 0 : 1)
                + "\ndisable full window drag:i:" + (farm.isFullWindowDragEnabled() ? 0 : 1)
                + "\ndisable menu anims:i:" + (farm.isAnimationEnabled() ? 0 : 1)
                + "\ndisable themes:i:" + (farm.isThemesEnabled() ? 0 : 1)
                + "\ndisable font smoothing:i:" + (farm.isFontSmoothingEnabled() ? 0 : 1)
                + "\ndisable desktop composition:i:" + (farm.isDesktopCompositionEnabled() ? 0 : 1)
                + "\nbitmapcachepersistenable:i:" + (farm.isBitmapCachingEnabled() ? 1 : 0)
                + "\nautoreconnection enabled:i:0"
                + "\nencryption:i:1"
                + "\nbitmapacceleration:i:" + (farm.isBitmapAccelerationEnabled() ? 1 : 0)
                + "\ncompression:i:1"
                + "\n";
            
            pntsc = new PNTSC( (Activity) activity );
            logger.debug( "PNTSC created" );
            pntsc.setOnRdpControlListener( this );
            logger.debug( "PNTSC is started" );
            pntsc.main( commandString );
            logger.debug( "PNTSC stopped" );
        }
        catch( IOException e )
        {
            logger.error( e.getMessage(), e );
            reason = e.getMessage();
        }
        catch( PNException e )
        {
            logger.error( e.getMessage(), e );
            reason = e.getMessage();
        }
        catch( Throwable e )
        {
            logger.error( e.getMessage(), e );
            reason = e.getMessage();
        }
        finally
        {
            try
            {
                pntsc.destroy();
            }
            catch( Throwable e ) { }
            pntsc = null;
        }
        
        logger.debug( "RDP connection stopped: " + reason );
        
        if ( stopListener != null )
        {
            stopListener.onThreadStop( this );
            stopListener = null;
        }
        
        if ( rdpControlListener != null )
        {
            logger.debug( "Notify that connection is stopped with reason: " + reason );
            rdpControlListener.OnRdpConnectionStopped( rdpView, reason );
            rdpControlListener = null;
        }
    }

    public void stopConnection()
    {
        try { pntsc.stop(); } catch ( Exception e ) { }
        try
        {
            this.interrupt();
        }
        catch( Throwable e )
        {
            logger.warn( "PNTSC interrupting error: " + e.getMessage(), e );
        }
        
        activity = null;
        client = null;
        farm = null;
        location = null;
        application = null;
        
        rdpView = null;
        rdpDesktop = null;
        
        logger.debug( "Connection stopped by user command" );
    }

    public String getApplicationName()
    {
        if ( application == null ) return null;
        
        return this.application.getName();
    }
    
    public Bitmap getRdpDesktop()
    {
        return this.rdpDesktop;
    }
    
    public View getRdpView()
    {
        return this.rdpView;
    }

    public void OnRdpViewCreated( View view, Bitmap desktop )
    {
        logger.debug( "RDP View object created" );
        
        this.rdpView = view;
        this.rdpDesktop = desktop;
        
        if ( rdpControlListener != null )
        {
            rdpControlListener.OnRdpViewCreated( view, desktop );
        }
    }

    public void OnRdpViewBackClicked( View view )
    {
        logger.debug( "RDP back button clicked" );
        
        if ( rdpControlListener != null )
        {
            rdpControlListener.OnRdpViewBackClicked( view );
        }
    }

    public void OnRdpConnectionStopped( View view, String reason )
    {
        logger.debug( "RDP connection stopped: " + reason );
        
        if ( rdpControlListener != null )
        {
            rdpControlListener.OnRdpConnectionStopped( view, reason );
        }
    }
    
    public long getId()
    {
        return id;
    }
}
