
package quest.android.appportal;

import org.apache.log4j.Logger;

import quest.android.appportal.controller.DefaultController;
import quest.android.appportal.controller.IController;
import quest.android.appportal.model.DefaultModel;
import quest.android.appportal.model.IModel;
import quest.android.appportal.view.IView;
import quest.android.appportal.view.UIView;

import android.app.Activity;
import android.content.res.Configuration;
import android.os.Bundle;

public class vWorkspaceConnectorForAndroid extends Activity
{
    private Logger logger = null;

    public final String APPLICATION_NAME = "vWorkspace Connector for Android";
    public final String VERSION          = "1.0";

    IView view = null;
    IModel model = null;
    IController controller = null;
    
    boolean isLandscape = true;
    int prevKeyboard = 0;
    int prevKeyboardHidden = 0;
    
    /** Called when the activity is first created. */
    @Override
    public void onCreate( Bundle savedInstanceState )
    {
        if ( logger != null )
        {
            logger.debug( "Create activity" );
        }
        else
        {
            logger = Logger.getLogger( vWorkspaceConnectorForAndroid.class );
            Logger.setEnabled( false );
        }
        
        super.onCreate( savedInstanceState );

        logger.debug( APPLICATION_NAME + " " + VERSION + " has been started" );
        
        // Detecting landscape orientation
        this.isLandscape = (this.getRequestedOrientation() & 1) == 1;
        
        // UIView must be created at first then Model and then Controller
        view = new UIView( this );
        model = new DefaultModel( this );
        controller = new DefaultController( this, model );
        controller.startSplashAnimation( 2000 );
    }
    
    @Override
    public void onConfigurationChanged( Configuration newConfig ) 
    {
        logger.debug( "Configuration changed" );
        
        if ( prevKeyboard != newConfig.keyboard )
        {
            logger.info( "Keyboard changed" );
        }
        
        if ( prevKeyboardHidden != newConfig.keyboardHidden )
        {
            logger.info( "Keyboard Hidden changed" );
        }
        
        prevKeyboard = newConfig.keyboard;
        prevKeyboardHidden = newConfig.keyboardHidden;
            
        isLandscape = (newConfig.orientation == Configuration.ORIENTATION_LANDSCAPE) ||
        (newConfig.orientation == Configuration.ORIENTATION_SQUARE);
        
        controller.updateOrientation( isLandscape );
        
        super.onConfigurationChanged( newConfig );
    }
    
    @Override
    protected void onPause()
    {
        if ( model != null )
        {
            model.save();
        }
        
        logger.debug( "Pause work" );
        
        super.onPause();        
    }
    
    @Override
    protected void onDestroy()
    {
        if ( model != null )
        {
            try
            {
                model.onExit();
            }
            catch( Throwable e )
            {
            }
        }
        
        if ( controller != null )
        {
            try
            {
                controller.onExit();
            }
            catch( Throwable e )
            {
            }
        }
        
        controller = null;
        view = null;
        model = null;
        
        logger.debug( "On destroy application" );
        super.onDestroy();
    }
    
    public void onBackPressed()
    {
        super.moveTaskToBack( true );
    }
}
