package quest.android.appportal.view.custom;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import quest.android.appportal.R;
import quest.android.appportal.farm.PNFarm;
import android.content.Context;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.View.OnClickListener;
import android.widget.BaseAdapter;
import android.widget.ImageView;
import android.widget.TextView;

public class FarmListAdapter extends BaseAdapter
{
    private Logger logger = Logger.getLogger( FarmListAdapter.class );
    
    private LayoutInflater mInflater;
    
    private List<PNFarm> farms;
    private ArrayList<String> items = new ArrayList<String>();
    
    private OnClickListener onDeleteClickListener = null;
    private OnClickListener onSettingsClickListener = null;
    
    public FarmListAdapter( Context context, List<PNFarm> items )
    {
        logger.debug( "Farm list adapter created" );
        
        mInflater = LayoutInflater.from(context);
        this.farms = items;
        
        for ( int i = 0; i < this.farms.size(); ++i )
        {
            this.items.add( this.farms.get( i ).getName() );
        }
        this.items.add( context.getString( R.string.farm_list_add_farm ) );
    }
    
    public void setOnDeleteClickListener( OnClickListener listener )
    {
        onDeleteClickListener = listener;
    }
    
    public void setOnSettingsClickListener( OnClickListener listener )
    {
        onSettingsClickListener = listener;
    }
    
    public int getCount()
    {
        return items.size();
    }

    public Object getItem( int position )
    {
        return position;
    }

    public long getItemId( int position )
    {
        return position;
    }

    public View getView( int position, View convertView, ViewGroup parent )
    {
        ViewHolder holder;

        if (convertView == null)
        {
            convertView = mInflater.inflate(R.layout.farm_list_item, null);

            holder = new ViewHolder();
            holder.text = (TextView) convertView.findViewById(R.id.FarmListTextItem);
            holder.settingsButton = (ImageView) convertView.findViewById(R.id.SettingsIcon);
            holder.deleteButton = convertView.findViewById(R.id.DeleteIcon);
            convertView.setTag(holder);
        }
        else 
        {
            holder = (ViewHolder) convertView.getTag();
        }
        
        FarmListItem item = new FarmListItem();
        PNFarm farm = null;
        
        item.farmUuid = position >= farms.size() ? null : (farm = farms.get( position )).getUuid();
        item.position = position;
       
        holder.text.setText( items.get( position ) );
        
        if ( farm != null )
        {
            holder.deleteButton.setVisibility( View.VISIBLE );
            holder.settingsButton.setVisibility( View.VISIBLE );
            
            if ( onDeleteClickListener != null )
            {
                holder.deleteButton.setOnClickListener( onDeleteClickListener );
                holder.deleteButton.setTag( item );
            }
            
            if ( onSettingsClickListener != null )
            {
                holder.settingsButton.setOnClickListener( onSettingsClickListener );
                holder.settingsButton.setTag( item );
            }
        }
        else
        {
            holder.deleteButton.setVisibility( View.GONE );
            holder.settingsButton.setVisibility( View.GONE );
        }

        return convertView;
    }
    
    private static class ViewHolder
    {
        private TextView  text = null;
        private ImageView settingsButton = null;
        private View deleteButton = null;
    }
    
    public class FarmListItem
    {
        private String farmUuid = null;
        private int position = 0;
        
        public String getFarmUuid()
        {
            return farmUuid;
        }
        
        public int getPosition()
        {
            return position;
        }
    }
}
