
package quest.android.appportal.view.custom;

import android.app.Activity;
import android.graphics.Matrix;
import android.graphics.PointF;
import android.util.FloatMath;
import android.view.MotionEvent;
import android.widget.ImageView;

public class PinchScaling
{
    public static final float MIN_SCALE_FACTOR           = 0.05F;

    public static final float MIN_SCALE                  = 1F;
    public static final float MAX_SCALE                  = 5F;
    public static final float NO_SCALE                   = 1F;

    public static final float UP_STICK_SCALE_THRESHOLD   = 1F;
    public static final float DOWN_STICK_SCALE_THRESHOLD = 0.95F;

    ImageView                 view                       = null;
    Matrix                    matrix                     = null;
    
    PointF middle = new PointF();
    float distance = 0.0F;

    public PinchScaling( Activity activity, ImageView imageView, Matrix matrix )
    {
        this.view = imageView;
        this.matrix = matrix;
    }

    public float getCurrentScale()
    {
        return getMatrixValue( Matrix.MSCALE_X );
    }
    
    private float getMatrixValue( int id )
    {
        final int maxMatrixValues = 9;
        
        if ( id < 0 || id >= maxMatrixValues ) return 0;
        
        float[] values = new float[maxMatrixValues];
        matrix.getValues( values );
        return values[id];
    }
    
    public void startScaling( MotionEvent event )
    {
        if ( event.getPointerCount() == 2 )
        {
            midPoint( middle, event );
            distance = getDistance( event );
        }
    }
    
    public void processScaling( MotionEvent event )
    {
        float currentScale = getMatrixValue( Matrix.MSCALE_X );

        float newDistance = getDistance( event );
        
        float scale = newDistance / distance;
        
        distance = newDistance; 
        
        float focusX = middle.x;
        float focusY = middle.y;
        
        if ( currentScale * scale < MIN_SCALE )
        {
            scale = MIN_SCALE / currentScale;
        }
        else if ( currentScale * scale > MAX_SCALE )
        {
            scale = MAX_SCALE / currentScale;
        }
        else if ( currentScale * scale > DOWN_STICK_SCALE_THRESHOLD
                && currentScale * scale < UP_STICK_SCALE_THRESHOLD )
        {
            scale = NO_SCALE / currentScale;
        }

        matrix.postScale( scale, scale, focusX, focusY );
        
        float[] values = new float[9];
        matrix.getValues( values );
        
        float dx = 0;
        float dy = 0;
        
        int viewWidth = view.getWidth();
        int viewHeight = view.getHeight();

        int width = view.getDrawable().getIntrinsicWidth();
        int height = view.getDrawable().getIntrinsicHeight();

        float scaledWidth = width * values[Matrix.MSCALE_X];
        float scaledHeight = height * values[Matrix.MSCALE_Y];
        
        if ( scaledWidth < viewWidth )
        {
            if ( values[Matrix.MTRANS_X] + dx < 0 )
            {
                dx = -values[Matrix.MTRANS_X];
            }
            else if ( values[Matrix.MTRANS_X] + scaledWidth + dx > viewWidth )
            {
                dx = viewWidth - (values[Matrix.MTRANS_X] + scaledWidth);
            }
        }
        else
        {
            if ( values[Matrix.MTRANS_X] + dx > 0 )
            {
                dx = -values[Matrix.MTRANS_X];
            }
            else if ( values[Matrix.MTRANS_X] + scaledWidth + dx < viewWidth )
            {
                dx = viewWidth - (values[Matrix.MTRANS_X] + scaledWidth);
            }
        }
        
        if ( scaledHeight < viewHeight )
        {
            if ( values[Matrix.MTRANS_Y] + dy < 0 )
            {
                dy = -values[Matrix.MTRANS_Y];
            }
            else if ( values[Matrix.MTRANS_Y] + scaledHeight + dy > viewHeight )
            {
                dy = viewHeight - (values[Matrix.MTRANS_Y] + scaledHeight);
            }
        }
        else
        {
            if ( values[Matrix.MTRANS_Y] + dy > 0 )
            {
                dy = -values[Matrix.MTRANS_Y];
            }
            else if ( values[Matrix.MTRANS_Y] + scaledHeight + dy < viewHeight )
            {
                dy = viewHeight - (values[Matrix.MTRANS_Y] + scaledHeight);
            }
        }
        
        matrix.postTranslate( dx, dy );
        middle.set( middle.x + dx, middle.y + dy );
        
        view.setImageMatrix( matrix );
    }

    private float getDistance( MotionEvent event ) 
    {
        float x = event.getX(0) - event.getX(1);
        float y = event.getY(0) - event.getY(1);
        return FloatMath.sqrt(x * x + y * y);
    }
    
    private void midPoint( PointF point, MotionEvent event ) 
    {
        float x = event.getX(0) + event.getX(1);
        float y = event.getY(0) + event.getY(1);
        point.set(x / 2F, y / 2F);
    }
    
}
