package quest.android.appportal.view.custom;

import java.util.List;

import quest.android.appportal.R;

import android.content.Context;
import android.graphics.Bitmap;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.BaseAdapter;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.TextView;

public class RdpViewGalleryAdapter extends BaseAdapter
{
    private static final int DEFAULT_SMALL_DESKTOP_SIZE = 180;
    
    private LayoutInflater mInflater;
    
    private List<RdpViewGalleryItem> views = null;
    private OnClickListener closeListener = null;

    public RdpViewGalleryAdapter( Context context, List<RdpViewGalleryItem> views, OnClickListener listener ) 
    {
        mInflater = LayoutInflater.from(context);
        this.views = views;
        this.closeListener = listener;
    }

    public long getItemId( int position )
    {
        return views.get( position ).id;
    }
    
    public int getCount() 
    {
        return views.size();
    }

    public Object getItem(int position) 
    {
        return views.get( position );
    }

    public View getView(int position, View convertView, ViewGroup parent) 
    {
        ViewHolder holder;

        if (convertView == null)
        {
            convertView = mInflater.inflate(R.layout.rdp_view_item, null);
            
            holder = new ViewHolder();
            holder.text = (TextView) convertView.findViewById(R.id.RdpGalleryFarmName);
            holder.desktop = (ImageView) convertView.findViewById(R.id.RdpGalleryImage);
            holder.closeButton = (ImageButton) convertView.findViewById( R.id.RdpGalleryClose );
            if ( closeListener != null ) 
            {
                holder.closeButton.setOnClickListener( closeListener );
                holder.closeButton.setTag( views.get( position ).id );
            }
            convertView.setTag(holder);
        }
        else 
        {
            holder = (ViewHolder) convertView.getTag();
        }
        
        Bitmap desktop = views.get( position ).bitmap;
        
        holder.desktop.setImageBitmap( desktop );
        holder.text.setText( views.get( position ).name );
        
        if ( desktop != null )
        {
            float modifier = (float) desktop.getWidth() / (float) desktop.getHeight();
            holder.desktop.setLayoutParams( new LinearLayout.LayoutParams( (int)(DEFAULT_SMALL_DESKTOP_SIZE * modifier), DEFAULT_SMALL_DESKTOP_SIZE ) );
        }

        return convertView;
    }
    
    private final class ViewHolder
    {
        private TextView  text = null;
        private ImageView desktop = null;
        private ImageButton closeButton = null;
    }
    
    public static final class RdpViewGalleryItem
    {
        private long id = 0;
        private String name = "";
        private Bitmap bitmap = null;
        
        public RdpViewGalleryItem( long id, String name, Bitmap bitmap )
        {
            this.id = id;
            this.name = name;
            this.bitmap = bitmap;
        }
    }    
}
