
package quest.android.appportal.view.custom;

import android.graphics.Matrix;
import android.graphics.PointF;
import android.view.MotionEvent;
import android.widget.ImageView;

public class ViewDragging
{
    public static final float STICK_DRAG_THRESHOLD = 8;
    public static final float MIN_PART_ON_SCREEN   = 50;

    ImageView                 view                 = null;

    Matrix                    matrix               = null;

    PointF                    start                = null;

    public ViewDragging( ImageView imageView, Matrix matrix )
    {
        this.view = imageView;
        this.matrix = matrix;
        this.start = new PointF();
    }

    public void startDragging( MotionEvent event )
    {
        start.set( event.getX(), event.getY() );
    }
    
    public void processDragging( MotionEvent event )
    {
        float[] values = new float[9];
        matrix.getValues( values );

        float dx = event.getX() - start.x;
        float dy = event.getY() - start.y;

        int viewWidth = view.getWidth();
        int viewHeight = view.getHeight();

        int width = view.getDrawable().getIntrinsicWidth();
        int height = view.getDrawable().getIntrinsicHeight();

        float scaledWidth = width * values[Matrix.MSCALE_X];
        float scaledHeight = height * values[Matrix.MSCALE_Y];

//        // Sticking to left / top
//        if ( Math.abs( values[Matrix.MTRANS_X] + dx ) < STICK_DRAG_THRESHOLD )
//        {
//            dx = -values[Matrix.MTRANS_X];
//        }
//        if ( Math.abs( values[Matrix.MTRANS_Y] + dy ) < STICK_DRAG_THRESHOLD )
//        {
//            dy = -values[Matrix.MTRANS_Y];
//        }
//
//        // Sticking to right / bottom
//        if ( Math.abs( values[Matrix.MTRANS_X] + scaledWidth + dx - viewWidth ) < STICK_DRAG_THRESHOLD )
//        {
//            dx = viewWidth - values[Matrix.MTRANS_X] - scaledWidth;
//        }
//        if ( Math.abs( values[Matrix.MTRANS_Y] + scaledHeight + dy - viewHeight ) < STICK_DRAG_THRESHOLD )
//        {
//            dy = viewHeight - values[Matrix.MTRANS_Y] - scaledHeight;
//        }
//
//        // Stop moving out the screen at left / top
//        if ( values[Matrix.MTRANS_X] + scaledWidth + dx < MIN_PART_ON_SCREEN )
//        {
//            dx = MIN_PART_ON_SCREEN - (values[Matrix.MTRANS_X] + scaledWidth);
//        }
//        if ( values[Matrix.MTRANS_Y] + scaledHeight + dy < MIN_PART_ON_SCREEN )
//        {
//            dy = MIN_PART_ON_SCREEN - (values[Matrix.MTRANS_Y] + scaledHeight);
//        }
//
//        // Stop moving out the screen at right / bottom
//        if ( values[Matrix.MTRANS_X] + dx > viewWidth - MIN_PART_ON_SCREEN )
//        {
//            dx = viewWidth - MIN_PART_ON_SCREEN - values[Matrix.MTRANS_X];
//        }
//        if ( values[Matrix.MTRANS_Y] + dy > viewHeight - MIN_PART_ON_SCREEN )
//        {
//            dy = viewHeight - MIN_PART_ON_SCREEN - values[Matrix.MTRANS_Y];
//        }
        
        if ( scaledWidth < viewWidth )
        {
            if ( values[Matrix.MTRANS_X] + dx < 0 )
            {
                dx = -values[Matrix.MTRANS_X];
            }
            else if ( values[Matrix.MTRANS_X] + scaledWidth + dx > viewWidth )
            {
                dx = viewWidth - (values[Matrix.MTRANS_X] + scaledWidth);
            }
        }
        else
        {
            if ( values[Matrix.MTRANS_X] + dx > 0 )
            {
                dx = -values[Matrix.MTRANS_X];
            }
            else if ( values[Matrix.MTRANS_X] + scaledWidth + dx < viewWidth )
            {
                dx = viewWidth - (values[Matrix.MTRANS_X] + scaledWidth);
            }
        }
        
        if ( scaledHeight < viewHeight )
        {
            if ( values[Matrix.MTRANS_Y] + dy < 0 )
            {
                dy = -values[Matrix.MTRANS_Y];
            }
            else if ( values[Matrix.MTRANS_Y] + scaledHeight + dy > viewHeight )
            {
                dy = viewHeight - (values[Matrix.MTRANS_Y] + scaledHeight);
            }
        }
        else
        {
            if ( values[Matrix.MTRANS_Y] + dy > 0 )
            {
                dy = -values[Matrix.MTRANS_Y];
            }
            else if ( values[Matrix.MTRANS_Y] + scaledHeight + dy < viewHeight )
            {
                dy = viewHeight - (values[Matrix.MTRANS_Y] + scaledHeight);
            }
        }

        matrix.postTranslate( dx, dy );
        view.setImageMatrix( matrix );
        
        start.set( event.getX(), event.getY() );
    }
}
